using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Xml;
using System.Net;
using System.IO;
using System.Windows.Forms;
using Microsoft.TeamFoundation.Server;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.WorkItemTracking.Client;
using Microsoft.TeamFoundation.VersionControl.Client;
using Microsoft.TeamFoundation.WorkItemTracking.Controls;
using CegTFSHelpers;
using CegWIManager;
using CegWICreator.Properties;
using System.Collections.Specialized;

namespace CegWICreator
{
	/// <summary>
	/// The main Form of this application, contains pretty much everything...
	/// </summary>
	public partial class MainForm : Form
	{
		#region DisableWIField class implementation
		/// <summary>
		/// Class that handles the disabling/enabling of Work Item Field change notifications.
		/// All notifications are disabled at construction, they're enabled again at destruction.
		/// This class is used when we want to disable GUI feedback during WI operations like saving.
		/// </summary>
		public class DisableWIField : IDisposable
		{
			/// <summary>
			/// Constructor
			/// </summary>
			/// <param name="owner">The Mainform of the application</param>
			public DisableWIField(MainForm owner)
			{
				m_Owner = owner;
				m_Owner.m_DisableFieldChangedNotif = true;
			}

			/// <summary>
			/// "Destructor"...
			/// </summary>
			public void Dispose()
			{
				m_Owner.m_DisableFieldChangedNotif = false;
			}

			MainForm		m_Owner;
		};
		#endregion

		#region Action class implementation
		/// <summary>
		/// The purpose of this class is to manage a user action regarding the GUI
		/// At construction it sets the Waiting Mouse Cursor and the given text in the status bar
		/// At destruction it updates the status bar with the given text (if any)
		/// </summary>
		public class Action : IDisposable
		{
			/// <summary>
			/// Constructor
			/// </summary>
			/// <param name="owner">The owner Form, has to be the MainForm</param>
			/// <param name="beginmsg">The message to display in the status bar at the start of the action</param>
			/// <param name="endmsg">The message to display in the status bar at the end of the action</param>
			public Action(MainForm owner, String beginmsg, String endmsg)
			{
				// Set the owner
				m_Owner = owner;

				// Push the action in the owner's stack
				m_Owner.m_Actions.Push(this);

				// Set the end message
				m_EndMsg = endmsg;

				// Update the status bar with the starting message
				m_Owner.StatusText.Text = beginmsg;
				m_Owner.StatusBar.Update();

				// Set the cursor to wait
				Cursor.Current = Cursors.WaitCursor;
			}

			/// <summary>
			/// Destructor
			/// </summary>
			public void Dispose()
			{
				// Pop from the stack
				m_Owner.m_Actions.Pop();
				if (m_EndMsg!=null)
				{
					m_Owner.StatusText.Text = m_EndMsg;
					m_Owner.StatusBar.Update();
				}
				Cursor.Current = Cursors.Default;					// Release the Wait Cursor by putting the default one
			}

			/// <summary>
			/// Contains the message to display at the end of the action (during Dispose()) in the status bar
			/// </summary>
			public	String		m_EndMsg;

			/// <summary>
			/// The owner Form, used to access to the Status Bar
			/// </summary>
					MainForm	m_Owner;
		};
		#endregion

		#region Variables and Properties
		/// <summary>
		/// Stack containing all the actions in process
		/// </summary>
		Stack<Action>			m_Actions = new Stack<Action>();

		/// <summary>
		/// The internal variable to the Team Foundation Server
		/// </summary>
		/// <remarks>DO NOT use this variable, use the <see cref="m_TFS"/> property instead.</remarks>
		TeamFoundationServer	_TFS		= null;

		/// <summary>
		/// The Team Foundation Server property.
		/// The get will return the TFS Object
		/// The set will set a new one, and the GUI will be updated of this change
		/// </summary>
		TeamFoundationServer	m_TFS
		{
			get
			{
				return _TFS;
			}
			set
			{
				_TFS  = value;
				m_WIP = null;
				SBServerName.Text = _TFS.Name;
			}
		}

		/// <summary>
		/// Work Item Store object (got from the m_TFS)
		/// </summary>
		WorkItemStore			m_WIS				= null;

		/// <summary>
		/// Version Control Server object (got from the m_TFS)
		/// </summary>
		VersionControlServer	m_SC				= null;

		WICObject				_Root				= null;

		XmlDocument				m_DocModel = null;

		/// <summary>
		/// The root object of the database
		/// </summary>
		WICObject				m_Root
		{
			get
			{
				return _Root;
			}

			set
			{
				if (_Root != null)
				{
					_Root.HierarchyChanged -= WICHierarchyChanged;
				}

				_Root = value;

				if (_Root != null)
				{
					_Root.HierarchyChanged += new WICObject.WICObjectHierarchyEventHandler(WICHierarchyChanged);
				}
			}
		}

		/// <summary>
		/// The Recycle Bin Work Item, if any.
		/// </summary>
		WorkItem				m_RecycleBin		= null;

		/// <summary>
		/// Area belonging to the recycle bin
		/// </summary>
		String					m_RecycleBinArea	= null;

		/// <summary>
		/// Network Model used by the application. This is mainly used to know what types of Work Item we can
		/// create under a given instance.
		/// </summary>
		WINetworkModel			m_NetModel			= null;

		/// <summary>
		/// Work Item Selection Form. The instance is always created, and showed up when the user asks for it.
		/// Otherwise it's hidden. This is done to keep the form settings through the use of the application.
		/// </summary>
		WISelectionForm			m_WISelForm			= null;

		/// <summary>
		/// The Dirty Counter internal variable
		/// </summary>
		int						_DirtyCounter		= 0;

		/// <summary>
		/// Dirty Counter property, this will update the GUI on Set operation.
		/// </summary>
		int						m_DirtyCounter
		{
			get
			{
				return _DirtyCounter;
			}
			set
			{
				_DirtyCounter		= value;
				m_SaveBtn.Enabled	= _DirtyCounter!=0;
			}
		}

		/// <summary>
		/// Determine if the database has changed since the last save.
		/// </summary>
		bool m_IsDirty
		{
			get
			{
				return _DirtyCounter != 0;
			}
		}

		/// <summary>
		/// If <c>true</c> Work Item Field Changed notification will be ignored, don't use directly that field, but
		/// use the <typeparamref name="DisableWIField"/> class instead
		/// </summary>
		bool			m_DisableFieldChangedNotif	= false;

		/// <summary>
		/// Internal variable for the selected WICObject.
		/// </summary>
		/// <remarks>DO NOT USE this variable, use the <see cref="m_SelWIC"/> property instead.</remarks>
		WICObject				_SelWIC		= null;

		/// <summary>
		/// The selected WICObject.
		/// The Get simply returns it.
		/// The Set also updates the GUI.
		/// </summary>
		WICObject				m_SelWIC
		{
			get
			{
				return _SelWIC;
			}
			set
			{
				if (_SelWIC == value)
				{
					return;
				}

				_SelWIC			= value;

				if (value != null)
				{
					m_WIForm.Item = value.WI;
				} else {
					m_WIForm.Item = null;
				}

				UpdateWIInfoBar(_SelWIC);
			}
		}

		/// <summary>
		/// The internal variable for the Work Item Project.
		/// </summary>
		/// <remarks>DO NOT USE this variable, use the <see cref="m_TP"/> property instead.</remarks>
		Project				   _WIP = null;

		/// <summary>
		/// Current Work Item Project
		/// Get will return the object.
		/// No set, this property is dependent from the Team Project.
		/// </summary>
		Project					m_WIP
		{
			get
			{
				return _WIP;
			}
			set
			{
				try
				{
					// Remove the Selected Work Item
					m_SelWIC = null;
					ClearWIDatabase();

					// Set the new value
					_WIP = value;

					// We tell the Network Model there's a Project change.
					if (m_NetModel != null)
					{
						m_NetModel.SetWIProject(_WIP);
					}

					// If there's an object
					if (_WIP != null)
					{
						// Try to locate the Recycle Bin Work Item (optional)
						int recid = 0;
						try
						{
							// Look for the XML Element
							XmlElement xrb = m_DocModel.SelectSingleNode(string.Format("WIModel/Projects/Project[@name=\"{0}\"]/RecycleBin", value.Name)) as XmlElement;
							if (xrb != null)
							{
								// Get the ID
								recid = int.Parse(xrb.GetAttribute("id"));

								// Get the Work Item
								// Note this Work Item is not managed by the WIChache of the NetModel
								m_RecycleBin = m_WIS.GetWorkItem(recid);

								// Get the Area Path (optional attribute)
								m_RecycleBinArea = xrb.GetAttribute("areapath");
							}
						}
						catch (Exception)
						{
							MessageBox.Show(String.Format("Invalid Recycle Bin ID [{0}] in the Configuration file, or Work Item inexistent", recid), "Work Item Creator", MessageBoxButtons.OK, MessageBoxIcon.Error);
						}

						// Retrieve information about the integration
						m_UserCanIntegrate = false;
						try
						{
							XmlElement xi = m_DocModel.SelectSingleNode(string.Format("WIModel/Projects/Project[@name=\"{0}\"]/Integration", value.Name)) as XmlElement;
							if (xi != null)
							{
								String introle = xi.GetAttribute("tfsrole");
								IGroupSecurityService gss = (IGroupSecurityService)m_TFS.GetService(typeof(IGroupSecurityService));
								Identity identity = gss.ReadIdentity(SearchFactor.AccountName, m_TFS.AuthenticatedUserName, QueryMembership.Direct);
								Identity devGroup = gss.ReadIdentity(SearchFactor.AccountName, introle, QueryMembership.None);

								foreach (string groupId in identity.MemberOf)
								{
									if (devGroup.Sid == groupId)
									{
										m_UserCanIntegrate = true;
										break;
									}
								}
							}
						}
						catch (Exception)
						{
						}

						// Build the Action Menu with this project
						BuildActionsMenu();

						// We can do things, so enable the Action menu
						m_ActionsMenu.Enabled = true;

						// Same for the Quick open
						m_QuickOpenWID.Enabled = true;

						// Update the status bar
						StatusText.Text = String.Format("Now using Team Project {0} from {1} logged as {2}", _WIP.Name, 
							m_TFS.Name, m_TFS.AuthenticatedUserDisplayName);
					}

					// We tell the Work Item Selection Form about the change too.
					if (m_WISelForm != null)
					{
						m_WISelForm.SetTeamProject(m_WIP);
					}

					// Update the Quick Open Button
					m_QuickOpenWID.Text = "";
					TBI_QuickWIDTextChanged(this, null);
				}
				
				catch (Exception ex)
				{
					StatusException(ex);
					return;
				}
			}
		}

		bool					m_UserCanIntegrate=false;

		#endregion

		#region StatusBar and Action misc methods
		/// <summary>
		/// Output an exception in the status bar
		/// </summary>
		/// <param name="ex"></param>
		private void StatusException (Exception ex)
		{
			MessageBox.Show(ex.ToString());
			StatusText.Text = ex.ToString();
			StatusBar.Update();

			// Parse the Action Stack and update the End Message of each entry, we want to be sure the message
			// is displayed
			foreach(Action action in m_Actions)
			{
				action.m_EndMsg = ex.ToString();
			}
		}

		/// <summary>
		/// Update the End Text of the Top Level action
		/// </summary>
		/// <param name="endmsg">The text that will replace the current one</param>
		public void UpdateTopLevelAction(String endmsg)
		{
			// Nothing on the stack, we display the message right away
			if (m_Actions.Count == 0)
			{
				StatusText.Text = endmsg;
				StatusBar.Update();
				return;
			}

			Action action = m_Actions.Peek();
			action.m_EndMsg = endmsg;
		}
		#endregion

		#region Construction and Load()
		/// <summary>
		/// Default constructor, initialize the GUI stuffs
		/// </summary>
		public MainForm()
		{
			InitializeComponent();
		}
		/// <summary>
		/// Load Event handler, called when the Form is first shown.
		/// </summary>
		/// <param name="sender">trigger object</param>
		/// <param name="e">parameter object</param>
		private void MainForm_Load(object sender, EventArgs e)
		{
			// Get the XML configuration file from the settings object
			String		wsurl		= Settings.Default.XmlConfig;
			String		xmodelurl;

			try
			{
				// Check if the URL points to a valid XML configuration file
				if (!SelectServer.CheckURL(wsurl))
				{
					// If it's not the case, give a change to the user to locate the configuration file directory
					SelectServer dlg = new SelectServer();
					if (dlg.ShowDialog() != DialogResult.OK)
					{
						throw new Exception("User couldn't find the server!");
					}

					// Set the new URL, and save it in the settings
					wsurl = dlg.m_URL.Text;
					Settings.Default.XmlConfig = wsurl;
					Settings.Default.Save();
				}

				// Append the XML configuration file
				xmodelurl = wsurl + @"/NetworkModels.xml";
			}

			// If there was an exception, it's probably due to the XML Configuration file, so we reset the setting and
			// save it.
			catch (Exception)
			{
				Settings.Default.XmlConfig="";
				Settings.Default.Save();
				throw;
			}
			
			// Get the configuration file from the given URL
			String		url = null;
			try
			{
				WebRequest request		= WebRequest.Create(xmodelurl);
				request.Credentials		= CredentialCache.DefaultCredentials;
				HttpWebResponse response= (HttpWebResponse)request.GetResponse();

				// Create and load the XML document
				m_DocModel = new XmlDocument();
				m_DocModel.Load(response.GetResponseStream());

				// Look for an optional TFSSettings element
				XmlElement	xtfsconfig = m_DocModel.SelectSingleNode("/WIModel/TFSSettings") as XmlElement;
				if (xtfsconfig != null)
				{
					// Get the URL of the Team Foundation Server
					url = xtfsconfig.GetAttribute("url");
				} else {
					// Couldn't find a TFS, so ask for one
					GetTFSURL	dlg = new GetTFSURL();
					if (dlg.ShowDialog(this)!=DialogResult.OK)
					{
						return;
					}

					url = dlg.m_TFSURL.Text;
				}

				StatusText.Text = String.Format("Loading configuration from {0}", xmodelurl);
			}
			catch (Exception)
			{
				MessageBox.Show(String.Format("Couldn't load the XML Model definition at {0}", xmodelurl), "Work Item Creator", MessageBoxButtons.OK, MessageBoxIcon.Error);
				throw;
			}

			try
			{
				// Connect to the Team Foundation Server, using the current credentials
				m_TFS = TeamFoundationServerFactory.GetServer(url);
				m_SC  = m_TFS.GetService(typeof(VersionControlServer)) as VersionControlServer;
				m_WIS = m_TFS.GetService(typeof(WorkItemStore)) as WorkItemStore;
//				IEventService EventSvc = (IEventService)m_TFS.GetService(typeof(IEventService));
//				Subscription[] sub = EventSvc.EventSubscriptions(m_TFS.AuthenticatedUserIdentity);

			}
			catch (Exception ex)
			{
				// Couldn't connect...
				MessageBox.Show(String.Format("Couldn't connect to the Team Foundation Server\r\n{0}", ex.ToString()), "Work Item Creator", MessageBoxButtons.OK, MessageBoxIcon.Error);
				throw;
			}

			// Initialize the WITree Image List
			XmlElement		xwi = m_DocModel.SelectSingleNode("/WIModel/WIIcons") as XmlElement;
			if (xwi != null)
			{
				// Remove all the images that were "factory added"
				TreeIL.Images.Clear();

				XmlNodeList xwis = m_DocModel.SelectNodes("/WIModel/WIIcons/WorkItem");

				// For each XML Element
				foreach (XmlElement xe in xwis)
				{
					try
					{
						// Make a Web Request to the file
						WebRequest request = WebRequest.Create(wsurl + "/" + xe.GetAttribute("location"));
						request.Credentials = CredentialCache.DefaultCredentials;
						HttpWebResponse response = (HttpWebResponse)request.GetResponse();

						// Read the answer in a memory stream (can't create the Icon directly from a ConnectStream
						// which is the type of the stream returned by GetResponseStream())
						byte[] buffer = new byte[10240];
						int read = response.GetResponseStream().Read(buffer, 0, 10240);
						MemoryStream stream = new MemoryStream (buffer);

						// Create the Icon from the Memory Stream, we want a 16*16 icon
						Icon ig = new Icon(stream, 16, 16);
						
						// Add the Icon to the list
						TreeIL.Images.Add(xe.GetAttribute("type"), ig);
					}
					catch (Exception)
					{
					}					
				}
			}

			// Enumerate all Team Project, add them in the project selection combobox
			ProjectCollection pc = m_WIS.Projects;
			m_ProjectSelection.Items.Clear();
			foreach (Project p in pc)
			{
				m_ProjectSelection.Items.Add(p.Name);
			}

			// As no project selected, we disable GUI triggers
			m_SaveBtn.Enabled		= false;
			m_ActionsMenu.Enabled	= false;
			m_QuickOpenWID.Enabled	= false;
			m_GoQuickOpen.Enabled	= false;

			// Create a new Network Model object, using the current WIS and the XML Configuration file we loaded
			m_NetModel = new WINetworkModel(m_WIS, m_DocModel);
			m_NetModel.m_WIPool.WIFieldChanged += new WorkItemFieldChangeEventHandler(WIFieldChanged);

			// Set the delegates to work the Hierarchy using WICObject objects instead of Work Items
			// It's important that the Network Model works with WICObjects, because we want him to work with newly
			// created Work Items (which have no ID but that are linked together)
			m_NetModel.SetDelegates(GetParentUsingWICObj, GetChildrenUsingWICObj, FromObjectUsingWICObj, CreateFromWIUsingWICObj);

			// We create the Work Item Selection Form, there will be only one instance, that will be showed up upon
			// user's request
			m_WISelForm = new WISelectionForm(m_TFS, m_WIS);

			// Restore the last used Team Project
			string lastteamproject = Settings.Default.LastTeamProject;
			if (lastteamproject!="")
			{
				if (m_ProjectSelection.Items.Contains(lastteamproject))
				{
					m_ProjectSelection.SelectedItem = lastteamproject;
				}
			}

			LoadMRU();
		}
		#endregion

		#region WICObject Delegates implementation

		/// <summary>
		/// Get the parent from a given object
		/// </summary>
		/// <param name="obj">The object to get the parent from</param>
		/// <returns>The parent or <c>null</c> if there's no parent or if <paramref name="obj"/> is invalid</returns>
		/// <remarks>This implementation will simply return the <c>Parent</c> member of the WICObject</remarks>
		private Object GetParentUsingWICObj(Object obj)
		{
			WICObject wic = obj as WICObject;
			if (wic == null)
			{
				return null;
			}

			return wic.Parent;
		}

		/// <summary>
		/// Get the Work Item from the given object
		/// </summary>
		/// <param name="obj">The object to get the Work Item from</param>
		/// <returns>The Work Item</returns>
		/// <remarks>This implementation return the Work Item stored in the <c>m_WorkItem</c> member of the WICObject</remarks>
		private WorkItem FromObjectUsingWICObj(Object obj)
		{
			if (obj == null)
			{
				return null;
			}
			
			WICObject wic = obj as WICObject;
			if (wic == null)
			{
				return null;
			}

			return wic.WI;
		}

		private Object CreateFromWIUsingWICObj(WorkItem wi, Object parent)
		{
			WICObject wic = new WICObject(m_NetModel, WICObjType.DBObj, wi);
			if (parent != null)
			{
				WICObject wicparent = parent as WICObject;
				if (wicparent != null)
				{
					wicparent.AddChild(wic);
				}
			}

			return wic;
		}

		/// <summary>
		/// Evaluate all the children of the given object
		/// </summary>
		/// <param name="obj">The object to enumerate all the children</param>
		/// <returns>The collection containing WICObject instances of all the children of <paramref name="obj"/></returns>
		/// <remarks>Progressive loading in implemented here. WICObjects are created (and Work Item loaded) the first
		/// time this method is called for a given object</remarks>
		private IEnumerable GetChildrenUsingWICObj(Object obj)
		{
			WICObject wic = obj as WICObject;
			if (wic == null)
			{
				return null;
			}

			// Load children if it's not already made
			if (!wic.ChildrenLoaded)
			{
				wic.LoadChildren();
			}

			// Return the children
			return wic.Children;
		}

		#endregion

		#region Event handlers

		/// <summary>
		/// Event triggered when the Team Project selection in the ComboBox changes
		/// </summary>
		/// <param name="sender">ComboBox</param>
		/// <param name="e">Parameters</param>
		private void TeamProjectChanged(object sender, EventArgs e)
		{
			// Ask to save, and issue a save if needed
			CheckWIToSave(false, true);

			// We clear the database
			ClearWIDatabase();

			try
			{
				if (CheckWIP((String)m_ProjectSelection.SelectedItem) == false)
				{
					return;
				}

				// Get the new Team Project and update the Property
				m_WIP = m_WIS.Projects[(String)m_ProjectSelection.SelectedItem];
			}
			catch (Exception ex)
			{
				ex.ToString();
				return;
			}

			Settings.Default.LastTeamProject = (String)m_ProjectSelection.SelectedItem;
			Settings.Default.Save();
		}

		/// <summary>
		/// Handler: Work Item selection changed from the Tree Control
		/// </summary>
		/// <param name="sender">...</param>
		/// <param name="e">...</param>
		/// <remarks>This method is called when the user changes the selection in the Tree Control.
		/// We simply update the m_SelWIC property.</remarks>
		private void WITree_ItemSelected(object sender, TreeViewEventArgs e)
		{
			m_SelWIC = e.Node.Tag as WICObject;
		}

		/// <summary>
		/// Handler: TreeNode mouse click
		/// </summary>
		/// <param name="sender">...</param>
		/// <param name="e">...</param>
		/// <remarks>We track right mouse click to build and display the context menu.</remarks>
		private void WITree_NodeClick(object sender, TreeNodeMouseClickEventArgs e)
		{
			// Right click, create the contextual menu and display it
			if (e.Button == MouseButtons.Right)
			{
				// Select the item
				m_WITree.SelectedNode = e.Node;

				// Make the menu
				BuildCtxMenuFor(e.Node, m_WITree.ContextMenuStrip);
			}
		}

		/// <summary>
		/// Handler: Tree Control's context menu item click.
		/// </summary>
		/// <param name="sender">...</param>
		/// <param name="e">...</param>
		/// <remarks>We simply route the Item to the corresponding commands</remarks>
		private void WITree_ItemClicked(object sender, ToolStripItemClickedEventArgs e)
		{
			try
			{
				if (e.ClickedItem.Name == "SetAsRoot")
				{
					CmdSetAsRoot(m_WITree.SelectedNode);
				}
				if (e.ClickedItem.Name == "Unlink")
				{
					CmdUnlinkWI(m_WITree.SelectedNode);
					return;
				}

				if (e.ClickedItem.Name == "LinkWI")
				{
					CmdLinkExistingWI(m_WITree.SelectedNode);
					return;
				}

				if (e.ClickedItem.Name == "SendToRecycle")
				{
					CmdSendToRecycle(m_WITree.SelectedNode);
					return;
				}

				if (e.ClickedItem.Name == "DeleteWI")
				{
					CmdDeleteWI(m_WITree.SelectedNode);
					return;
				}

				if (e.ClickedItem.Name == "CancelModifications")
				{
					CmdCancelModifications(m_WITree.SelectedNode);
					return;
				}

				if (e.ClickedItem.Name == "UpdateAreaPath")
				{
					CmdUpdateAreaPath(m_WITree.SelectedNode);
					return;
				}

				if (e.ClickedItem.Name == "UpdateIteration")
				{
					CmdUpdateIteration(m_WITree.SelectedNode);
					return;
				}

			}
			catch (Exception ex)
			{
				StatusException(ex);
			}
		}

		/// <summary>
		/// Handler: TreeNode expanding event
		/// </summary>
		/// <param name="sender">...</param>
		/// <param name="e">...</param>
		/// <remarks>We make sure the children are loaded!</remarks>
		private void WITree_Expanding(object sender, TreeViewCancelEventArgs e)
		{
			WICObject wic = e.Node.Tag as WICObject;
			if (wic != null)
			{
				wic.Expanded = true;
				if (wic.ChildrenLoaded == false)
				{
					using (new Action(this, String.Format("Loading children of {0}", wic.WI.Title), "Done"))
					{
						wic.LoadChildren();
						UpdateWITree(wic, true, true);
					}
				}
				else
				{
					if ((e.Node.Nodes.Count==1) && (e.Node.Nodes[0].Name=="DummyNode"))
					{
						UpdateWITree(wic, true, true);
					}
				}
			}
		}

		/// <summary>
		/// Update the expand/collapse state of the corresponding WICObject
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void WITree_Collapsing(object sender, TreeViewCancelEventArgs e)
		{
			WICObject wic = e.Node.Tag as WICObject;
			if (wic != null)
			{
				wic.Expanded = false;
			}
		}

		/// <summary>
		/// Handler: A field of a loaded Work Item has changed
		/// </summary>
		/// <param name="sender">...</param>
		/// <param name="e">...</param>
		/// <remarks>This method can be filtered if <paramref name="m_DisabledFieldChangeNotif"/> is set to true.
		/// The main purpose is to update the Dirty Counter and the GUI.</remarks>
		void WIFieldChanged(object sender, WorkItemEventArgs e)
		{
			// Notification to ignore ?
			if (m_DisableFieldChangedNotif == true)
			{
				return;
			}

			try
			{
				// Update the Filter Tree
				if ((e!=null) && (e.Field!=null))
				{
					m_WIFilter.UpdateField(e.Field);
				}

				// Get the corresponding Work Item
				WorkItem wi = sender as WorkItem;
				IEnumerable<WICObject> list = WICsFromWI(wi);

				foreach (WICObject cur in list)
				{
					// Update the Dirty counter/state
					UpdateDirty(cur, UpdateMode.UM_Modify, false);

					// Update the GUI
					UpdateWIGUI(cur);
				}
			}
			catch (Exception ex)
			{
				StatusException(ex);
			}
		}

		/// <summary>
		/// Handler to change the TFS Server
		/// </summary>
		/// <param name="sender">...</param>
		/// <param name="e">...</param>
		private void MI_ChangeServer(object sender, EventArgs e)
		{
			SelectServer dlg = new SelectServer();
			if (dlg.ShowDialog(this) != DialogResult.OK)
			{
				return;
			}
			Settings.Default.XmlConfig = dlg.m_URL.Text;
			Settings.Default.Save();

			MainForm_Load(this, null);
		}

		/// <summary>
		/// Handler called when the user wants to exit
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			e.Cancel = CheckWIToSave(false, false) == false;
		}

		#endregion

		#region TreeView implementation
		/// <summary>
		/// Create a Tree Item from a WICObject
		/// </summary>
		/// <param name="wic">The WICObject to create the Tree Item from</param>
		/// <param name="createdummy">If <c>true</c> a Dummy child will be also created</param>
		/// <returns>The created TreeNode</returns>
		TreeNode CreateTreeItem (WICObject wic, bool createdummy)
		{
			// Get the label
			String		name	= GetTreeItemName(wic);

			// Create the node
			TreeNode	tn		= new TreeNode(name);

			// Get the Work Item
			WorkItem	wi		= wic.WI;

			// Associate it with the given Work Item
			tn.Tag = wic;

			// Picture index from the Work Item's type
			int ii = TreeIL.Images.IndexOfKey(wi.Type.Name);
			if (ii==-1)
			{
				ii = TreeIL.Images.IndexOfKey("Default");
			}

			tn.ImageIndex = tn.SelectedImageIndex = ii;

			// If the Work Item is invalid, the label is displayed in Red
			if ((wi.IsOpen==true) && (wi.IsValid()==false))
			{
				tn.ForeColor = Color.Red;
			}
			
			// Create a dummy child to have a Expand/Collapse box ?
			if (createdummy)
			{
				TreeNode dummytn = tn.Nodes.Add("DummyNode");
				dummytn.Name = "DummyNode";
			}

			// Return the create TreeNode
			return tn;
		}

		/// <summary>
		/// Helper that formats the name of a Tree Item
		/// </summary>
		/// <param name="wic">WICObject to get the Tree Item label</param>
		/// <returns>The label</returns>
		String GetTreeItemName (WICObject wic)
		{
			WorkItem wi = wic.WI;
			return String.Format("{0}[{1}] {2}-{3}", wi.IsDirty ? "* " : "", (wi.Id == 0) ? "New" : wi.Id.ToString(), wi.Title, wi.State);
		}

		/// <summary>
		/// Update the application's TreeView
		/// </summary>
		/// <param name="wic">The update starts from the TreeNode corresponding to the given WIC Object</param>
		/// <param name="removedummy">Check if there's dummy TreeNode, and remove them if needed</param>
		/// <param name="recurse"><c>true</c> to also update children.</param>
		void UpdateWITree(WICObject wic, bool removedummy, bool recurse)
		{
			// Checking
			if (wic == null)
			{
				return;
			}

			// Deactivate the rendering on the Tree Control, because we're going to insert a lot of stuffs
			m_WITree.BeginUpdate();

			try
			{
				_UpdateWITree(wic, removedummy, recurse);
			}
			catch (Exception ex)
			{
				StatusException(ex);
			}

			// Reactivate the rendering of the control
			m_WITree.EndUpdate();
		}

		/// <summary>
		/// Recursive part of UpdateWITree.
		/// </summary>
		/// <param name="wic">See UpdateWITree.</param>
		/// <param name="removedummy">See UpdateWITree.</param>
		/// <param name="recurse">See UpdateWITree.</param>
		void _UpdateWITree(WICObject wic, bool removedummy, bool recurse)
		{
			// Filtered
			if (wic.IsFiltered(true) || m_WIFilter.IsFiltered(wic, true))
			{
				// If there's something, remove the whole subtree and quit
				if (wic.TN != null)
				{
					wic.TN = null;
				}

				return;
			}

			// If there's no TreeNode, create one and insert it at the right place
			if (wic.TN == null)
			{
				// Create the TreeNode
				wic.TN = CreateTreeItem(wic, m_NetModel.GetFieldFirstChildWI(wic.WI) != 0);
			}

			if (wic.TN.Parent == null)
			{
				// If the wic has no parent, it's the root, so it's a special case
				if (wic == m_Root)
				{
					if (m_WITree.Nodes.Contains(wic.TN) == false)
					{
						m_WITree.Nodes.Add(wic.TN);
					}
				}

				// Get the parent, and insert the TreeNode at the right place
				else
				{
					TreeNode parenttn = wic.Parent.TN;
					WICObject prev = wic.Previous;

					if (prev == null)
					{
						parenttn.Nodes.Insert(0, wic.TN);
					}
					else
					{
						int ii = parenttn.Nodes.IndexOf(prev.TN);
						if ((ii + 1) == parenttn.Nodes.Count)
						{
							parenttn.Nodes.Add(wic.TN);
						}
						else
						{
							parenttn.Nodes.Insert(ii + 1, wic.TN);
						}
					}
				}
			}
			else
			{
				wic.TN.Text = GetTreeItemName(wic);
			}

			// Change the Text Color is the TreeNode is filtered
			if (wic.Type == WICObjType.DBObj)
			{
				// Change the Text Color only if the Work Item is valid
				// If it's invalid, keep the red color
				if (wic.WI.IsValid())
				{
					if (wic.IsFiltered(false) || m_WIFilter.IsFiltered(wic, false))
					{
						wic.TN.ForeColor = Color.Gray;
					}
					else
					{
						wic.TN.ForeColor = Color.Black;

						// If we're in filtering mode ensure the item is visible
// 						if (IsFiltering())
// 						{
// 							wic.TN.EnsureVisible();
// 						}
					}
				}
			}

			// We must remove the dummy ?
			if (removedummy && wic.ChildrenLoaded && (wic.TN.Nodes.Count > 0))
			{
				TreeNode dummytn = wic.TN.Nodes[0];
				if (dummytn.Name == "DummyNode")
				{
					dummytn.Remove();
				}
			}

			// Recurse if required
			if (recurse)
			{
				foreach (WICObject cur in wic.Children)
				{
					_UpdateWITree(cur, removedummy, true);
				}
			}

			// To Expand ?
			if (wic.Expanded)
			{
				wic.TN.Expand();
			}
			else
			{
				wic.TN.Collapse();
			}
		}

		/// <summary>
		/// Determine if WITree filtering is on.
		/// </summary>
		/// <returns>Returns <c>true</c> if filtering is on.</returns>
		bool IsFiltering()
		{
			return ((m_TitleFilter.Text != "") && (m_TitleFilter.Text != "<Title Filter>")) ||
					 m_WIFilter.IsFiltering();
		}

		/// <summary>
		/// Setup a contextual menu for the given TreeNode
		/// </summary>
		/// <param name="tn">The TreeNode to create the contextual menu from</param>
		/// <param name="menu">The menu to add the items</param>
		private void BuildCtxMenuFor(TreeNode tn, ContextMenuStrip menu)
		{
			WICObject wic = tn.Tag as WICObject;
			WorkItem wi = wic.WI;
			WorkItem parentwi = m_NetModel.WIGetParent(wi);
			bool previtem = false;
			ToolStripItem it;

			// Clear the menu
			menu.Items.Clear();

			// If there's a parent, add the "Unlink" command
			if (parentwi == null)
			{
				if (previtem)
				{
					menu.Items.Add(new ToolStripSeparator());
				}

				it = menu.Items.Add("Set as Root");
				it.Name = "SetAsRoot";
				it.Tag = wic;
			}
			else
			{
				it = menu.Items.Add("Unlink from Parent");
				it.Name = "Unlink";
				it.Tag = wic;
			}

			// Recycle Bin
			if (m_RecycleBin != null)
			{
				it = menu.Items.Add("Send to Recycle Bin");
				it.Name = "SendToRecycle";
				it.Tag = wic;
			}

			// Add "Link existing"
			it = menu.Items.Add("Link existing Work Item");
			it.Name = "LinkWI";
			it.Tag = wic;

			// If it's brand new Work Item, we still can delete it
			if (wi.Id == 0)
			{
				it = menu.Items.Add("Delete");
				it.Name = "DeleteWI";
				it.Tag = wic;
			}

			// Else if it's modified, we proposed to cancel these modifications
			else if (wi.IsDirty)
			{
				it = menu.Items.Add("Cancel modifications");
				it.Name = "CancelModifications";
				it.Tag = wic;
			}

			// Separator
			menu.Items.Add(new ToolStripSeparator());

			// Update Children's area
			it = menu.Items.Add("Update children's Area Path");
			it.Name = "UpdateAreaPath";
			it.Tag = wic;

			// Update Children's iteration
			it = menu.Items.Add("Update chidren's Iteration");
			it.Name = "UpdateIteration";
			it.Tag = wic;


			previtem = true;

			// Sub Work Item creation
			try
			{
				// Get the list of the WIT we can create for the given WIC
				ArrayList witc = m_NetModel.GetCreableWITs(m_WIP.Name, wic);

				// There must be at least one
				if (witc.Count > 0)
				{
					// Add a separator if needed
					if (previtem)
					{
						menu.Items.Add(new ToolStripSeparator());
					}

					// Create the sub menu item
					ToolStripMenuItem micreate = new ToolStripMenuItem("Create Sub Work Item");
					micreate.Owner = menu;
					micreate.DropDownItemClicked += new ToolStripItemClickedEventHandler(WITree_CreateSubWI);
					menu.Items.Add(micreate);

					// Iterate through all the available WITs
					foreach (WorkItemType witype in witc)
					{
						// Create the menu item
						it = micreate.DropDownItems.Add(witype.Name);
						it.Name = witype.Name;
						it.Tag = wic;
					}

					bool firstitem = true;

					foreach (KeyValuePair<string, WITemplate> kvp in m_NetModel.m_WITemplates)
					{
						WITemplate witpl = kvp.Value;

						if (witpl.IsForProject(m_WIP.Name) && witpl.IsRootOfType(witc))
						{
							if (firstitem)
							{
								firstitem = false;
								micreate.DropDownItems.Add(new ToolStripSeparator());
							}

							it = micreate.DropDownItems.Add(witpl.Name);
							it.ToolTipText = witpl.Comments;
							it.Name = witpl.Name;
							it.Tag = witpl;
						}
					}
				}
			}
			catch (Exception ex)
			{
				StatusException(ex);
			}
		}

		void BuildActionsMenu()
		{
			// Sub Work Item creation
			try
			{
				m_ActionsMenu.DropDownItems.Clear();

				ToolStripItem it;
				it = m_ActionsMenu.DropDownItems.Add("Open Database from Work Item");
				it.Name = "OpenDB";

				// Get the list of the WIT we can create for the given WIC
				ArrayList witc = m_NetModel.GetCreableWITs(m_WIP.Name, null);

				// Iterate through all the available WITs
				foreach (WorkItemType witype in witc)
				{
					// Create the menu item
					it = m_ActionsMenu.DropDownItems.Add(String.Format("Create {0}", witype.Name));
					it.Name = witype.Name;
				}

				// The user can integrate ? Add a menu item
				if (m_UserCanIntegrate)
				{
					m_ActionsMenu.DropDownItems.Add(new ToolStripSeparator());
					it = m_ActionsMenu.DropDownItems.Add("Make a new release...");
					it.Name = "MakeRelease";

					it = m_ActionsMenu.DropDownItems.Add("Branches Comparison...");
					it.Name = "BranchesComparison";

					it = m_ActionsMenu.DropDownItems.Add("Check database integrity...");
					it.Name = "CheckDB";
				}
			}
			catch (Exception ex)
			{
				StatusException(ex);
			}
		}

		void WITree_CreateSubWI(object sender, ToolStripItemClickedEventArgs e)
		{
			WICObject wic = e.ClickedItem.Tag as WICObject;
			if (wic != null)
			{
				WorkItemType wit = m_WIP.WorkItemTypes[e.ClickedItem.Name];
				if (wit == null)
				{
					UpdateTopLevelAction(String.Format("Unknown Work Item Type: {0}", e.ClickedItem.Name));
					return;
				}
				WorkItem wi = CreateWorkItem(m_WITree.SelectedNode, wit);
				return;
			}

			WITemplate	witpl = e.ClickedItem.Tag as WITemplate;
			if (witpl != null)
			{
				WICObject wicp = m_WITree.SelectedNode.Tag as WICObject;
				wicp.LoadChildren();
				
				wic = m_NetModel.ApplyTemplate(witpl, wicp) as WICObject;

				UpdateDirty(wic, UpdateMode.UM_Modify, false);
				UpdateWITree(wicp, true, true);

				m_WITree.SelectedNode = wic.TN;

				return;
			}
		}

		/// <summary>
		/// Refresh the WITree from the given TN
		/// </summary>
		/// <param name="tn">The TreeNode to start the refresh from</param>
		/// <remarks>This method is recursive</remarks>
		private void RefreshWIs(TreeNode tn)
		{
			WICObject wic = tn.Tag as WICObject;
			WorkItem wi = wic.WI;
			if (wi == null)
			{
				return;
			}

			// Get the latest version
			wi.SyncToLatest();
			tn.Text = GetTreeItemName(wic);

			// Refresh the children
			foreach (TreeNode cur in tn.Nodes)
			{
				if (cur.Name != "DummyNode")
				{
					RefreshWIs(cur);
				}
			}
		}

		#endregion

		#region Database management

		/// <summary>
		/// Clear the whole database and update the GUI
		/// </summary>
		void	ClearWIDatabase()
		{
			// Clear/reset everything
			m_WITree.Nodes.Clear();
			m_Root = null;
			if (m_NetModel!=null)
			{
				m_NetModel.m_WIPool.Clear();
			}
			m_DirtyCounter = 0;
			m_TitleFilter.Text = "<Title Filter>";
			m_TitleFilter.ForeColor = Color.Gray;
			m_WITree.Focus();
			m_WIFilter.Clear();
		}

		/// <summary>
		/// Construct the whole WICObject database belonging to the given Work Item
		/// </summary>
		/// <param name="wi">The Work Item to construct the database from.</param>
		/// <returns>The WICObject corresponding to <paramref name="wi"/></returns>
		/// <remarks>This method will go up in the hierarchy to find the root and
		/// will build everything from the root to <paramref name="wi"/>. <paramref name="wi"/> will be
		/// selected and ensure visible</remarks>
		WICObject BuildWIDatabase(WorkItem wi)
		{
			// Nothing to do ? Exit.
			if (wi == null)
			{
				return null;
			}

			// Start from the given Work Item, find the Root, and store the whole path between these two
			List<WorkItem> wilist = new List<WorkItem>();
			
			WorkItem	curwi = wi;
			while (curwi != null)
			{
				wilist.Insert(0, curwi);				// As we go up in the hierarchy, we insert always at the first position
				curwi = m_NetModel.WIGetParent(curwi);	// to keep the right direction after
			}

			WICObject	obj = null;
			try
			{
				// Create the root object
				m_Root	= new WICObject(m_NetModel, WICObjType.DBObj, wilist[0]);
				obj		= m_Root;
				UpdateDirty(m_Root, UpdateMode.UM_Modify, false);

				// Update the Tree Control
				UpdateWITree(obj, false, true);

				// Process every level from the root to the one for which we called this method
				for (int i = 1; i < wilist.Count; i++)
				{
					obj.LoadChildren();

					// Get the children
					WICObjectList children = obj.Children;

					// Update the Tree Control
					UpdateWITree(obj, true, true);

					// Parse them to find the one we're looking for to make our path to the requested wi
					foreach (WICObject curobj in children)
					{
						// This is the WI we're looking for in the current level ? Break and go to the next level.
						if (curobj.WI.Id == wilist[i].Id)
						{
							obj = curobj;
							break;
						}
					}
				}

				// Make sure the TreeNode of the requested Work item is visible, and select it
				obj.TN.EnsureVisible();
				m_WITree.SelectedNode = obj.TN;
			}
			catch (Exception ex)
			{
				StatusException(ex);
			}

			// Returns the TreeNode corresponding to the Work Item we got as a parameter
			return obj;
		}

		/// <summary>
		/// Dirty Update modes
		/// </summary>
		enum UpdateMode
		{
			/// <summary>
			/// The object's content is modified
			/// </summary>
			UM_Modify = 1,

			/// <summary>
			/// The object is newly added to the database
			/// </summary>
			UM_Add = 2,

			/// <summary>
			/// The object is removed from the database
			/// </summary>
			UM_Remove = 3
		}

		/// <summary>
		/// Update the dirty counter
		/// </summary>
		/// <param name="wic">The WICObject that causes the update</param>
		/// <param name="mode">The update mode, <see cref="UpdateMode"/>.</param>
		/// <param name="recursive">If <c>true</c> the method will recurse to the children of <paramref name="wic"/></param>
		void UpdateDirty(WICObject wic, UpdateMode mode, bool recursive)
		{
			// Get the work item
			WorkItem wi = wic.WI;

			if (wi != null)
			{
				// Remove case
				if (mode == UpdateMode.UM_Remove)
				{
					// If the Work Item to be removed is new, we decrement the DirtyCounter as its one object less to be saved
					if (wi.Id == 0)
					{
						m_DirtyCounter--;
					}

					// If the Work Item were already saved, it will move to the ToRecycle list, and the DirtyCounter
					// will be update during the next save.
				}

				// Modify case
				else if (mode == UpdateMode.UM_Modify)
				{
					// The dirty state of the Work Item changed ?
					if (wic.DirtyState != wi.IsDirty)
					{
						// Update the dirty counter
						m_DirtyCounter += wi.IsDirty ? 1 : -1;

						// Update the dirty state
						wic.DirtyState = wi.IsDirty;
					}
				}

				// Add case
				else if (mode == UpdateMode.UM_Add)
				{
					// There's a possibility the WI is already dirty (due to a previous unlink, then a relink done here)
					// If it's not the case, then it's a first add, so we update the Dirty Counter and State
					if (wic.DirtyState == false)
					{
						wic.DirtyState = true;		// Set the state as dirty because the WI will have its hierarchy fields changed
						m_DirtyCounter++;
					}
				}
			}

			// Recurse if needed
			if (recursive)
			{
				foreach (WICObject cur in wic.Children)
				{
					UpdateDirty(cur, mode, true);
				}
			}
		}

		/// <summary>
		/// Load all the Work Item of the database
		/// </summary>
		/// <param name="wic">WICObject to start the loading from.</param>
		/// <remarks>This method is mainly used when the filtering mode is activated for the first time</remarks>
		private void LoadAll(WICObject wic)
		{
			if (!wic.ChildrenLoaded)
			{
				using (new Action(this, String.Format("Loading children of {0}", wic.WI.Title), "Done"))
				{
					wic.LoadChildren();
				}
			}

			foreach(WICObject cur in wic.Children)
			{
				LoadAll(cur);
			}
		}

		#endregion

		#region Migration from old hierarchy format to the new one
		WICache		m_MigCache = null;

		/// <summary>
		/// Migration of the "Link based" hierarchy to the "Fields based" one
		/// </summary>
		/// <param name="wiid">The ID of the Work Item to migrate</param>
		/// <returns>Success flag</returns>
		bool DoHierarchyMigration(int wiid)
		{
			{
				// Get the Work Item
				WorkItem wi = m_WIS.GetWorkItem(wiid);
				if (wi == null)
				{
					// Nothing to migrate...Still true is returned
					return true;
				}

				// Find the Root
				while (!m_NetModel.IsRootWorkItem(wi))
				{
					wi = m_WIS.GetWorkItem(m_NetModel.GetFieldParentWI(wi));
				}

				// Update the ID too
				wiid = wi.Id;

				// First we check if there's a FirstSonWI is different of zero, if it's the case, the migration were
				// already made
				if (m_NetModel.GetFieldFirstChildWI(wi) != 0)
				{
					// Nothing to do, it's the new format
					return true;
				}

				// Now we look if there's at least one child
				bool nochildren = true;
				foreach (Link link in wi.Links)
				{
					// <WIManager> en commentaire pour signifier un lien  grer
					if (link.Comment == "<WIManager>")
					{
						nochildren = false;
						break;
					}
				}

				// Nothing ? We quit
				if (nochildren)
				{
					return true;
				}
			}

			// Ok, now we know we have to migrate
			if (MessageBox.Show("The database needs to migrate to the new hierarchy format, you must do it before continuing",
				"Cegedim Work Item Creator", MessageBoxButtons.OKCancel, MessageBoxIcon.Warning) == DialogResult.Cancel)
			{
				return false;
			}

			// Create the WICache
			m_MigCache = new WICache(m_WIS, true, false);

			// Let go migrate !
			try
			{
				WorkItem wi = m_MigCache.Get(wiid);
				using (new Action(this, "Setting up new hierarchy format", "Done.."))
				{
					MigrateHierarchy(wi);
				}

				using (new Action(this, "Removing old hierarchy format", "Done.."))
				{
					MigrateRemoveOldHierarchy(wi);
				}

				using (new Action(this, "Saving Work Item Database", "Done.."))
				{
					String res = "";
					m_MigCache.Save(ref res);
				}

			}
			catch (Exception ex)
			{
				MessageBox.Show(String.Format("ERROR: Couldn't perform the migration!\r\n{0}", ex.ToString()), 
					"Cegedim Work Item Creator", MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
			m_MigCache = null;
			
			return true;
		}

		/// <summary>
		/// Update the hierarchy of the given Work Item and its children.
		/// </summary>
		/// <param name="wi">Work Item to update</param>
		/// <remarks>This method will set up the new hierarchy fields (FirstChildWI and NextWI)</remarks>
		void MigrateHierarchy(WorkItem wi)
		{
			bool		firstchild	= true;
			WorkItem	previous	= null;

			// Parse the links to find the children
			foreach (Link link in wi.Links)
			{
				// <WIManager> in comment, this is a link to change
				if (link.Comment == "<WIManager>")
				{
					// Get the associate Work Item
					RelatedLink rl = link as RelatedLink;
					if (rl == null)
					{
						continue;
					}
					WorkItem linkedwi = m_MigCache.Get(rl.RelatedWorkItemId);
					if (linkedwi == null)
					{
						continue;
					}

					// Check if the link points to a child
					if (!m_NetModel.IsFatherAndSon(wi, linkedwi))
					{
						continue;
					}

					// Open the Work Item for editing
					linkedwi.Open();

					// Reset the Next WI now, it will be modified the next iteration, except for the last WI
					m_NetModel.SetFieldNextWI(linkedwi, 0);

					// First ?
					if (firstchild)
					{
						m_NetModel.SetFieldFirstChildWI(wi, linkedwi.Id);
						firstchild = false;
					}

					// Another
					else
					{
						m_NetModel.SetFieldNextWI(previous, linkedwi.Id);
					}

					// Recurse
					MigrateHierarchy(linkedwi);

					// New previous
					previous = linkedwi;
				}
			}
		}

		/// <summary>
		/// Remove the previous hierarchy data (WI links)
		/// </summary>
		/// <param name="wi">Work Item to remove the data</param>
		/// <remarks>This method is recursive</remarks>
		void MigrateRemoveOldHierarchy(WorkItem wi)
		{
			// Parse the links to remove the ones about the hierarchy
			for (int i=0; i<wi.Links.Count;)
			{
				Link link = wi.Links[i];

				// <WIManager> in comment, this is a link to change
				if (link.Comment == "<WIManager>")
				{
					// Get the associate Work Item
					RelatedLink rl = link as RelatedLink;
					if (rl == null)
					{
						i++;
						continue;
					}
					WorkItem linkedwi = m_MigCache.Get(rl.RelatedWorkItemId);
					if (linkedwi == null)
					{
						i++;
						continue;
					}

					// Check if the link points to a child
					if (!m_NetModel.IsFatherAndSon(wi, linkedwi))
					{
						i++;
						continue;
					}

					// Remove the link
					wi.Links.Remove(link);
				} else {
					i++;
				}
			}

			// Recurse children with the new hierarchy
			WorkItem cur = m_MigCache.Get(m_NetModel.GetFieldFirstChildWI(wi));
			while (cur != null)
			{
				MigrateRemoveOldHierarchy(cur);

				cur = m_MigCache.Get(m_NetModel.GetFieldNextWI(cur));
			}
		}
		#endregion

		#region Work Item creation

		/// <summary>
		/// Generic method of Work Item creation
		/// </summary>
		/// <param name="parent">TreeNode of the parent</param>
		/// <param name="wit">The type of Work Item to create</param>
		/// <returns>The newly created Work Item</returns>
		/// <remarks>This method will copy some fields (Iteration, Area) from the parent Work Item</remarks>
		private WorkItem CreateWorkItem(TreeNode parent, WorkItemType wit)
		{
			try
			{
				// Get the WIC Parent
				WICObject wicparent		= parent.Tag as WICObject;
				wicparent.LoadChildren();

				// Create the Work Item, copy some fields from its parent
				WorkItem wiparent		= wicparent.WI;
				WorkItem wi				= m_NetModel.m_WIPool.Create(wit);

				wi.Title				= String.Format("{0} #{1}", wit.Name, wicounter++);
				wi.IterationPath		= wiparent.IterationPath;
				wi.IterationId			= wiparent.IterationId;
				wi.AreaPath				= wiparent.AreaPath;
				wi.AreaId				= wiparent.AreaId;

				// Create a WIC for the new WorkItem
				WICObject wic			= new WICObject(m_NetModel, WICObjType.DBObj, wi);
				UpdateDirty				(wic, UpdateMode.UM_Modify, false);
				wicparent.AddChild		(wic);
				UpdateWITree			(wicparent, true, true);
				m_WITree.SelectedNode	= wic.TN;

				return wi;

			}
			catch (Exception ex)
			{
				StatusException(ex);
			}
			return null;
		}

		/// <summary>
		/// Internal counter to create unique names for the Work Items
		/// </summary>
		static int wicounter = 1;

		private void MIActionsMenuClicked(object sender, ToolStripItemClickedEventArgs e)
		{
			if (e.ClickedItem.Name=="OpenDB")
			{
				if (!CheckWIToSave(false, false))
				{
					return;
				}

				// We clear the database
				ClearWIDatabase();

				if (m_WISelForm.ShowDialog() == DialogResult.OK)
				{
					using (new Action(this, "Opening Work Item Database...", "Database opened"))
					{
						// Make sure the Opened Database is to the new hierarchy format
						if (!DoHierarchyMigration(m_WISelForm.m_SelectedWorkItemID))
						{
							UpdateTopLevelAction("Migration cancel, databased opening aborted!");
							return;
						}

						// Build the Tree
						BuildWIDatabase(m_NetModel.GetWorkItem(m_WISelForm.m_SelectedWorkItemID));
						AddToMRU(m_WISelForm.m_SelectedWorkItemID);
					}
				}
				return;
			}

			else if (e.ClickedItem.Name=="MakeRelease")
			{
				MakeReleaseForm form = new MakeReleaseForm(this, m_TFS);
				form.ShowDialog(this);

				return;
			}

			else if (e.ClickedItem.Name=="BranchesComparison")
			{
				BranchesComparisonSettingsForm form = new BranchesComparisonSettingsForm(this, m_TFS);
				if (form.ShowDialog(this) != DialogResult.OK)
				{
					return;
				}

				BranchesComparisonForm bcform = new BranchesComparisonForm(this, m_TFS, form.BranchesList.Items, form.FromDate.Value, form.ToDate.Value);
				bcform.ShowDialog();

				return;
			}

			else if (e.ClickedItem.Name=="CheckDB")
			{
				CheckDB form = new CheckDB(this, m_TFS, m_WIP);
				form.ShowDialog(this);
			}

			// Create a new Root Work Item from a specific type
			else
			{
				// First close the existing DB
				if (!CheckWIToSave(false, false))
				{
					return;
				}

				// We clear the database
				ClearWIDatabase();

				// Create a new Work Item
				WorkItemType wit = m_WIP.WorkItemTypes[e.ClickedItem.Name];
				if (wit == null)
				{
					UpdateTopLevelAction(String.Format("Unknown Work Item Type: {0}", e.ClickedItem.Name));
					return;
				}

				WorkItem wi = m_NetModel.m_WIPool.Create(wit);
				wi.Title = String.Format("{0} #{1}", wit.Name, wicounter++);

				// Build the database from it
				BuildWIDatabase(wi);
			}
		}
		#endregion

		#region Saving procedure
		/// <summary>
		/// Check if the database should be saved, ask the user if so, and save if he wants to.
		/// </summary>
		/// <param name="must">If <c>true</c> the save is an obligation, if the user doesn't want to save, the 
		/// method return false, the operation that calls this method should consider to fail.</param>
		/// <returns>Success flag</returns>
		/// <param name="nocancel">If <c>true</c> there won't be any Cancel button in the MessageBox.</param>
		/// <returns>Success flag</returns>
		public bool CheckWIToSave(bool must, bool nocancel)
		{
			// Nothing to save
			if ((m_Root==null) || (m_IsDirty==false))
			{
				return true;
			}

			// Message box data setup
			MessageBoxButtons	btn = MessageBoxButtons.OKCancel;
			String				msg	= "You must save the database to continue!";
			if (must == false)
			{
				btn = nocancel ? MessageBoxButtons.YesNo : MessageBoxButtons.YesNoCancel;
				msg = "Do you want to save the Work Item database ?";
			}
			
			// Show the Message box, and react from its result
			switch (MessageBox.Show(msg, "Cegedim Work Item Creator", btn))
			{
				// Ok or Yes, we save!
				case DialogResult.OK:
				case DialogResult.Yes:
					{
						if (!CmdSaveDatabase())
						{
							return false;
						}
						return true;
					}

				// No, we remove the dirty flag, and the changes are considered as lost
				case DialogResult.No:
					{
						return true;
					}

				// Cancelled, return false because the user doesn't want to do what we want to do !
				case DialogResult.Cancel:
					{
						return false;
					}
			}
			return false;
		}

		/// <summary>
		/// Main saving procedure
		/// </summary>
		/// <param name="wic">The WIC to save</param>
		/// <param name="recurse">If <c>true</c> the children of <paramref name="wic"/> will also be saved</param>
		/// <param name="step">Step of the saving process to perform</param>
		/// <returns>Succes flag</returns>
		/// <remarks>The saving process occurs in many steps:
		/// <list type="bullet">
		/// <item>step 0: The main step, will call _SaveWorkItems with the following step and perform additionnal operations</item>
		/// <item>step 1: Save the newly created Work Item</item>
		/// <item>step 2: Update the hierarchy</item>
		/// <item>step 3: Save the Work Item that were modified</item>
		/// </list>
		/// After the completion of these steps, the method will save the content of the Recycle Bin.
		/// </remarks>
		public bool _SaveWorkItems(WICObject wic, bool recurse, int step)
		{
			// Optimistic...
			bool success = true;

			try
			{
				// First step
				if (step == 0)
				{
					using (new Action(this, "Saving the database...", "Done"))
					{
						// We save the newly created WIs to get their ID (which is essential for the hierarchy creation)
						success &= _SaveWorkItems(wic, recurse, 1);

						// From now, disable the notification of WI modification
						using (new DisableWIField(this))
						{
							// Update Hierarchy
							using (new Action(this, "Update hierarchy of the database...", "Done"))
							{
								success &= _SaveWorkItems(wic, recurse, 2);
							}

							// We save the Work Items of the database opened
							success &= _SaveWorkItems(wic, recurse, 3);

							// Finally we save the Work Item to recycle
							foreach (WorkItem cur in m_NetModel.m_ToRecycle)
							{
								if (cur.IsDirty && cur.IsValid())
								{
									cur.Save();
									m_DirtyCounter--;
								}
								else
								{
									//TODO PB !
								}
							}
							m_NetModel.m_ToRecycle.Clear();

							// Save the Recycle Bin Work Items
							if (m_RecycleBin != null)
							{
								if (m_RecycleBin.IsDirty && m_RecycleBin.IsValid())
								{
									// Save the Recycle Bin Work Item
									m_RecycleBin.Save();
								}

								// Save its direct children
								WorkItem cur = m_NetModel.GetFirstChildWorkItem(m_RecycleBin);
								while (cur != null)
								{
									if (cur.IsDirty && cur.IsValid())
									{
										cur.Save();
									}
									cur = m_NetModel.GetNextWorkItem(cur);
								}
							}

							m_DirtyCounter = 0;

						}
					}

					return success;
				}

				// Second pass: save the new Work Items
				else if (step == 1)
				{
					if ((wic.Type == WICObjType.DBObj) && (wic.WI.Id == 0))
					{
						WorkItem wi = wic.WI;
						if (wi.IsValid())
						{
							// Saving
							using (new Action(this, String.Format("Saving the newly created Work Item {0}...", wi.Title), "Saved"))
							{
								wi.Save();
							}
						}
						else
						{
							throw new Exception("Invalid Work Item");
						}
					}
				}

				// Third pass: update the hierarchy
				// We construct the Work Item Hierarchy from the Tree Control's one, either or not it has changed
				else if (step == 2)
				{
					if (wic.Type == WICObjType.DBObj)
					{
						WorkItem wi = wic.WI;

						// Update the ParentWI field
						if (wic.Parent != null)
						{
							WorkItem parentwi = wic.Parent.WI;
							m_NetModel.SetFieldParentWI(wi, parentwi.Id);
						}

						// Update the FirstChildWI (only if the children are loaded)
						if (wic.ChildrenLoaded)
						{
							if (wic.Children.Count > 0)
							{
								WorkItem firstchildwi = wic.Children[0].WI;
								m_NetModel.SetFieldFirstChildWI(wi, firstchildwi.Id);
							}
							else
							{
								m_NetModel.SetFieldFirstChildWI(wi, 0);
							}
						}

						// Update the NextWI
						if (wic.Next != null)
						{
							WorkItem nextwi = wic.Next.WI;
							m_NetModel.SetFieldNextWI(wi, nextwi.Id);
						}
						else
						{
							m_NetModel.SetFieldNextWI(wi, 0);
						}
					}
				}

				// Fourth pass: save everything
				else if (step == 3)
				{
					if (wic.Type == WICObjType.DBObj)
					{
						WorkItem wi = wic.WI;
						TreeNode tn = wic.TN;

						if (wi.IsDirty)
						{
							if (wi.IsValid())
							{
								// Saving
								using (new Action(this, String.Format("Saving the Work Item {0}...", wi.Title), "Saved"))
								{
									// Save the Work Item
									wi.Save();

									// Update the Dirty Counter
									UpdateDirty(wic, UpdateMode.UM_Modify, false);
								}
							}
							else
							{
								throw new Exception("CWC0001 Invalid Work Item");
							}
						}

						// We made it so far, so make sure the label color is black
						if (tn != null)
						{
							tn.ForeColor = Color.Black;

							// At last update the text of the item
							tn.Text = GetTreeItemName(wic);
						}
					}
				}

			}
			catch (Exception ex)
			{
				// If there were a problem during the save, we put the TreeNode in Red, make it visible,
				// and warn in the status bar
				TreeNode tn = wic.TN;
				if (tn != null)
				{
					tn.EnsureVisible();
					tn.ForeColor = Color.Red;
				}
				UpdateTopLevelAction(String.Format("Can't save the Work Item {0} because {1}.",
					wic.WI.Title, ex.ToString().Contains("CWC0001") ? "it is invalid" : "there's a more recent version on the server"));
				success = false;
				wic.DirtyState = false;		// Reset the dirty state has the WI will need modification anyway to be saved
			}

			// Recurse to children if asked
			if (recurse)
			{
				// Children loop
				foreach (WICObject cur in wic.Children)
				{
					success &= _SaveWorkItems(cur, true, step);
				}
			}

			return success;
		}
		#endregion

		#region WICObject/TreeNode/WorkItem interactions

		/// <summary>
		/// Find all the WICObject instances corresponding to the given Work Item
		/// </summary>
		/// <param name="wi">Work Item to get the WICObject instances</param>
		/// <returns>An enumerable object containing the WICObject that handle <paramref name="wi"/></returns>
		/// <remarks>In future version of the WICreator, dependencies will be managed, in that case there could be
		/// more than one WICObject for a given Work Item</remarks>
		IEnumerable<WICObject> WICsFromWI(WorkItem wi)
		{
			List<WICObject>	list = new List<WICObject>();
			if (m_Root == null)
			{
				return list;
			}

			_FromWI(m_Root, wi, list, true);

			return list;
		}

		/// <summary>
		/// Get the WICObject corresponding to a WorkItem.
		/// </summary>
		/// <param name="wi">The Work Item to get the corresponding WICObject</param>
		/// <returns>The WICObject</returns>
		/// <remarks>This method doesn't consider the WICObject representing the dependencies</remarks>
		WICObject FromWI(WorkItem wi)
		{
			List<WICObject>	list = new List<WICObject>();
			if (m_Root == null)
			{
				return null;
			}

			_FromWI(m_Root, wi, list, false);

			return (list.Count>0) ? list[0] : null;
		}

		/// <summary>
		/// Recursive part of WICsFromWI and FromWI
		/// </summary>
		/// <param name="wic">WIC to analyse</param>
		/// <param name="wi">Work Item of the WIC we're looking for</param>
		/// <param name="list">The list to store the WIC if it matches <paramref name="wi"/></param>
		/// <param name="includedep">If <c>true</c> the dependencies will be analysed</param>
		void _FromWI(WICObject wic, WorkItem wi, List<WICObject> list, bool includedep)
		{
			WorkItem wicwi = wic.WI;
			if (wic.Type == WICObjType.DBObj)
			{
				// ID is null, the Work Item is just created, not saved yet, so we're testing with the objects instead
				// of the IDs
				if (wicwi.Id==0)
				{
					if (wi == wicwi)
					{
						list.Add(wic);
					}
				} else {
					if (wi.Id == wicwi.Id)
					{
						list.Add(wic);
					}
				}
			}

			foreach (WICObject cur in wic.Children)
			{
				_FromWI(cur, wi, list, includedep);
			}
		}

		/// <summary>
		/// Remove the given WIC from the database
		/// </summary>
		/// <param name="wic">The WIC to remove</param>
		/// <returns>Succcess flag</returns>
		/// <remarks>If the WIC or one of his children is dirty, a Message will be displayed to warn about the changes
		/// lost.</remarks>
		private bool RemoveWIC(WICObject wic)
		{
			// Check if at least sonwic or its children are dirty
			if (wic.IsDirtyRecursive(true))
			{
				// Warn the user we'll discard all the content modified before unlinking if he unlinks
				if (MessageBox.Show("This operation will discard the changes made to the object you are unlinking and also its children!",
					"Cegedim Work Item Creator", MessageBoxButtons.OKCancel, MessageBoxIcon.Warning) == DialogResult.Cancel)
				{
					return false;
				}
			}

			// Remove the WIC, then the related TreeNode
			UpdateDirty(wic, UpdateMode.UM_Remove, true);
			using (new DisableWIField(this))
			{
				//Special case, the WI is not new and not dirty, we increment the dirty counter to enable the save
				if ((wic.WI.Id != 0) && (wic.WI.IsDirty == false))
				{
					m_DirtyCounter++;
				}

				// Remove the WIC
				wic.Remove();
			}

			return true;
		}

		#endregion

		#region Misc GUI
		
		bool CheckWIP(String prjname)
		{
			Project prj = m_WIS.Projects[prjname];
			if (prj == null)
			{
				return true;
			}

			foreach (WorkItemType wit in prj.WorkItemTypes)
			{
				try
				{
					FieldDefinition f = wit.FieldDefinitions["Cegedim.ParentWI"];
					f = wit.FieldDefinitions["Cegedim.FirstChildWI"];
					f = wit.FieldDefinitions["Cegedim.NextWI"];
				}
				catch (Exception)
				{
					return UpgradeWITs(prjname);
				}
			}
			return true;
		}

		bool UpgradeWITs(String prjname)
		{
			if (MessageBox.Show(string.Format(
					"The Work Item Types of the '{0}' project don't have the fields required to handle hierarchy.\r\n"+
					"Do you want to upgrade the WIT definition to handle hierarchy?\r\n"+
					"If refuse, WICreator won't handle this project.", prjname), "Work Item Creator", 
				MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.No)
			{
				return false;
			}

			List<String> witnames = new List<string>();
			{
				Project prj = m_WIS.Projects[prjname];
				prj.WorkItemTypes.ImportEventHandler += new ImportEventHandler(WorkItemType_ValidationEventHandler);
				foreach (WorkItemType wit in prj.WorkItemTypes)
				{
					witnames.Add(wit.Name);
				}
			}

			foreach(String witname in witnames)
			{
				m_WIS.RefreshCache();
				m_WIS.SyncToCache();

				Project prj = m_WIS.Projects[prjname];
				WorkItemType wit = prj.WorkItemTypes[witname];

				try
				{
					FieldDefinition f = wit.FieldDefinitions["Cegedim.ParentWI"];
					f = wit.FieldDefinitions["Cegedim.FirstChildWI"];
					f = wit.FieldDefinitions["Cegedim.NextWI"];
				}
				catch (Exception)
				{
					UpgradeWIT(prj, wit);
				}
			}

			m_WIS.RefreshCache();
			m_WIS.SyncToCache();

			return true;
		}

		bool WITValidateSucceed = true;
		bool UpgradeWIT(Project prj, WorkItemType wit)
		{
			using (new Action(this, String.Format("Modifying definition of '{0}'...", wit.Name), String.Format("Work Item Type '{0}' now supports hierarchy.", wit.Name)))
			{
				// Get the Xml Document for the given WIT
				XmlDocument doc = wit.Export(false);
				XmlElement fields = doc.SelectSingleNode("*/WORKITEMTYPE/FIELDS") as XmlElement;

				// Look for each field, and add the ones that are not present
				XmlElement field = fields.SelectSingleNode("FIELD[@refname=\"Cegedim.ParentWI\"]") as XmlElement;
				if (field == null)
				{
					field = doc.CreateElement("FIELD");
					field.SetAttribute("name", "ParentWI");
					field.SetAttribute("refname", "Cegedim.ParentWI");
					field.SetAttribute("type", "Integer");
					field.InnerXml = "<HELPTEXT>The ID of the Parent Work Item</HELPTEXT><DEFAULT from=\"value\" value=\"0\" />";
					fields.AppendChild(field);
				}

				field = fields.SelectSingleNode("FIELD[@refname=\"Cegedim.FirstChildWI\"]") as XmlElement;
				if (field == null)
				{
					field = doc.CreateElement("FIELD");
					field.SetAttribute("name", "FirstChildWI");
					field.SetAttribute("refname", "Cegedim.FirstChildWI");
					field.SetAttribute("type", "Integer");
					field.InnerXml = "<HELPTEXT>The ID of the first children</HELPTEXT><DEFAULT from=\"value\" value=\"0\" />";
					fields.AppendChild(field);
				}

				field = fields.SelectSingleNode("FIELD[@refname=\"Cegedim.NextWI\"]") as XmlElement;
				if (field == null)
				{
					field = doc.CreateElement("FIELD");
					field.SetAttribute("name", "NextWI");
					field.SetAttribute("refname", "Cegedim.NextWI");
					field.SetAttribute("type", "Integer");
					field.InnerXml = "<HELPTEXT>The ID of the next Work Item</HELPTEXT><DEFAULT from=\"value\" value=\"0\" />";
					fields.AppendChild(field);
				}

				// Validate the type
				try
				{
					WITValidateSucceed = true;
					WorkItemType.Validate(prj, doc.OuterXml);
					if (WITValidateSucceed == false)
					{
						MessageBox.Show("Work Item Type couldn't pass validation, something went wrong", "Work Item Creator", MessageBoxButtons.OK, MessageBoxIcon.Error);
						return false;
					}

					prj.WorkItemTypes.Import(doc.OuterXml);

				}
				catch (Exception ex)
				{
					MessageBox.Show(string.Format("Work Item Type couldn't be imported.\r\n{0}", ex.ToString()), "Work Item Creator", MessageBoxButtons.OK, MessageBoxIcon.Error);
					return false;
				}
			}
			return true;
		}

		void WorkItemType_ValidationEventHandler(object sender, ImportEventArgs e)
		{
			WITValidateSucceed = e.Severity == ImportSeverity.Error;
		}

		/// <summary>
		/// Update the GUI corresponding to the given WIC
		/// </summary>
		/// <param name="wic">The WIC to update the GUI</param>
		/// <remarks>The TreeNode associated will be updated, and if the Work Item corresponding to the WIC is selected
		/// the info bar will also be updated.</remarks>
		void UpdateWIGUI(WICObject wic)
		{
			try
			{
				TreeNode	tn	= wic.TN;
				WorkItem	wi  = wic.WI;
				if (wi.IsOpen==true)
				{
					wi.Validate();
				}

				if (tn != null)
				{
					if ((wi.IsOpen == false) || (wi.IsValid() == true))
					{
						tn.ForeColor = Color.Black;
					}
					else
					{
						tn.ForeColor = Color.Red;
					}
					tn.Text = GetTreeItemName(wic);
				}

				if (wic == m_SelWIC)
				{
					UpdateWIInfoBar(wic);
				}
			}
			catch(Exception ex)
			{
				StatusException(ex);
			}
		}

		/// <summary>
		/// Update the Infobar with the given WIC
		/// </summary>
		/// <param name="wic">WIC to use to display the information</param>
		void UpdateWIInfoBar (WICObject wic)
		{
			String wiinfo = "";
			
			if (wic != null)
			{
				WorkItem wi = wic.WI;
				if ((wi.IsOpen == false) || (wi.IsValid() == true))
				{
					wiinfo = String.Format("{0} {1} {2}: {3}", wi.Type.Name, wi.Id, wi.IsDirty ? "(modified) " : "", wi.Title);
				}
				else
				{
					ArrayList list = wi.Validate();
					Field f = list[0] as Field;

					String status;
					switch (f.Status)
					{
						case FieldStatus.InvalidEmpty: status = "can't be empty"; break;
						default: status = f.Status.ToString(); break;
					}
					wiinfo = String.Format("{0} {1} {2}: The field \"{3}\" {4}", wi.Type.Name, wi.Id, wi.IsDirty ? "(modified) " : "", f.Name, status);
				}
			}

			m_WIInfoBar.Text = wiinfo;
			m_WIInfoBar.Update();
		}

		/// <summary>
		/// Update the enable/disable state of the Hierarchy move buttons
		/// </summary>
		/// <remarks>If filtering is on, the button are disabled because if wouldn't make any sense to modify the order
		/// with a filtered content.</remarks>
		private void UpdateMoveBtn()
		{
			bool nofilter = IsFiltering() == false;
			m_WIPrevBtn.Enabled		= nofilter;
			m_WIToTopBtn.Enabled	= nofilter;
			m_WIDownBtn.Enabled		= nofilter;
			m_WIToBottom.Enabled	= nofilter;
		}

		void LoadMRU()
		{
			// Load the MRU
			if (Settings.Default.MRU != null)
			{
				SortedDictionary<int, int> smru = new SortedDictionary<int, int>();

				foreach (String curstr in Settings.Default.MRU)
				{
					try
					{
						string[] kvp = curstr.Split(new char[] { ';' });
						if (kvp.Length == 2)
						{
							int key = int.Parse(kvp[1]) * 100;
							int val = int.Parse(kvp[0]);

							while (smru.ContainsKey(key))
							{
								key++;
							}

							smru.Add(key, val);
						}
					}
					catch (Exception)
					{
					}
				}

				if (smru.Count > 0)
				{
					m_FavoritesMenu.DropDownItems.Clear();
					WorkItemStore wis = new WorkItemStore(m_TFS);
					foreach (KeyValuePair<int, int> kvp in smru)
					{
						try
						{
							WorkItem wi = wis.GetWorkItem(kvp.Value);
							ToolStripMenuItem tsi = new ToolStripMenuItem(string.Format("[{0}] {1}", wi.Id, wi.Title));
							tsi.Tag = wi.Id;
							m_FavoritesMenu.DropDownItems.Insert(0, tsi);

						}
						catch (Exception)
						{
							// The Work item is not present on that server
						}
					}
				}
			}
		}
		void AddToMRU(int wiid)
		{
			StringCollection mru = Settings.Default.MRU;
			if (mru == null)
			{
				Settings.Default.MRU = new StringCollection();
				mru = Settings.Default.MRU;
			}

			for (int i = 0; i < mru.Count; i++)
			{
				String curval = mru[i];
				if (curval.Contains(wiid.ToString() + ";"))
				{
					string[] kvp = curval.Split(new char[] { ';' });
					int counter = int.Parse(kvp[1])+1;
					mru[i] = wiid.ToString() + ";" + counter.ToString();
					Settings.Default.Save();
					LoadMRU();
					return;
				}
			}

			if (mru.Count == 10)
			{
				int mini=0, minval = int.MaxValue;

				for (int i=0; i<mru.Count; i++)
				{
					string[] kvp = mru[i].Split(new char[] { ';' });
					int counter = int.Parse(kvp[1]);

					if (counter < minval)
					{
						mini = i;
						minval = counter;
					}
				}

				mru[mini] = wiid.ToString() + ";1";
				Settings.Default.Save();
				LoadMRU();
				return;

			}

			mru.Add(wiid.ToString() + ";1");
			Settings.Default.Save();
			LoadMRU();
		}

		#endregion

		#region Main commands
		/// <summary>
		/// Set the current node as the root one
		/// </summary>
		/// <param name="parent">TreeNode to set as the root</param>
		/// <returns>Success flag</returns>
		/// <remarks>This method will set the ParentWI field of the Work Item to -1</remarks>
		private bool CmdSetAsRoot(TreeNode parent)
		{
			WICObject	wic = parent.Tag as WICObject;
			WorkItem	wi	= wic.WI;
			if (wi==null)
			{
				return false;
			}

			using (new Action(this, "", String.Format("Set Work Item [{0}] as root.", wi.Id.ToString())))
			{
				m_NetModel.SetFieldParentWI(wi, -1);
			}
			
			return true;
		}

		/// <summary>
		/// Unlink the Work Item from its parent
		/// </summary>
		/// <param name="parent">TreeNode corresponding to the Work Item to unlink</param>
		/// <returns>Success flag</returns>
		private bool CmdUnlinkWI(TreeNode parent)
		{
			CmdUnlinkWorkItem(parent);
			return true;
		}

		/// <summary>
		/// Unlink the Work Item from its parent
		/// </summary>
		/// <param name="tn">TreeNode corresponding to the Work Item to unlink</param>
		/// <returns>Success flag</returns>
		private bool CmdUnlinkWorkItem(TreeNode tn)
		{
			try
			{
				TreeNode  parenttn = tn.Parent;
				if (parenttn==null)
				{
					return false;
				}
		
				WICObject parentwic = parenttn.Tag as WICObject;
				WICObject sonwic	= tn.Tag as WICObject;
				if (sonwic.Type!=WICObjType.DBObj)
				{
					return false;
				}

				return RemoveWIC(sonwic);

			}
			catch (Exception ex)
			{
				StatusException(ex);
			}
			return false;
		}

		/// <summary>
		/// Link an existing Work Item to a given parent
		/// </summary>
		/// <param name="parenttn">The TreeNode of the parent Work Item</param>
		/// <returns>Success flag</returns>
		/// <remarks>This method will display a Work Item selection form, for the user to choose the Work Item to link.
		/// </remarks>
		private bool CmdLinkExistingWI(TreeNode parenttn)
		{
			WICObject parentwic = parenttn.Tag as WICObject;
			WICObject sonwic = null;
			WorkItem parentwi = parentwic.WI;

			// Display the dialog of Work Item selection
			if (m_WISelForm.ShowDialog() == DialogResult.OK)
			{
				// Start the operation
				WorkItem selwi = m_NetModel.GetWorkItem(m_WISelForm.m_SelectedWorkItemID);
				using (new Action(this, String.Format("Linking {0} to {1}...", selwi.Title, parentwi.Title), "Linking done"))
				{
					try
					{
						sonwic = new WICObject(m_NetModel, WICObjType.DBObj, selwi);
						m_NetModel.CanLinkWorkItems(m_WIP.Name, parentwic, sonwic);
					}
					catch (Exception ex)
					{
						// Already Linked
						if (ex.Message.Contains("CWC0001"))
						{
							DialogResult dr = MessageBox.Show(String.Format("The Work Item {0} has already a parent.\rDo you want to move it?", selwi.Title),
								"Cegedim Work Item Creator", MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation);
							if (dr == DialogResult.No)
							{
								return false;
							}

							// The Work Item is already linked, look for its WIC and remove it if found
							WICObject exwic = FromWI(selwi);
							if (exwic != null)
							{
								exwic.Remove();
							}

							// There's no WIC, the Work Item is in another project
							else
							{
								// Let's unlink it using the NetworkModel, add changed Work Item to the ToRecycle list
								// for them to be saved during the next saving process
								m_NetModel.UnlinkWorkItem(selwi, WINetworkModel.WMAction.WMA_TORECYCLE);
							}
						}

						// Cyclic
						else if (ex.Message.Contains("CWC0002"))
						{
							MessageBox.Show(ex.Message, "Cegedim Work Item Creator", MessageBoxButtons.OK, MessageBoxIcon.Error);
							return false;
						}
						// Rules failed
						else if (ex.Message.Contains("CWC0003"))
						{
							MessageBox.Show(ex.Message, "Cegedim Work Item Creator", MessageBoxButtons.OK, MessageBoxIcon.Error);
							return false;
						}
						else
						{
							StatusException(ex);
						}
					}

					parentwic.AddChild(sonwic);
					UpdateDirty(sonwic, UpdateMode.UM_Add, true);
					UpdateWITree(parentwic, true, true);
					m_WITree.SelectedNode = sonwic.TN;
				}
			}
			return true;
		}

		/// <summary>
		/// Send the given Work Item to the recycle bin.
		/// </summary>
		/// <param name="parent">The TreeNode corresponding to the Work Item to send</param>
		/// <returns>Success flags</returns>
		/// <remarks>This method will close (if possible) the Work Item and set the ArearPath to the one corresponding
		/// to the Recycle Bin (if any).</remarks>
		private bool CmdSendToRecycle(TreeNode parent)
		{
			WICObject wic = parent.Tag as WICObject;
			WorkItem wi = wic.WI;
			if ((wi == null) || (wi.Id == m_Root.WI.Id))
			{
				return false;
			}

			if (MessageBox.Show(this, String.Format("Are you sure you want to send {0} and all its children to the Recycle Bin?",
				wi.Title), "Cegedim Work Item Creator", MessageBoxButtons.YesNo, MessageBoxIcon.Information) == DialogResult.No)
			{
				return false;
			}

			// No Work Item Notification
			using (new DisableWIField(this))
			{
				WICObject upwic = wic.Previous;
				if (upwic == null)
				{
					upwic = wic.Parent;
				}
				UpdateDirty(upwic, UpdateMode.UM_Add, false);

				// First revert all changes made to the Work Items, and change their State to Closed and Area path to the Recycle Bin's one
				using (new Action(this, "Closing Work Items...", "Done..."))
				{
					_SendToRecycle(wic);
				}

				using (new Action(this, "Sending to recycle bin...", String.Format("Work Item [{0}] {1} was sent to recycle bin", wi.Id, wi.Title)))
				{
					// Remove the WIC
					RemoveWIC(wic);

					// Relink in the Recycle Bin
					m_NetModel.AddChildWorkItem(m_RecycleBin, wi, false);
				}
			}
			return true;
		}

		private bool _SendToRecycle(WICObject wic)
		{
			bool res = true;

			WorkItem wi = wic.WI;
			if (wi == null)
			{
				return false;
			}

			// Revert changes, close, change Area and save

			// Revert changes, if any
			if (wi.IsDirty)
			{
				wi.Reset();
			}

			// Open for edit if it's not already
			if (wi.IsOpen == false)
			{
				wi.Open();
			}

			// Change the Area path
			if (m_RecycleBinArea != null)
			{
				wi.AreaPath = m_RecycleBinArea;
			}

			// Close the Work Item
			wi.State = "Closed";

			// Save the Work Item
			try
			{
				wi.Save();

			}

			// If an exception is raised, this is probably due to the fact we can't change the WI's state from its
			// current one to close
			catch (Exception)
			{
				// Revert again
				wi.Reset();

				// Change the Area path
				if (m_RecycleBinArea != null)
				{
					wi.AreaPath = m_RecycleBinArea;
				}

				// Last change, save if valid only
				if (wi.IsValid())
				{
					wi.Save();
				}

				else
				{
					// TODO WARNING
				}
			}

			// Children not loaded ? Load them
			if (wic.ChildrenLoaded == false)
			{
				wic.LoadChildren();
			}

			// Recurse
			foreach (WICObject cur in wic.Children)
			{
				res &= _SendToRecycle(cur);
			}

			return res;
		}

		/// <summary>
		/// Delete a newly created (unsaved) Work Item
		/// </summary>
		/// <param name="parent">The TreeNode corresponding to the Work Item to delete</param>
		/// <returns>Success flag</returns>
		/// <remarks>This method will work only for unsaved Work Item (as TFS doesn't allow an existing Work Item to
		/// be deleted)
		/// </remarks>
		private bool CmdDeleteWI(TreeNode parent)
		{
			WICObject wic = parent.Tag as WICObject;
			if (wic == null)
			{
				return false;
			}

			// Destruction of a newly created Work Item
			if (wic.Type == WICObjType.DBObj)
			{
				if (wic.WI.Id != 0)
				{
					return false;
				}

				return RemoveWIC(wic);
			}

			return false;
		}

		/// <summary>
		/// Cancel the modifications made to the given Work Item
		/// </summary>
		/// <param name="node">TreeNode corresponding to the Work Item to cancel the modification</param>
		/// <returns>Success Flag</returns>
		/// <remarks>The revert is calls to Reset() and SyncToLatest()</remarks>
		private bool CmdCancelModifications(TreeNode node)
		{
			WICObject wic = node.Tag as WICObject;
			WorkItem wi = wic.WI;
			if (wi == null)
			{
				return false;
			}

			String msg = String.Format("Do you want to cancel the modifications made on the Work Item [{0}] {1} ?", wi.Id, wi.Title);
			if (MessageBox.Show(msg, "Work Item Creator", MessageBoxButtons.YesNo, MessageBoxIcon.Question) != DialogResult.Yes)
			{
				return true;
			}

			// We reset the Work Item and get the latest version from the server
			wi.Reset();
			wi.SyncToLatest();

			// Mise  jour de la GUI
			UpdateWIGUI(wic);

			return true;
		}


		private bool CmdUpdateAreaPath(TreeNode node)
		{
			WICObject wic = node.Tag as WICObject;
			using (new Action(this, String.Format("Updating AreaPath of the children of {0}...", wic.WI.Title), "Done"))
			{
				UpdateAreaPath(wic, wic.WI.AreaPath);
			}

			return true;
		}

		private void UpdateAreaPath(WICObject wic, string newarea)
		{
			if (wic.WI.State != "Closed")
			{
				wic.WI.AreaPath = newarea;
			}
			
			if (wic.ChildrenLoaded == false)
			{
				wic.LoadChildren();
			}

			foreach(WICObject cur in wic.Children)
			{
				UpdateAreaPath(cur, newarea);
			}
		}

		private bool CmdUpdateIteration(TreeNode node)
		{
			WICObject wic = node.Tag as WICObject;
			using (new Action(this, String.Format("Updating Iteration of the children of {0}...", wic.WI.Title), "Done"))
			{
				UpdateIteration(wic, wic.WI.IterationPath);
			}

			return true;
		}

		private void UpdateIteration(WICObject wic, string newiteration)
		{
			if (wic.WI.State != "Closed")
			{
				wic.WI.IterationPath = newiteration;
			}
			
			if (wic.ChildrenLoaded == false)
			{
				wic.LoadChildren();
			}

			foreach(WICObject cur in wic.Children)
			{
				UpdateIteration(cur, newiteration);
			}
		}

		/// <summary>
		/// Save the database if dirty
		/// </summary>
		/// <returns>Success flag</returns>
		public bool CmdSaveDatabase()
		{
			if ((m_IsDirty == false) || (m_Root == null))
			{
				return false;
			}

			_SaveWorkItems(m_Root, true, 0);

			return true;
		}
		
		#endregion

		#region Toolbar triggers
		private void TBI_Save(object sender, EventArgs e)
		{
			CmdSaveDatabase();
		}

		private void TBI_QuickWIDTextChanged(object sender, EventArgs e)
		{
			try
			{
				int id = int.Parse(m_QuickOpenWID.Text);
				WorkItem wi = m_NetModel.GetWorkItem(id);

				if ((wi != null) && (wi.Project == m_WIP))
				{
					m_GoQuickOpen.Enabled = true;
				}
			}
			catch (Exception)
			{
				m_GoQuickOpen.Enabled = false;
			}
		}

		private void TBI_SelectGoQuickOpen(object sender, EventArgs e)
		{
			if (!CheckWIToSave(false, false))
			{
				return;
			}

			// We clear the database
			ClearWIDatabase();

			using (new Action(this, "Opening Work Item Database...", "Database opened"))
			{
				int wid = int.Parse(m_QuickOpenWID.Text);

				// Make sure the Opened Database is to the new hierarchy format
				if (!DoHierarchyMigration(wid))
				{
					UpdateTopLevelAction("Migration cancel, databased opening aborted!");
					return;
				}

				BuildWIDatabase(m_NetModel.GetWorkItem(wid));
				AddToMRU(wid);
			}
		}

		private void TBI_QuickWIDKeyPress(object sender, KeyPressEventArgs e)
		{
			// Return ? Go !
			if ((m_GoQuickOpen.Enabled==true) && (e.KeyChar=='\r'))
			{
				TBI_SelectGoQuickOpen(sender, e);
			}
		}

		private void TBI_RefreshDB(object sender, EventArgs e)
		{
			// On demande a sauv si besoin, car le refresh risquerait d'craser les changements en cours
			if (CheckWIToSave(true, false)==false)
			{
				StatusText.Text = "Refresh aborted";
				return;
			}

			using (new Action(this, "Refreshing Work Item database...", "Work Item database refreshed!"))
			{
				foreach (TreeNode tn in m_WITree.Nodes)
				{
					RefreshWIs(tn);
				}
			}
		}

		private void TBI_WIToFirst(object sender, EventArgs e)
		{
			TreeNode seltn = m_WITree.SelectedNode;
			if (seltn == null)
			{
				return;
			}

			WICObject wic = seltn.Tag as WICObject;
			if (!wic.MoveToFirst())
			{
				return;
			}

			// Remove the WIC's TN from its parent to force a new insertion at the right place during the coming update
			wic.TN.Remove();

			UpdateDirty(wic, UpdateMode.UM_Add, false);
			UpdateWITree(wic, false, false);

			// Select again the Tree Node because the selection was lost during the TN.Remove()
			m_WITree.SelectedNode = wic.TN;
		}

		private void TBI_WIPrev(object sender, EventArgs e)
		{
			TreeNode seltn = m_WITree.SelectedNode;
			if (seltn == null)
			{
				return;
			}

			WICObject wic = seltn.Tag as WICObject;
			if (!wic.MoveToPrevious())
			{
				return;
			}

			// Remove the WIC's TN from its parent to force a new insertion at the right place during the coming update
			wic.TN.Remove();

			UpdateDirty(wic, UpdateMode.UM_Add, false);
			UpdateWITree(wic, false, false);

			// Select again the Tree Node because the selection was lost during the TN.Remove()
			m_WITree.SelectedNode = wic.TN;
		}

		private void TBI_WINext(object sender, EventArgs e)
		{
			TreeNode seltn = m_WITree.SelectedNode;
			if (seltn == null)
			{
				return;
			}

			WICObject wic = seltn.Tag as WICObject;
			if (!wic.MoveToNext())
			{
				return;
			}

			// Remove the WIC's TN from its parent to force a new insertion at the right place during the coming update
			wic.TN.Remove();

			UpdateDirty(wic, UpdateMode.UM_Add, false);
			UpdateWITree(wic, false, false);

			// Select again the Tree Node because the selection was lost during the TN.Remove()
			m_WITree.SelectedNode = wic.TN;
		}

		private void TBI_WIToLast(object sender, EventArgs e)
		{
			TreeNode seltn = m_WITree.SelectedNode;
			if (seltn == null)
			{
				return;
			}

			WICObject wic = seltn.Tag as WICObject;
			if (!wic.MoveToLast())
			{
				return;
			}

			// Remove the WIC's TN from its parent to force a new insertion at the right place during the coming update
			wic.TN.Remove();

			UpdateDirty(wic, UpdateMode.UM_Add, false);
			UpdateWITree(wic, false, false);

			// Select again the Tree Node because the selection was lost during the TN.Remove()
			m_WITree.SelectedNode = wic.TN;
		}
		#endregion

		#region Tree Filter
		private void TitleFilter_Enter(object sender, EventArgs e)
		{
			if (m_TitleFilter.Text == "<Title Filter>")
			{
				m_TitleFilter.Text = "";
				m_TitleFilter.ForeColor = Color.Black;
			}
		
			UpdateMoveBtn();
		}

		private void TitleFilter_Leave(object sender, EventArgs e)
		{
			if (m_TitleFilter.Text == "")
			{
				m_TitleFilter.Text = "<Title Filter>";
				m_TitleFilter.ForeColor = Color.Gray;
			}

			UpdateMoveBtn();
		}

		private void TitleFilter_TextChanged(object sender, EventArgs e)
		{
			if (m_Root == null)
			{
				return;
			}

			// Parse the whole WIC Database, and filter WIs that don't contain the filter text
			_DoTextFilter((m_TitleFilter.Text=="<Title Filter>") ? "" : m_TitleFilter.Text, m_Root);

			// Update the Tree Control
			UpdateWITree(m_Root, true, true);

			UpdateMoveBtn();
		}

		private void _DoTextFilter(String filter, WICObject wic)
		{
			if (wic == null)
			{
				return;
			}

			if (wic.Type != WICObjType.DBObj)
			{
				return;
			}

			if (filter=="")
			{
				wic.TextFiltered = false;
			}
			else
			{
				wic.TextFiltered = wic.WI.Title.ToLower().Contains(filter.ToLower()) == false;
			}

			// Ensure children are loaded
			if (wic.ChildrenLoaded==false)
			{
				using (new Action(this, String.Format("Loading children of {0}", wic.WI.Title), "Done"))
				{
					wic.LoadChildren();
				}
			}

			// Recurse
			foreach (WICObject cur in wic.Children)
			{
				_DoTextFilter(filter, cur);
			}
		}

		private	TreeFilterForm	m_WIFilter	= new TreeFilterForm();

		private void FilterBtn_Click(object sender, EventArgs e)
		{
			// Load all the database
			LoadAll(m_Root);

			if (m_WIFilter.ShowDialog(this) != DialogResult.OK)
			{
				return;
			}

			// Update the Filter Item
			m_WIFilter.UpdateTree();

			// Update the ordering buttons
			UpdateMoveBtn();

			// Refresh the WITree
			UpdateWITree(m_Root, true, true);
		}

		private void WICHierarchyChanged(object sender, WICObjectHierarchyEventArgs e)
		{
			_WIHIUpdate(e.WIC, e.Insert);
		}

		private void _WIHIUpdate(WICObject wic, bool insert)
		{
			m_WIFilter.Update(wic, insert);

			foreach(WICObject cur in wic.Children)
			{
				_WIHIUpdate(cur, insert);
			}
		}
		#endregion

		private void m_FavoritesMenu_DropDownItemClicked(object sender, ToolStripItemClickedEventArgs e)
		{
			try
			{
				int wiid = (int)e.ClickedItem.Tag;
				if (!CheckWIToSave(false, false))
				{
					return;
				}

				// We clear the database
				ClearWIDatabase();

				using (new Action(this, "Opening Work Item Database...", "Database opened"))
				{
					// Make sure the Opened Database is to the new hierarchy format
					if (!DoHierarchyMigration(wiid))
					{
						UpdateTopLevelAction("Migration cancel, databased opening aborted!");
						return;
					}

					BuildWIDatabase(m_NetModel.GetWorkItem(wiid));
					AddToMRU(wiid);
				}

			}
			catch (Exception)
			{
			}
		}
	}
}