using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.WorkItemTracking.Client;
using Microsoft.TeamFoundation.VersionControl.Client;
using Microsoft.TeamFoundation.WorkItemTracking.Controls;

namespace CegWICreator
{
	/// <summary>
	/// Work Item filter Form
	/// </summary>
	public partial class TreeFilterForm : Form
	{
		/// <summary>
		/// Default Constructor
		/// </summary>
		public TreeFilterForm()
		{
			InitializeComponent();
		}

		private void Tree_AfterCheck(object sender, TreeViewEventArgs e)
		{
			foreach(TreeNode cur in e.Node.Nodes)
			{
				cur.Checked = e.Node.Checked;
			}
		}

		private void Tree_Collapse(object sender, TreeViewEventArgs e)
		{
			FilterItem item = e.Node.Tag as FilterItem;
			if (item != null)
			{
				item.Expanded = false;
			}
		}

		private void Tree_Expand(object sender, TreeViewEventArgs e)
		{
			FilterItem item = e.Node.Tag as FilterItem;
			if (item != null)
			{
				item.Expanded = true;
			}
		}

		private void TreeFilterForm_Shown(object sender, EventArgs e)
		{
			// Delete the whole content of the Tree View
			m_Tree.Nodes.Clear();

			// Add the items
			AddItem(State,		m_Tree.Nodes);
			AddItem(Type,		m_Tree.Nodes);
			AddItem(CreatedBy,	m_Tree.Nodes);
			AddItem(AssignedTo,	m_Tree.Nodes);
		}

		private	void	AddItem	(FilterItem item, TreeNodeCollection owner)
		{
			// Create the TreeNode, set its property. The Tag member point to the FilterItem object
			TreeNode tn = new TreeNode(item.Value.ToString());
			tn.Checked	= item.Filtered == false;
			tn.Tag		= item;

			// Add the TreeNode to the owner collection
			owner.Add(tn);

			// Parse the children and add them if their UsedCount is non null
			foreach (FilterItem cur in item.Children)
			{
				if (cur.UsedCount>0)
				{
					AddItem(cur, tn.Nodes);
				}
			}

			// Expand the TreeNode if required
			if (item.Expanded)
			{
				tn.Expand();
			}
		}

		/// <summary>
		/// Update the filter state with the given WICObject
		/// </summary>
		/// <param name="wic">WICObject used to update the filter values</param>
		/// <param name="insert">If <c>true</c> the WICObject is inserted in the database, if <c>false</c> it is
		/// removed.
		/// </param>
		/// <remarks>This method updates the available values for each filter type.</remarks>
		public void Update(WICObject wic, bool insert)
		{
			// Get the Work Item
			WorkItem wi = wic.WI;
			if (wi == null)
			{
				return;
			}

			// Update the Filter Items
			State.UpdateItem(wi.State,			insert);
			Type.UpdateItem(wi.Type.Name,		insert);
			CreatedBy.UpdateItem(wi.CreatedBy,	insert);

			if (wi.Fields.Contains("Assigned To"))
			{
				Field f = wi.Fields["Assigned To"];
				if (f != null)
				{
					AssignedTo.UpdateItem(f.Value, insert);
				}
			}
		}

		/// <summary>
		/// Update the Form's TreeView with the actual filter content
		/// </summary>
		public void UpdateTree()
		{
			foreach(TreeNode tn in m_Tree.Nodes)
			{
				_Update(tn);
			}
		}

		private void _Update(TreeNode tn)
		{
			FilterItem item = tn.Tag as FilterItem;
			if (item != null)
			{
				item.Filtered = tn.Checked == false;
			}

			foreach(TreeNode cur in tn.Nodes)
			{
				_Update(cur);
			}
		}

		/// <summary>
		/// Determiner if the given WICObject is filtered or not (possibility to recurse)
		/// </summary>
		/// <param name="wic">The WICObject to determine</param>
		/// <param name="recurse">If true all children must also be filtered for the method to return <c>true</c>.</param>
		/// <returns>Returns <c>true</c> if the WICObject is filtered, <c>false</c> if it must be shown (or if at least
		/// one of its children if recurse is on).</returns>
		public bool IsFiltered(WICObject wic, bool recurse)
		{
			// Filtered state
			bool filtered = false;

			// Get the Work Item
			WorkItem wi = wic.WI;
			if (wi != null)
			{
				filtered |= State.IsFiltered(wi.State);
				filtered |= Type.IsFiltered(wi.Type.Name);
				filtered |= CreatedBy.IsFiltered(wi.CreatedBy);
				
				if (wi.Fields.Contains("Assigned To"))
				{
					Field f = wi.Fields["Assigned To"];
					if (f != null)
					{
						filtered |= AssignedTo.IsFiltered(f.Value);
					}
				}
			}

			if (!filtered)
			{
				return false;
			}

			if (recurse)
			{
				foreach(WICObject cur in wic.Children)
				{
					if (IsFiltered(cur, true) == false)
					{
						return false;
					}
				}
			}

			return true;
		}

		/// <summary>
		/// Determine if we're in filtering mode or not.
		/// </summary>
		/// <returns>Returns <c>true</c> if we're filtering</returns>
		/// <remarks>All filter value must be checked for this method to returns <c>true</c>.</remarks>
		public bool IsFiltering()
		{
			if (_IsFiltering(State))
			{
				return true;
			}
			if (_IsFiltering(Type))
			{
				return true;
			}
			if (_IsFiltering(CreatedBy))
			{
				return true;
			}
			if (_IsFiltering(AssignedTo))
			{
				return true;
			}

			return false;
		}

		private	bool	_IsFiltering(FilterItem item)
		{
			if (item.Filtered)
			{
				return true;
			}

			foreach (FilterItem cur in item.Children)
			{
				if (_IsFiltering(cur))
				{
					return true;
				}
			}

			return false;
		}

		public void Clear()
		{
			State		= new FilterItem("State",		null);
			Type		= new FilterItem("Type",		null);
			CreatedBy	= new FilterItem("CreatedBy",	null);
			AssignedTo	= new FilterItem("AssignedTo",	null);
		}

		public void UpdateField(Field f)
		{
			// Filter Tree Update
			if (f.Name == "State")
			{
				State.UpdateItem(f.Value, true);
			}
			else if (f.Name == "Created By")
			{
				CreatedBy.UpdateItem(f.Value, true);
			}
			else if (f.Name == "Assigned To")
			{
				AssignedTo.UpdateItem(f.Value, true);
			}
			else if (f.Name == "Type")
			{
				Type.UpdateItem(f.Value, true);
			}
		}

		private	FilterItem	State		= new FilterItem("State",		null);
		private	FilterItem	Type		= new FilterItem("Type",		null);
		private FilterItem	CreatedBy	= new FilterItem("CreatedBy",	null);
		private FilterItem	AssignedTo	= new FilterItem("AssignedTo",	null);

	}

	class FilterItemList : List<FilterItem>
	{

	};

	class FilterItem
	{
		public		FilterItem(object value, FilterItem owner)
		{
			// Insert the Item in the owner
			if (owner != null)
			{
				owner.Children.Add(this);
			}

			// Set the used count to 1
			UsedCount = 1;

			// Set the value corresponding to the Item
			Value = value;
		}

		public	void UpdateItem(object value, bool inc)
		{
			// Look for the Item in the children
			FilterItem	item = null;
			foreach(FilterItem cur in Children)
			{
				if (cur.Value.ToString() == value.ToString())
				{
					item = cur;
					break;
				}
			}

			// Found ?
			if (item != null)
			{
				if ((inc==false) && (item.UsedCount==0))
				{
					//TODO EXCEPTION
					return;
				}

				// Update the Used Count
				item.UsedCount += inc ? 1 : -1;

				return;
			}

			// Not found, create one if it's an inc
			if (inc == false)
			{
				//TODO EXCEPTION
				return;
			}

			// Create the Item
			item = new FilterItem(value, this);
		}

		public	bool	IsFiltered(object value)
		{
			// Look for the Item in the children
			foreach(FilterItem cur in Children)
			{
				if (cur.Value.ToString() == value.ToString())
				{
					return cur.Filtered;
				}
			}
			return false;
		}

		public	FilterItemList	Children	= new FilterItemList();
		public	int				UsedCount;
		public	object			Value;
		public	bool			Expanded	= false;
		public	bool			Filtered	= false;
	};
}