using System;
using System.Net;
using System.IO;
using System.Diagnostics;
using System.Web;
using System.Xml;
using System.Xml.Serialization;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.WorkItemTracking.Client;

namespace CegWIManager
{
	/// <summary>
	/// This class handles the Team System Eventing Service notification.
	/// Typically, when the Notify() WebMethod is called, an instance of this class is created then 
	/// ProcessWIChangedAlert() is called with the XML content passed to the Notify method.
	/// </summary>
	public class WINotifyHandler
	{
		bool					m_Connected		= false;
		TeamFoundationServer	m_TFS			= null;
		WorkItemStore			m_WIS			= null;
		EventLog				m_Log			= null;
		XmlDocument				m_DocWI			= null;
		XmlDocument				m_Model			= null;
		WINetworkModel			m_NetModel		= null;

		/// <summary>
		/// Unique constructor of the class
		/// </summary>
		/// <param name="xmodel">The XML Document containing the definition of the WINetwork model.</param>
		public WINotifyHandler(XmlDocument xmodel)
		{
			// Initialize the Log
			m_Log = new EventLog("Application", ".", "WINetwork");

			// Store the XML Document of the NetModel and Configuration
			m_Model = xmodel;
		}

		/// <summary>
		/// This method connects to the Team Foundation Server if it's not already done
		/// </summary>
		/// <returns><c>true</c> in case of success, <c>faslse</c> otherwise?</returns>
		private bool EnsureConnect()
		{
			// We check if the connection were already made, and quite promptly if it's the case
			if (m_Connected)
			{
				return true;
			}

			try
			{
				// We check if there's a TFSSettings XML Element that tells us how to connect to the server
				XmlElement xtfsconfig = m_Model.SelectSingleNode("/WIModel/TFSSettings") as XmlElement;
				if (xtfsconfig != null)
				{
					String		url		= xtfsconfig.GetAttribute("url");
					String		login	= xtfsconfig.GetAttribute("login");
					String		password= xtfsconfig.GetAttribute("password");
					String		domain	= xtfsconfig.GetAttribute("domain");

					// Need of specific credentials ?
					if (login!="")
					{
						NetworkCredential networkCredential = new NetworkCredential(login, password, domain);
						ICredentials credential = (ICredentials)networkCredential;

						// Server's URL
						m_TFS = new TeamFoundationServer(url, credential);
					}

					// If no login information was given, we use the current user
					else 
					{
						m_TFS = new TeamFoundationServer(url);
					}
				}

				// Get the Work Item Storage singleton
				m_WIS = m_TFS.GetService(typeof(WorkItemStore)) as WorkItemStore;

				// Create an instance of the Work Item Network Model object
				m_NetModel = new WINetworkModel(m_WIS, m_Model);

				// Yes, we're connected now
				m_Connected = true;

				return true;
			}
			catch (Exception)
			{
				throw;
			}
		}

		/// <summary>
		/// Get a field of type <c>string</c> in the WorkItemChangedEvent XML feed.
		/// </summary>
		/// <param name="name">Name of the field to get the value</param>
		/// <param name="newversion">If <c>true</c> returns the new version of the field, if <c>false</c> the old
		/// one will be.</param>
		/// <returns>The value corresponding to <c>name</c>, or an empty string if the field doesn't exists</returns>
		private string GetStringField(String name, bool newversion)
		{
			try
			{
				// Let's build the XPath request to get the field element
				String xpath = String.Format(@"WorkItemChangedEvent/CoreFields/StringFields/Field[ReferenceName='{0}']/{1}", name, newversion ? "NewValue" : "OldValue");
				return m_DocWI.SelectSingleNode(xpath).InnerText;
			}
			catch (Exception)
			{
				return String.Empty;
			}
		}

		/// <summary>
		/// Get a field of type <c>integer</c> in the WorkItemChangedEvent XML feed.
		/// </summary>
		/// <param name="name">Name of the field to get the value</param>
		/// <param name="newversion">If <c>true</c> returns the new version of the field, if <c>false</c> the old
		/// one will be.</param>
		/// <returns>The value corresponding to <c>name</c>, or 0 if the field doesn't exists</returns>
		private int GetIntField(String name, bool newversion)
		{
			try
			{
				// Let's build the XPath request to get the field element
				String xpath = String.Format(@"WorkItemChangedEvent/CoreFields/IntegerFields/Field[ReferenceName='{0}']/{1}", name, newversion ? "NewValue" : "OldValue");
				return int.Parse(m_DocWI.SelectSingleNode(xpath).InnerText);
			}
			catch (Exception)
			{
				return 0;
			}
		}

		/// <summary>
		/// Main method which processes the alert, everything is done here...
		/// </summary>
		/// <param name="xml">The XML feed passed through the Notify() Web Method.</param>
		/// <returns><c>true</c> if everything went ok, <c>false</c> otherwise</returns>
		public bool ProcessWIChangedAlert(string xml)
		{
			try
			{
				// Create an XML Document from the feed
				m_DocWI = new XmlDocument();
				m_DocWI.LoadXml(xml);

				// Ensure we're connected to the Team Foundation Server
				if (!EnsureConnect())
				{
					// Impossibility to connect, we quit
					return false;
				}

				// We get the actual time to perform some profiling
				TimeSpan	startprof = DateTime.Now.TimeOfDay;

				// We get the ID of the Work Item that triggered the event
				int id = GetIntField("System.Id", true);

				// Get the Work Item, check if it exists (apparently, there's no reason, but well, we check anyway)
				WorkItem wi = m_WIS.GetWorkItem(id);
				if (wi==null)
				{
					throw new Exception(String.Format("Unknown Work Item of ID: {0}", id));
				}

				// Write a log entry about the reception of the event
				m_Log.WriteEntry(String.Format("Received WIChanged on WorkItem [{0}] - {1}", wi.Id.ToString(), wi.Title),
					EventLogEntryType.Information);

				// Let's process the different Network Model for that Work Item
				String res = m_NetModel.ProcessModels(wi);

				// Compute the time it took, and write it in the log string
				TimeSpan delta = DateTime.Now.TimeOfDay - startprof;
				res += String.Format("Elapsed time for request execution: {0}\r\n", delta.ToString());

				// Is there something to log ? If so, let's log it!
				if (res != "")
				{
					EventLogEntryType type = (res.IndexOf("ERROR")==-1) ? EventLogEntryType.Information : EventLogEntryType.Error;
					m_Log.WriteEntry(res, type);
				}
			}

			// In case of a mysterious exception, we log it in the info event viewer
			catch (Exception e)
			{
				m_Log.WriteEntry(String.Format("Exception occurred {0}", e.ToString()),	EventLogEntryType.Error);
				return false;
			}
			return true;
		}
	}
}
