using System;
using System.Collections;
using System.Collections.Generic;
using System.Windows.Forms;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.WorkItemTracking.Client;
using Microsoft.TeamFoundation.VersionControl.Client;
using Microsoft.TeamFoundation.WorkItemTracking.Controls;
using WIManager;

namespace WICreator
{
	/// <summary>
	/// Type of WICObject
	/// </summary>
	public enum WICObjType
	{
		/// <summary>
		/// Database object
		/// </summary>
		DBObj		= 1,

		/// <summary>
		/// Dependency logical folder
		/// </summary>
		DepFolder	= 2,

		/// <summary>
		/// Dependency object
		/// </summary>
		DepObj		= 3
	}

	/// <summary>
	/// List of WICObject instances
	/// </summary>
	public class WICObjectList : List<WICObject>
	{
	};

	/// <summary>
	/// Argument class for the HierarchyChanged event
	/// </summary>
	public class WICObjectHierarchyEventArgs : EventArgs
	{
		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="insert">true for insertion, false for remove</param>
		/// <param name="wic">The WIC object that changed</param>
		public			WICObjectHierarchyEventArgs(bool insert, WICObject wic)
		{
			_Insert = insert;
			_WIC	= wic;
		}

		private	bool		_Insert;
		private WICObject	_WIC;
		
		/// <summary>
		/// Insert/Remove property.
		/// </summary>
		/// <remarks>If <c>true</c> the WIC is being inserted, <c>false</c> if it is removed</remarks>
		public	bool		Insert
		{
			get
			{
				return _Insert;
			}
		}

		/// <summary>
		/// Access to the WIC that triggered the event
		/// </summary>
		public WICObject	WIC
		{
			get
			{
				return _WIC;
			}
		}
	}

	/// <summary>
	/// Work Item Creator Object
	/// </summary>
	/// <remarks>This class handles a single object of the Work Item Creator database. Its main purpose is to
	/// reference the TreeNode associated with the data its stores (WorkItem, logical folder, dependency Work Item)
	/// </remarks>
	public class WICObject
	{
				WICObjType	   _Type;

		/// <summary>
		/// Type property. Get only.
		/// </summary>
		public	WICObjType		Type
		{
			get
			{
				return _Type;
			}
		}
				WINetworkModel	NetModel;

				bool	   _TextFiltered = false;

		/// <summary>
		/// Object is filtered due to Text Filtering
		/// </summary>
		public	bool		TextFiltered
		{
			get
			{
				return _TextFiltered;
			}

			set
			{
				_TextFiltered = value;
			}
		}

		/// <summary>
		/// Dirty state of the object corresponding to its Work Item
		/// </summary>
		public	bool		DirtyState	= false;

		/// <summary>
		/// Expanded state of the corresponding TreeNode
		/// </summary>
		public	bool		Expanded	= false;

				WorkItem   _WI			= null;

		/// <summary>
		/// Work Item associated with the WICObject
		/// </summary>
		public	WorkItem	WI
		{
			get
			{
				return _WI;
			}
		}

				TreeNode   _TN = null;

		/// <summary>
		/// TreeNode associated with the WICObject
		/// </summary>
		public	TreeNode	TN
		{
			get
			{
				return _TN;
			}

			set
			{
				if (value==null)
				{
					if (_TN!=null)
					{
						_TN.Remove();
					}

					foreach (WICObject cur in Children)
					{
						cur.TN = null;
					}
				}
				_TN = value;
			}
		}

				WICObject		   _Parent			= null;

		/// <summary>
		/// Access to the parent WICObject
		/// </summary>
		public	WICObject			Parent
		{
			get
			{
				return _Parent;
			}
		}

				bool			  _ChildrenLoaded	= false;

		/// <summary>
		/// Property that tells if the children are already loaded or not.
		/// </summary>
		public	bool			   ChildrenLoaded
		{
			get
			{
				return _ChildrenLoaded;
			}
		}
				WICObjectList		   _Children		= new WICObjectList();

		/// <summary>
		/// List of children
		/// </summary>
		/// <remarks><see cref="LoadChildren"/> must have been called at first.</remarks>
		public	WICObjectList			Children
		{
			get
			{
				return _Children;
			}
		}

		/// <summary>
		/// Access to the database root
		/// </summary>
		public	WICObject				Root
		{
			get
			{
				WICObject cur = this;
				while (cur.Parent != null)
				{
					cur = cur.Parent;
				}

				return cur;
			}
		}

		/// <summary>
		/// Delegate of the HierarchyChanged event
		/// </summary>
		/// <param name="sender">Sender object</param>
		/// <param name="e">Parameter data</param>
		public delegate void WICObjectHierarchyEventHandler(object sender, WICObjectHierarchyEventArgs e);

		/// <summary>
		/// HierarchyChanged event.
		/// </summary>
		/// <remarks>Is raised when a given WICObject is inserted or removed from the database</remarks>
		public event	WICObjectHierarchyEventHandler HierarchyChanged;

		/// <summary>
		/// Tells if the WICObject or at least one of his children are dirty.
		/// </summary>
		/// <param name="ignorenew">WICObject of newly created Work Item will be ignored if <c>true</c></param>
		/// <returns>Dirty state</returns>
		public	bool IsDirtyRecursive(bool ignorenew)
		{
			return _IsDirty(ignorenew);
		}

		/// <summary>
		/// Recursive part of <see cref="IsDirtyRecursive"/>
		/// </summary>
		/// <param name="ignorenew">WICObject of newly created Work Item will be ignored if <c>true</c></param>
		/// <returns>Dirty state</returns>
		private bool _IsDirty(bool ignorenew)
		{
			if (DirtyState)
			{
				if ((ignorenew==false) || (WI.Id!=0))
				{
					return true;
				}
			}

			// Recurse
			foreach(WICObject cur in Children)
			{
				if (cur._IsDirty(ignorenew))
				{
					return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Load the children of the WICObject
		/// </summary>
		/// <remarks>By default, the children are not loaded to make possible progressive loading of database content.
		/// You have to explicitly call this method to ensure the children are loaded or the <see cref="Children"/>
		/// property will be empty.</remarks>
		public	void LoadChildren()
		{
			// Children already loaded ? Return the list
			if (_ChildrenLoaded)
			{
				return;
			}

			// WIC is DBObj, parse the children of WI
			if (Type == WICObjType.DBObj)
			{
				// Get the first child
				WorkItem cur = NetModel.GetFirstChildWorkItem(WI);

				// As long as there's a child
				while (cur != null)
				{
					// Create the WICObject for that Work Item
					WICObject wic = new WICObject(NetModel, WICObjType.DBObj, cur);

					// Add the child
					AddChild(wic);

					// Get its next sibling
					cur = NetModel.GetNextWorkItem(cur);
				}
			}

			_ChildrenLoaded = true;
		}

		/// <summary>
		/// Add a child to the WICObject
		/// </summary>
		/// <param name="wic">The childrene to add to <c>this</c></param>
		/// <remarks>This method will raise the <see cref="HierarchyChanged"/> event</remarks>
		public	void AddChild(WICObject wic)
		{
			if (wic == null)
			{
				return;
			}

			_Children.Add(wic);
			wic._Parent = this;

			// Exec the notif
			WICObjectHierarchyEventArgs	e = new WICObjectHierarchyEventArgs(true, wic);
			WICObject root = Root;
			if (root != null)
			{
				root.HierarchyChanged(root, e);
			}
		}

		/// <summary>
		/// Remove this WICObject from the database
		/// </summary>
		/// <remarks>If the corresponding Work Item is dirty, changes will be reverted.
		/// The <see cref="HierarchyChanged"/> event will be raised.</remarks>
		public	void Remove()
		{
			// First, if there were changed made on the Work Item, revert them
			if (WI.IsDirty)
			{
				WI.Reset();
			}
			// Reset the hierarchy links (only for saved WIs)
			if (WI.Id != 0)
			{
				NetModel.SetFieldParentWI(WI, 0);
				NetModel.SetFieldNextWI(WI, 0);
				NetModel.AddToRecycle(WI);
			}

			// Remove the TreeNode if any
			if (TN!=null)
			{
				TN.Remove();
				TN = null;
			}

			// Unlink the object from its parent if any
			if (_Parent!=null)
			{
				// Exec the notification
				WICObjectHierarchyEventArgs	e = new WICObjectHierarchyEventArgs(false, this);
				WICObject root = Root;
				if (root != null)
				{
					root.HierarchyChanged(root, e);
				}

				_Parent.Children.Remove(this);
				_Parent = null;
			}

			// Clear the list of the children, we don't take care of them anymore
			_Children.Clear();

			// Remove the Work Item
			_WI = null;
		}

		/// <summary>
		/// Access to the previous sibling
		/// </summary>
		public WICObject Previous
		{
			get
			{
				if (Parent == null)
				{
					return null;
				}

				int i = Parent._Children.IndexOf(this);
				if ((i==-1) || (i==0))
				{
					return null;
				}

				return Parent._Children[i-1];
			}
		}

		/// <summary>
		/// Access to the next sibling
		/// </summary>
		public	WICObject		Next
		{
			get
			{
				if (Parent == null)
				{
					return null;
				}

				int i = Parent._Children.IndexOf(this);
				if ((i==-1) || ((i+1)==Parent._Children.Count))
				{
					return null;
				}

				return Parent._Children[i+1];
			}
		}

		// Dependencies
				WICObjectList		_Dependencies	= null;

		/// <summary>
		/// Access to the list of dependencies
		/// </summary>
		public	WICObjectList		Dependencies
		{
			get
			{
				return _Dependencies;
			}
		}


		/// <summary>
		/// Default Constructor
		/// </summary>
		/// <param name="netmodel">The Network Model used to parse the hierarchy</param>
		/// <param name="type">The type of the WICObject to construct</param>
		/// <param name="wi">The Work Item associated</param>
		public WICObject(WINetworkModel netmodel, WICObjType type, WorkItem wi)
		{
			NetModel	= netmodel;
			_Type		= type;
			_WI			= wi;
			DirtyState	= (wi.Id==0) ? false : wi.IsDirty;		// Make the Dirty State updated only for existing WIs
		}

		/// <summary>
		/// Determined if the WICObject is filtered (taking into account its children or not)
		/// </summary>
		/// <param name="recurse">If <c>true</c> children will also be analysed and they must be all filtered for
		/// the method to return <c>true</c></param>
		/// <returns>Return <c>true</c> if the WICObject is filtered, <c>false</c> otherwise</returns>
		public bool IsFiltered(bool recurse)
		{
			if (TextFiltered == false)
			{
				return false;
			}

			if (recurse)
			{
				foreach (WICObject cur in Children)
				{
					if (cur.IsFiltered(recurse) == false)
					{
						return false;
					}
				}
			}

			return true;
		}

		/// <summary>
		/// Move the WICObject into the first position among its siblings
		/// </summary>
		/// <returns>Success flag, will likely return <c>false</c> if the WICObject has no parent</returns>
		public bool MoveToFirst()
		{
			// Checking
			if (Parent == null)
			{
				return false;
			}

			// Get the index of this object into its parent's list
			int i = Parent._Children.IndexOf(this);

			// Already the first ?
			if (i==0)
			{
				return true;
			}

			// Move the object to the first position
			Parent._Children.RemoveAt(i);				// Remove it
			Parent._Children.Insert(0, this);			// Insert at the first position

			return true;
		}

		/// <summary>
		/// Move the WICObject into the last position among its siblings
		/// </summary>
		/// <returns>Success flag, will likely return <c>false</c> if the WICObject has no parent</returns>
		public bool MoveToLast()
		{
			// Checking
			if (Parent == null)
			{
				return false;
			}

			// Get the index of this object into its parent's list
			int i = Parent._Children.IndexOf(this);

			// Already the last ?
			if (i==(Parent._Children.Count-1))
			{
				return true;
			}

			// Move the object to the last position
			Parent._Children.RemoveAt(i);				// Remove it
			Parent._Children.Add(this);					// Insert at the last position

			return true;
		}

		/// <summary>
		/// Move the WICObject one position behind.
		/// </summary>
		/// <returns>Success flag, will likely return <c>false</c> if the WICObject has no parent</returns>
		public bool	MoveToPrevious()
		{
			// Checking
			if (Parent == null)
			{
				return false;
			}

			// Get the index of this object into its parent's list
			int i = Parent._Children.IndexOf(this);

			// The first ? Can't move it
			if (i==0)
			{
				return false;
			}

			// Move the object to the last position
			Parent._Children.RemoveAt(i);				// Remove it
			Parent._Children.Insert(i-1, this);			// Insert at the position before

			return true;
		}

		/// <summary>
		/// Move the WICObject one position ahead
		/// </summary>
		/// <returns>Success flag, will likely return <c>false</c> if the WICObject has no parent</returns>
		public bool MoveToNext()
		{
			// Checking
			if (Parent == null)
			{
				return false;
			}

			// Get the index of this object into its parent's list
			int i = Parent._Children.IndexOf(this);

			// Already the last ? Can't move it.
			if (i==(Parent._Children.Count-1))
			{
				return false;
			}

			// Move the object to the last position
			Parent._Children.RemoveAt(i);				// Remove it
			Parent._Children.Insert(i+1, this);			// Insert at the position after

			return true;
		}
	}
}
