using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using WICreatorAdmin.Properties;
using System.Xml;
using Microsoft.TeamFoundation.Server;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.WorkItemTracking.Client;
using Microsoft.TeamFoundation.VersionControl.Client;
using System.Security.Principal;
using TFSHelpers;
using System.IO;

namespace WICreatorAdmin
{
	public partial class MainForm : Form
	{
		#region Data and Properties
			protected	String		ConfigPFN;
			protected	XmlDocument	Config;
			protected	XmlElement	Projects;
			
			internal	bool		_ConfigDirty = false;
			protected	bool		ConfigDirty
			{
				get
				{
					return _ConfigDirty;
				}

				set
				{
					_ConfigDirty = value;
					saveConfigurationToolStripMenuItem.Enabled = value;
				}
			}

			TeamFoundationServer		TFS;
			WorkItemStore				WIS;
			VersionControlServer		SC;
			IEventService				EventSvc;
			Project						CurrentProject = null;
			TeamProject					CurrentTeamProject = null;

			SortedList<String, String>	WITList = new SortedList<string, string>();
		#endregion

		public MainForm()
		{
			InitializeComponent();
			ConfigDirty = false;
		}

		private void MainForm_Load(object sender, EventArgs e)
		{
			LoadConfig(true);
		}

		private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			e.Cancel = CloseConfig() == false;
		}

		#region Configuration File
		private bool SaveConfig()
		{
			// Commit the settings of the current project (if any)
			if (CurrentProject != null)
			{
				CommitProject(CurrentProject.Name);
			}

			// Commit the global settings
			CommitServerSettings();

			if (ConfigDirty == false)
			{
				return true;
			}

			saveConfigurationToolStripMenuItem.Enabled = false;
			Config.Save(ConfigPFN);
			ConfigDirty = false;
			StatusText.Text = "Configuration file saved.";

			return true;
		}

		private bool CloseConfig()
		{
			if (ConfigDirty == true)
			{
				DialogResult dr = MessageBox.Show("Do you want to save the Configuration ?", "Work Item Creator Admin", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question);
				if (dr == DialogResult.Cancel)
				{
					return false;
				}

				if (dr == DialogResult.Yes)
				{
					SaveConfig();
				}

				Config = null;
			}

			StatusText.Text = "Configuration file closed";
			closeConfigurationToolStripMenuItem.Enabled = false;

			return true;
		}

		private bool LoadConfig(bool usesettings)
		{
			if (CloseConfig() == false)
			{
				return false;
			}

			string xmlconfiglocation = usesettings ? Settings.Default.XMLConfigLocation : null;
			if ((xmlconfiglocation == null) || (xmlconfiglocation == ""))
			{
				OpenFileDialog	form = new OpenFileDialog();
				DirectoryInfo di = new DirectoryInfo(Application.ExecutablePath);
				form.InitialDirectory = di.Parent.Name;
				form.CheckFileExists = true;
				form.DefaultExt = "xml";
				form.FileName = "NetworkModels.xml";
				form.Multiselect = false;
				form.Title = "Choose the NetworkModels.xml file to administer";
				if (form.ShowDialog() != DialogResult.OK)
				{
					return false;
				}
				xmlconfiglocation = form.FileName;
			}

			try
			{
				// Load the XML Config file
				ConfigPFN = xmlconfiglocation;
				Config = new XmlDocument();
				Config.Load(ConfigPFN);

				// Check the root element
				XmlElement elem = Config.SelectSingleNode("WIModel") as XmlElement;
				if (elem == null)		throw new Exception();

				// Check the version of the file
				string version = elem.GetAttribute("version");
				if (version==null || version!="1.1")
				{
					DialogResult dr = MessageBox.Show("The version of the configuration file is outdated, you must upgrade it to continue", "Work Item Creator Admin", MessageBoxButtons.OKCancel, MessageBoxIcon.Exclamation);
					if (dr != DialogResult.OK)
					{
						return false;
					}

					SetupGUIFromConfig();

					// Set the right version
					elem.SetAttribute("version", "1.1");

					// Create a root element that will store the information for all project, one by one
					XmlElement projectsel = Config.CreateElement("Projects");
					elem.InsertBefore(projectsel, elem.ChildNodes[0]);
					ConfigDirty = true;
					SaveConfig();
				}

				Projects = elem.SelectSingleNode("Projects") as XmlElement;

				XmlElement adminel = elem.SelectSingleNode("WIAdmin") as XmlElement;

				// Not there ? We create one
				if (adminel == null)
				{
					DialogResult dr = MessageBox.Show(string.Format("Looks like this is the first time you run the admin tool, the account {0} will be used as the administrator, are you ok with it ?", WindowsIdentity.GetCurrent().Name), "Work Item Creator Admin", MessageBoxButtons.OKCancel, MessageBoxIcon.Information);
					if (dr != DialogResult.OK)
					{
						return false;
					}

					adminel = Config.CreateElement("WIAdmin");
					elem.InsertBefore(adminel, elem.ChildNodes[0]);
					XmlElement accountel = Config.CreateElement("AccountName");
					accountel.InnerText = WindowsIdentity.GetCurrent().Name;
					adminel.AppendChild(accountel);
					ConfigDirty = true;
				}

				// Check if the logged user is the admin
				else
				{
					if (WindowsIdentity.GetCurrent().Name != adminel.SelectSingleNode("AccountName").InnerText)
					{
						Config = null;
						MessageBox.Show("You are not logged with the account that can administrate the Work Item Creator!", "Work Item Creator Admin", MessageBoxButtons.OK, MessageBoxIcon.Error);
						return false;
					}
				}

			}
			catch (Exception)
			{
				Config = null;
				MessageBox.Show("This is not a valid configuration file", "Work Item Creator Admin", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}

			CheckWITCache();

			StatusText.Text = string.Format("Loaded the configuration: {0}", ConfigPFN);

			Settings.Default.XMLConfigLocation = ConfigPFN;
			Settings.Default.Save();

			closeConfigurationToolStripMenuItem.Enabled = true;

			SetupGUIFromConfig();

			return true;
		}

		private void loadConfigurationFileToolStripMenuItem_Click(object sender, EventArgs e)
		{
			LoadConfig(false);
		}

		private void saveConfigurationToolStripMenuItem_Click(object sender, EventArgs e)
		{
			SaveConfig();
		}

		private void closeConfigurationToolStripMenuItem_Click(object sender, EventArgs e)
		{
			CloseConfig();
		}

		private void CheckWITCache()
		{
			try
			{
				FileInfo fi = new FileInfo(ConfigPFN);
				XmlDocument webconfig = new XmlDocument();
				webconfig.Load(fi.DirectoryName + @"\" + "web.config");

				XmlNamespaceManager nsmgr = new XmlNamespaceManager(webconfig.NameTable);
				nsmgr.AddNamespace("nc", "http://schemas.microsoft.com/.NetConfiguration/v2.0");
				XmlElement el = webconfig.SelectSingleNode("nc:configuration/nc:appSettings/nc:add[@key=\"WorkItemTrackingCacheRoot\"]", nsmgr) as XmlElement;
				if (el == null)
				{
					el = webconfig.SelectSingleNode("configuration/appSettings/add[@key=\"WorkItemTrackingCacheRoot\"]") as XmlElement;
				}

				DirectoryInfo di = new DirectoryInfo(el.GetAttribute("value"));
				if (di.Exists == false)
				{
					MessageBox.Show(string.Format("WARNING, you must create a directory '{0}' that can be accessed by the WebService for the Work Item caching.\r\nIf you want to change the location, you still can modify the web.config.", di.FullName), "Work Item Creator Admin", MessageBoxButtons.OK, MessageBoxIcon.Error);
				}
			}
			catch (Exception ex)
			{
				MessageBox.Show(string.Format("Couldn't get the configuration of the Work Item caching directory in the web.config file (see documentation of Work Item Creator).\r\n", ex.ToString()), "Work Item Creator Admin", MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
		}

		private void SetupGUIFromConfig()
		{
			XmlElement tfsel = Config.SelectSingleNode("WIModel/TFSSettings") as XmlElement;
			if (tfsel == null)
			{
				return;
			}

			TFSURL.Text = tfsel.GetAttribute("url");
			TFSLogin.Text = tfsel.GetAttribute("login");
			TFSLoginPassword.Text = tfsel.GetAttribute("password");
			TFSLoginDomain.Text = tfsel.GetAttribute("domain");

			if (TFSURL.Text != "")
			{
				ConnectToTFS();
			}
			else
			{
				DataGridViewComboBoxColumn witlist = this.IconList.Columns[0] as DataGridViewComboBoxColumn;
				witlist.Items.Clear();
				IconList.Rows.Clear();
			}
		}

		void CommitServerSettings()
		{
			XmlElement tfsel = Config.SelectSingleNode("WIModel/TFSSettings") as XmlElement;
			if (tfsel == null)
			{
				return;
			}

			tfsel.SetAttribute("url", TFSURL.Text);
			if (TFSLogin.Text != "")			tfsel.SetAttribute("login", TFSLogin.Text);
			if (TFSLoginPassword.Text != "")	tfsel.SetAttribute("password", TFSLoginPassword.Text);
			if (TFSLoginDomain.Text != "")		tfsel.GetAttribute("domain", TFSLoginDomain.Text);

			XmlElement iconlist = Config.SelectSingleNode("WIModel/WIIcons") as XmlElement;
			foreach (DataGridViewRow row in IconList.Rows)
			{
				DataGridViewComboBoxCell combo = row.Cells[0] as DataGridViewComboBoxCell;

				if (combo.Value == null)		continue;

				XmlElement el = Config.SelectSingleNode(string.Format("WIModel/WIIcons/WorkItem[@type=\"{0}\"]", combo.Value as String)) as XmlElement;
				if (el == null)
				{
					el = Config.CreateElement("WorkItem");
					el.SetAttribute("type", combo.Value as String);
					iconlist.AppendChild(el);
				}
				el.SetAttribute("location", row.Cells[1].Value as String);
			}
		}

		private void ContentChanged(object sender, EventArgs e)
		{
			ConfigDirty = true;
		}

		#endregion

		#region TFS Related
		private bool ConnectToTFS()
		{
			// Clear the Icon list, and init the combo that list all the WITs
			WITList.Clear();
			DataGridViewComboBoxColumn witlist = this.IconList.Columns[0] as DataGridViewComboBoxColumn;
			witlist.Items.Clear();
			witlist.Items.Add("Default");

			// Connect to the Team Foundation Server, using the current credentials
			try
			{
				TFS = TeamFoundationServerFactory.GetServer(TFSURL.Text);
				WIS = TFS.GetService(typeof(WorkItemStore)) as WorkItemStore;
				SC  = TFS.GetService(typeof(VersionControlServer)) as VersionControlServer;
				EventSvc = (IEventService)TFS.GetService(typeof(IEventService));
			}
			catch (Exception)
			{
				TFSCheckResult.Text = "Wrong URL...";
				return false;
			}
			TFSCheckResult.Text = "Ok !";

			// Add in a dictionary all the types of Work Item of all Team Projects
			foreach (Project wip in WIS.Projects)
			{
				foreach (WorkItemType wit in wip.WorkItemTypes)
				{
					if (WITList.ContainsKey(wit.Name)==false)
					{
						WITList.Add(wit.Name, wit.Name);
					}
				}
			}
			
			// Set the content of the IconList combo box with all the available WITs
			foreach (String wit in WITList.Values)
			{
				witlist.Items.Add(wit);
			}

			// Setup the project panel
			TeamProjectList.Items.Clear();
			foreach (Project project in WIS.Projects)
			{
				TeamProjectList.Items.Add(project.Name);
			}
			ResetProjectSettings();

			XmlNodeList iconlistnodes = Config.SelectNodes("WIModel/WIIcons/WorkItem");
			foreach (XmlElement curel in iconlistnodes)
			{
				String type = curel.GetAttribute("type");
				if (WITList.Keys.Contains(type) || (type == "Default"))
				{
					IconList.Rows.Add(new string[] { type, curel.GetAttribute("location") });
				}
			}

			return true;
		}

		private void TFSCheckButton_Click(object sender, EventArgs e)
		{
			if (ConnectToTFS() == true)
			{
				ConfigDirty = true;
			}
		}
		#endregion

		#region Icon List Settings
		private void IconList_CellContentClick(object sender, DataGridViewCellEventArgs e)
		{
			if (e.RowIndex == -1)
			{
				return;
			}

			if (e.ColumnIndex == 2)
			{
				OpenFileDialog form = new OpenFileDialog();
				form.CheckFileExists = true;
				form.Multiselect = false;
				form.Title = "Choose the picture to associate with the Work Item type";
				form.InitialDirectory = Path.GetDirectoryName(ConfigPFN);
				if (form.ShowDialog() != DialogResult.OK)
				{
					return;
				}

				string relpath = form.FileName.Remove(0, form.InitialDirectory.Length + 1);

				(IconList.Rows[e.RowIndex].Cells[1] as DataGridViewTextBoxCell).Value = relpath;

				return;
			}
		}
		#endregion

		#region Project Settings Management
		private void TeamProjectList_SelectedIndexChanged(object sender, EventArgs e)
		{
			if (sender == TeamProjectList)
			{
				if (TeamProjectList.SelectedItem == null)
				{
					ResetProjectSettings();
					return;
				}

				// Commit current project in the XML File
				if (CurrentProject != null)
				{
					CommitProject(CurrentProject.Name);
				}

				// Get the new project to display
				String projectname = TeamProjectList.SelectedItem as string;
				CurrentProject = WIS.Projects[projectname];
				CurrentTeamProject = SC.GetTeamProject(projectname);

				// Setup the GUI for that new project
				SetupProject(CurrentProject);
			}
		}

		private void ResetProjectSettings()
		{
			RecycleWI.Text = "";
			ChooseRecycleWIBtn.Enabled = false;
			ChooseRecycleAreaBtn.Enabled = false;
			RecycleArea.Text = "";
			RemoveRecycleBtn.Enabled = false;
			AlertStatus.Text = "";
			ChangeAlertStatusBtn.Enabled = false;
			ChangeIntegrationRoleBtn.Enabled = false;
			ChangeHierarchyFieldsBtn.Enabled = false;
		}
		private void SetupProject(Project project)
		{
			//Get/Create the project XML Element
			XmlElement projectel = Projects.SelectSingleNode(string.Format("Project[@name=\"{0}\"]", project.Name)) as XmlElement;
			if (projectel == null)
			{
				projectel = Config.CreateElement("Project");
				projectel.SetAttribute("name", project.Name);
				Projects.AppendChild(projectel);
			}

			// Recycle Bin display
			XmlElement recycleel = projectel.SelectSingleNode("RecycleBin") as XmlElement;
			if (recycleel != null)
			{
				RecycleWI.Text = recycleel.GetAttribute("id");
				RecycleArea.Text = recycleel.GetAttribute("areapath");
			}
			else
			{
				RecycleWI.Text = "";
				RecycleArea.Text = "";
			}

			// Integration
			XmlElement integel = projectel.SelectSingleNode("Integration") as XmlElement;
			if (integel != null)
			{
				IntegrationRoleText.Text = integel.GetAttribute("tfsrole");
			}
			else
			{
				IntegrationRoleText.Text = "";
			}

			// Alert
			UpdateAlertStatus(project);

			// Hierarchy Fields
			UpdateHierarchyStatus(project);

			// Enable the buttons
			ChooseRecycleWIBtn.Enabled = true;
			ChooseRecycleAreaBtn.Enabled = true;
			RemoveRecycleBtn.Enabled = true;
			ChangeAlertStatusBtn.Enabled = true;
			ChangeIntegrationRoleBtn.Enabled = true;
			ChangeHierarchyFieldsBtn.Enabled = true;
		}

		private void CommitProject(String projectname)
		{
			if (projectname==null)
			{
				return;
			}

			XmlElement projectel = Projects.SelectSingleNode(string.Format("Project[@name=\"{0}\"]", projectname)) as XmlElement;
			if (projectel==null)
			{
				return;
			}

			// Recycle Bin
			XmlElement recycleel = projectel.SelectSingleNode("RecycleBin") as XmlElement;

			// Remove ?
			if (RecycleWI.Text == "")
			{
				if (recycleel != null)
				{
					recycleel.ParentNode.RemoveChild(recycleel);
				}
			}

			// Change
			else
			{
				if (recycleel == null)
				{
					recycleel = Config.CreateElement("RecycleBin");
					projectel.AppendChild(recycleel);
				}

				recycleel.SetAttribute("id", RecycleWI.Text);
				if (RecycleArea.Text != "")
				{
					recycleel.SetAttribute("areapath", RecycleArea.Text);
				}
				else
				{
					recycleel.RemoveAttribute("areapath");
				}
			}

			// Integration
			XmlElement integel = projectel.SelectSingleNode("Integration") as XmlElement;

			// Remove ?
			if (IntegrationRoleText.Text == "")
			{
				if (integel != null)
				{
					integel.ParentNode.RemoveChild(integel);
				}
			}

			// Change
			else
			{
				if (integel == null)
				{
					integel = Config.CreateElement("Integration");
					projectel.AppendChild(integel);
				}

				integel.SetAttribute("tfsrole", IntegrationRoleText.Text);
			}

		}

		#endregion

		#region Recycle Bin
		private void ChooseRecycleWIBtn_Click(object sender, EventArgs e)
		{
			WISelectionForm form = new WISelectionForm(TFS, WIS);
			form.SetTeamProject(CurrentProject);
			DialogResult dr = form.ShowDialog();
			if (dr != DialogResult.OK)
			{
				return;
			}

			int wiid = (int)form.m_SelectedWorkItemID;
			RecycleWI.Text = wiid.ToString();
		}

		private void ChooseRecycleAreaBtn_Click(object sender, EventArgs e)
		{
			BrowseAreaIteration form = new BrowseAreaIteration(TFS, CurrentTeamProject, true);
			DialogResult dr = form.ShowDialog();
			if (dr != DialogResult.OK)
			{
				return;
			}

			string areapath = form.m_SelItem.Text;
			RecycleArea.Text = areapath;
		}

		private void RemoveRecycleBtn_Click(object sender, EventArgs e)
		{
			RecycleWI.Text = "";
			RecycleArea.Text = "";

		}
		#endregion

		#region Alert
		private void UpdateAlertStatus(Project project)
		{
			Subscription[] subs = EventSvc.EventSubscriptions(WindowsIdentity.GetCurrent().Name);
			Subscription sub = GetAlertForProject(project, subs);
			if (sub != null)
			{
				AlertStatus.Text = "Registered";
				ChangeAlertStatusBtn.Text = "Unregister...";
				ChangeAlertStatusBtn.Tag = sub;
			}
			else
			{
				AlertStatus.Text = "Not registered";
				ChangeAlertStatusBtn.Text = "Register...";
				ChangeAlertStatusBtn.Tag = null;
			}
		}
		private Subscription GetAlertForProject(Project wip, Subscription[] sub)
		{
			foreach (Subscription cur in sub)
			{
				if (cur.EventType != "WorkItemChangedEvent") continue;
				if (cur.DeliveryPreference.Type != DeliveryType.Soap) continue;
				if (cur.ConditionString.Contains(string.Format("'{0}'", wip.Name)) == false) continue;
				if (cur.DeliveryPreference.Address.Contains("WINetwork/Service.asmx") == false) continue;

				return cur;
			}
			return null;
		}

		private void ChangeAlertStatusBtn_Click(object sender, EventArgs e)
		{
			// Register
			if (ChangeAlertStatusBtn.Tag == null)
			{
				string expr = string.Format("\"PortfolioProject\" = '{0}'", CurrentProject.Name);
				DeliveryPreference delivery = new DeliveryPreference();
				delivery.Address = String.Format("{0}WINetwork/Service.asmx", TFS.Uri.AbsoluteUri);
				delivery.Schedule = DeliverySchedule.Immediate;
				delivery.Type = DeliveryType.Soap;
				StatusText.Text = string.Format("Registering alert for '{0}'...", CurrentProject.Name);
				EventSvc.SubscribeEvent(WindowsIdentity.GetCurrent().User.ToString(), "WorkItemChangedEvent", expr, delivery);
				StatusText.Text = string.Format("Registering alert for '{0}' done!", CurrentProject.Name);
			}

			// Unregister
			else
			{
				Subscription cursub = ChangeAlertStatusBtn.Tag as Subscription;
				StatusText.Text = string.Format("Unregistering alert for '{0}'...", CurrentProject.Name);
				EventSvc.UnsubscribeEvent(cursub.ID);
				StatusText.Text = string.Format("Unregistering alert for '{0}' done!", CurrentProject.Name);
			}
			UpdateAlertStatus(CurrentProject);
		}
		#endregion

		#region Integration
		private void ChangeIntegrationRoleBtn_Click(object sender, EventArgs e)
		{
			SelectSecurityGroup form = new SelectSecurityGroup(TFS, CurrentTeamProject, true);
			DialogResult dr = form.ShowDialog();
			if (dr != DialogResult.OK)
			{
				return;
			}

			IntegrationRoleText.Text = form.SelectedGroup;
		}

		private void RemoveIntegrationRoleBtn_Click(object sender, EventArgs e)
		{
			IntegrationRoleText.Text = "";
		}
		#endregion

		#region Upgrade WITs
		private void UpdateHierarchyStatus(Project project)
		{
			HierarchyFieldsStatus.Text = "Included";
			ChangeHierarchyFieldsBtn.Text = "Remove them...";
			HierarchyFieldsStatus.Tag = (bool)true;
			foreach (WorkItemType wit in project.WorkItemTypes)
			{
				try
				{
					FieldDefinition f = wit.FieldDefinitions["Cegedim.ParentWI"];
					f = wit.FieldDefinitions["Cegedim.FirstChildWI"];
					f = wit.FieldDefinitions["Cegedim.NextWI"];
				}
				catch (Exception)
				{
					HierarchyFieldsStatus.Text = "No present";
					ChangeHierarchyFieldsBtn.Text = "Add them...";
					HierarchyFieldsStatus.Tag = (bool)false;
					break;
				}
			}
		}

		private void ChangeHierarchyFieldsBtn_Click(object sender, EventArgs e)
		{
			UpgradeWITs(CurrentProject.Name, !(bool)HierarchyFieldsStatus.Tag);
		}

		bool WITValidateSucceed = true;
		bool UpgradeWITs(String prjname, bool add)
		{
			List<String> witnames = new List<string>();
			{
				Project prj = WIS.Projects[prjname];
				prj.WorkItemTypes.ImportEventHandler += new ImportEventHandler(WorkItemType_ValidationEventHandler);
				foreach (WorkItemType wit in prj.WorkItemTypes)
				{
					witnames.Add(wit.Name);
				}
			}

			foreach (String witname in witnames)
			{
				WIS.RefreshCache();
				WIS.SyncToCache();

				Project prj = WIS.Projects[prjname];
				WorkItemType wit = prj.WorkItemTypes[witname];

				if (add)
				{
					try
					{
						FieldDefinition f = wit.FieldDefinitions["Cegedim.ParentWI"];
						f = wit.FieldDefinitions["Cegedim.FirstChildWI"];
						f = wit.FieldDefinitions["Cegedim.NextWI"];
					}
					catch (Exception)
					{
						UpgradeWIT(prj, wit);
					}
				}
				else
				{
					RemoveWITFields(prj, wit);
				}
			}

			WIS.RefreshCache();
			WIS.SyncToCache();

			UpdateHierarchyStatus(WIS.Projects[prjname]);
			
			return true;
		}

		bool UpgradeWIT(Project prj, WorkItemType wit)
		{
			StatusText.Text = String.Format("Modifying definition of '{0}'...", wit.Name);
			Application.DoEvents();

			// Get the Xml Document for the given WIT
			XmlDocument doc = wit.Export(false);
			XmlElement fields = doc.SelectSingleNode("*/WORKITEMTYPE/FIELDS") as XmlElement;

			// Look for each field, and add the ones that are not present
			XmlElement field = fields.SelectSingleNode("FIELD[@refname=\"Cegedim.ParentWI\"]") as XmlElement;
			if (field == null)
			{
				field = doc.CreateElement("FIELD");
				field.SetAttribute("name", "ParentWI");
				field.SetAttribute("refname", "Cegedim.ParentWI");
				field.SetAttribute("type", "Integer");
				field.InnerXml = "<HELPTEXT>The ID of the Parent Work Item</HELPTEXT><DEFAULT from=\"value\" value=\"0\" />";
				fields.AppendChild(field);
			}

			field = fields.SelectSingleNode("FIELD[@refname=\"Cegedim.FirstChildWI\"]") as XmlElement;
			if (field == null)
			{
				field = doc.CreateElement("FIELD");
				field.SetAttribute("name", "FirstChildWI");
				field.SetAttribute("refname", "Cegedim.FirstChildWI");
				field.SetAttribute("type", "Integer");
				field.InnerXml = "<HELPTEXT>The ID of the first children</HELPTEXT><DEFAULT from=\"value\" value=\"0\" />";
				fields.AppendChild(field);
			}

			field = fields.SelectSingleNode("FIELD[@refname=\"Cegedim.NextWI\"]") as XmlElement;
			if (field == null)
			{
				field = doc.CreateElement("FIELD");
				field.SetAttribute("name", "NextWI");
				field.SetAttribute("refname", "Cegedim.NextWI");
				field.SetAttribute("type", "Integer");
				field.InnerXml = "<HELPTEXT>The ID of the next Work Item</HELPTEXT><DEFAULT from=\"value\" value=\"0\" />";
				fields.AppendChild(field);
			}

			// Validate the type
			try
			{
				WITValidateSucceed = true;
				Microsoft.TeamFoundation.WorkItemTracking.Client.WorkItemType.Validate(prj, doc.OuterXml);
				if (WITValidateSucceed == false)
				{
					MessageBox.Show("Work Item Type couldn't pass validation, something went wrong", "Work Item Creator Admin", MessageBoxButtons.OK, MessageBoxIcon.Error);
					return false;
				}

				prj.WorkItemTypes.Import(doc.OuterXml);

			}
			catch (Exception ex)
			{
				MessageBox.Show(string.Format("Work Item Type couldn't be imported.\r\n{0}", ex.ToString()), "Work Item Creator Admin", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}

			StatusText.Text = String.Format("Hierarchy fields was added to '{0}'...", wit.Name);
			Application.DoEvents();

			return true;
		}

		bool RemoveWITFields(Project prj, WorkItemType wit)
		{
			StatusText.Text = String.Format("Modifying definition of '{0}'...", wit.Name);
			Application.DoEvents();

			// Get the Xml Document for the given WIT
			XmlDocument doc = wit.Export(false);
			XmlElement fields = doc.SelectSingleNode("*/WORKITEMTYPE/FIELDS") as XmlElement;

			// Look for each field, and add the ones that are not present
			XmlElement field = fields.SelectSingleNode("FIELD[@refname=\"Cegedim.ParentWI\"]") as XmlElement;
			if (field != null)
			{
				field.ParentNode.RemoveChild(field);
			}

			field = fields.SelectSingleNode("FIELD[@refname=\"Cegedim.FirstChildWI\"]") as XmlElement;
			if (field != null)
			{
				field.ParentNode.RemoveChild(field);
			}

			field = fields.SelectSingleNode("FIELD[@refname=\"Cegedim.NextWI\"]") as XmlElement;
			if (field != null)
			{
				field.ParentNode.RemoveChild(field);
			}

			// Validate the type
			try
			{
				WITValidateSucceed = true;
				Microsoft.TeamFoundation.WorkItemTracking.Client.WorkItemType.Validate(prj, doc.OuterXml);
				if (WITValidateSucceed == false)
				{
					MessageBox.Show("Work Item Type couldn't pass validation, something went wrong", "Work Item Creator Admin", MessageBoxButtons.OK, MessageBoxIcon.Error);
					return false;
				}

				prj.WorkItemTypes.Import(doc.OuterXml);

			}
			catch (Exception ex)
			{
				MessageBox.Show(string.Format("Work Item Type couldn't be imported.\r\n{0}", ex.ToString()), "Work Item Creator Admin", MessageBoxButtons.OK, MessageBoxIcon.Error);
				return false;
			}

			StatusText.Text = String.Format("Hierarchy fields was removed from '{0}'...", wit.Name);
			Application.DoEvents();

			return true;
		}

		void WorkItemType_ValidationEventHandler(object sender, ImportEventArgs e)
		{
			WITValidateSucceed = e.Severity == ImportSeverity.Error;
		}
		#endregion

		}
}






