using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Text;
using System.Xml;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.WorkItemTracking.Client;

namespace WIManager
{

	public class WITemplate
	{
		protected String _Name;
		public String Name
		{
			get
			{
				return _Name;
			}
		}

		protected String _Comments;
		public String Comments
		{
			get
			{
				return _Comments;
			}
		}

		protected XmlElement _Description;
		public XmlElement Description
		{
			get
			{
				return _Description;
			}
		}

		protected String _Projects;

		public WITemplate(string name, string comments, string projects, XmlElement desc)
		{
			_Name = name;
			_Comments = comments;
			_Description = desc;
			_Projects = "$" + projects + "$";
		}

		public bool IsForProject(string project)
		{
			return (_Projects == "$$") || _Projects.Contains("$" + project + "$");
		}

		public bool IsRootOfType(ArrayList witlist)
		{
			try
			{
				XmlElement elem = Description.SelectSingleNode("WorkItem") as XmlElement;
				String roottype = elem.GetAttribute("type");
				foreach (WorkItemType wit in witlist)
				{
					if (wit.Name == roottype)
					{
						return true;
					}
				}
			}
			catch (Exception)
			{
				return false;
			}
			return false;
		}
	}

	/// <summary>
	/// This class handles a cache of Work Items, it makes sure there's only one instance of the 
	/// <see cref="Microsoft.TeamFoundation.WorkItemTracking.Client.WorkItem"/> class per ID.
	/// </summary>
	public class WICache : List<WorkItem>
	{
		WorkItemStore m_WIS;
		bool m_AutoOpen;
		bool m_TrackChanges;

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="wis">The Work Item Store to use</param>
		/// <param name="autoopen">If <c>true</c> the Work Item will be automatically opened at the first Get</param>
		/// <param name="trackfieldchanges">If <c>true</c> changes on the Work Item will be tracked, use the 
		/// <paramref name="WIFieldChanged"/> event to be aware of them.</param>
		public WICache(WorkItemStore wis, bool autoopen, bool trackfieldchanges)
		{
			m_WIS = wis;
			m_AutoOpen = autoopen;
			m_TrackChanges = trackfieldchanges;
		}

		/// <summary>
		/// Get the Work Item instance corresponding to the given ID
		/// </summary>
		/// <param name="id">The ID of the Work Item to retrieve</param>
		/// <returns>The Work Item or <c>null</c> if the ID is invalid</returns>
		/// <remarks>If there's already a Work Item instance of that ID present in the cache, it'll be returned.
		/// Otherwise we'll get one from the WIS</remarks>
		public WorkItem Get(int id)
		{
			// Can't be 0
			if (id == 0)
			{
				return null;
			}

			// Look in the list
			foreach (WorkItem cur in this)
			{
				// Found ? Return it.
				if (cur.Id == id)
				{
					return cur;
				}
			}

			// Not found, get it from the WIS
			try
			{
				// Get the Work Item
				WorkItem wi = m_WIS.GetWorkItem(id);

				// Open the Work Item if asked
				if (m_AutoOpen)
				{
					wi.Open();
				}

				// Edit Mode ? Add a notification to the field changed event
				if (m_TrackChanges)
				{
					wi.FieldChanged += new WorkItemFieldChangeEventHandler(WIFieldChangedHandler);
				}

				// Add the Work Item to the list
				Add(wi);

				// Return it
				return wi;
			}
			catch (Exception)
			{
				return null;
			}
		}

		/// <summary>
		/// Create a new Work Item
		/// </summary>
		/// <param name="wit">Type of the Work Item to create</param>
		/// <returns>The created Work Item</returns>
		/// <remarks>The Work Item is added to the WI Pool of the NetModel</remarks>
		public WorkItem Create(WorkItemType wit)
		{
			// Create the Work Item instance
			WorkItem wi = new WorkItem(wit);

			// Open the Work Item if asked
			if (m_AutoOpen)
			{
				wi.Open();
			}

			// Add a notification to track changes if needed
			if (m_TrackChanges)
			{
				wi.FieldChanged += new WorkItemFieldChangeEventHandler(WIFieldChangedHandler);
			}

			// Add the Work Item to the pool
			Add(wi);

			// Return the Work Item
			return wi;
		}

		/// <summary>
		/// Event that will be raised each time a Work Item in the Cache will have a field change
		/// </summary>
		public event WorkItemFieldChangeEventHandler WIFieldChanged;

		// Handler registered on each Work Item we got (edit mode only)
		// The handler will route the event, but at the WICache level (one handler for all the WI stored is enough)
		void WIFieldChangedHandler(object sender, WorkItemEventArgs e)
		{
			if (WIFieldChanged != null)
			{
				WIFieldChanged(sender, e);
			}
		}

		/// <summary>
		/// Save the Work Item Write Cache
		/// </summary>
		/// <param name="res">Result string, for information purpose</param>
		/// <remarks>Parses all the Work Item present in the cache, and save them.</remarks>
		public void Save(ref String res)
		{
			foreach (WorkItem cur in this)
			{
				if (cur.IsDirty == true)
				{
					cur.Validate();
					if (cur.IsValid())
					{
						try
						{
							cur.Save();
						}
						catch (Exception ex)
						{
							res += ex.ToString() + "\r\n";
						}
					}
					else
					{
						res += String.Format("Can't Save Work Item {0} because it's not valid\r\n", cur.Id.ToString());
					}
				}
			}
		}
	};

	/// <summary>
	/// Work Item Network Model implementation
	/// </summary>
	public class WINetworkModel
	{
		// We implement a set of delegates to let the possibility to handle the relationship in another way than
		// the Work Item hierarchy links.

		/// <summary>
		/// Retrieve the parent of the given object.
		/// </summary>
		/// <param name="obj">The object to get the parent</param>
		/// <returns>The parent object, or <c>null</c> if there's no parent</returns>
		public delegate Object DelGetParent(Object obj);

		/// <summary>
		/// Get all the children of a given object.
		/// </summary>
		/// <param name="obj">The object to get the children</param>
		/// <returns>An enumerable interface containing all the children</returns>
		public delegate IEnumerable DelGetChildren(Object obj);

		/// <summary>
		/// Get the Work Item from the given object
		/// </summary>
		/// <param name="obj">The object to get the Work Item from</param>
		/// <returns>The corresponding Work Item or <c>null</c> if not found</returns>
		public delegate WorkItem DelFromObject(Object obj);

		public delegate Object DelCreateFromWI(WorkItem wi, Object parent);

		/// <summary>
		/// The GetParent delegate, will return the parent of the given object
		/// </summary>
		public DelGetParent GetParent = null;

		/// <summary>
		/// The GetChildren delegate, will return all the children of the given object
		/// </summary>
		public DelGetChildren GetChildren = null;

		/// <summary>
		/// The FromObject delegate, will return the Work Item corresponding to the given object
		/// </summary>
		public DelFromObject FromObject = null;

		public DelCreateFromWI CreateFromWI = null;

		/// <summary>
		/// Work Item Store from the Team Foundation Server
		/// </summary>
		WorkItemStore m_WIS = null;

		/// <summary>
		/// The Work Item Tracking Project corresponding to the current Team Project
		/// </summary>
		Project m_WIP = null;

		/// <summary>
		/// The XML Document containing the Network Model rules and definition
		/// </summary>
		XmlDocument m_Doc = null;

		/// <summary>
		/// Work Item pool used by the WINetwork for the modified Work Items that will be saved at the end of the process
		/// </summary>
		WICache m_WINetToSave = null;

		/// <summary>
		/// Main Work Item Pool for the WINetworkModel.
		/// This instance contains all the opened Work Items, in edit mode, but not automatically opened.
		/// </summary>
		public WICache m_WIPool = null;

		/// <summary>
		/// This list contains the Work Item to recycle (save when possible, then remove)
		/// </summary>
		public List<WorkItem> m_ToRecycle = new List<WorkItem>();

		/// <summary>
		/// The list of Work Item Templates available
		/// </summary>
		public Dictionary<String, WITemplate> m_WITemplates = new Dictionary<string, WITemplate>();

		/// <summary>
		/// Add a Work Item to the ToRecycle list (make sure only it's referenced only one time)
		/// </summary>
		/// <param name="wi"></param>
		public void AddToRecycle(WorkItem wi)
		{
			if (m_ToRecycle.Contains(wi))
			{
				return;
			}
			m_ToRecycle.Add(wi);
		}

		/// <summary>
		/// Get a Work Item from its ID. This method must be used instead of a direct access from the Work Item Store
		/// because if the Work Item already exists, the instance is returned. Access to the Work Item Store are
		/// very costly, this is why there's a cache in place.
		/// </summary>
		/// <param name="id">ID of the Work Item to retrieve</param>
		/// <returns>The Work Item object or <c>null</c> if the ID is invalid</returns>
		public WorkItem GetWorkItem(int id)
		{
			if (m_WIPool == null)
			{
				return null;
			}

			return m_WIPool.Get(id);
		}

		///////////////////////////////////////////////////////////////////////////////////////////////////////////////
		//							WORK ITEM DELEGATE IMPLEMENTATION (default implementation)
		///////////////////////////////////////////////////////////////////////////////////////////////////////////////
		/// <summary>
		/// Retrieve the parent of a given Work Item
		/// </summary>
		/// <param name="obj">Is a Work Item typed object, its parent will be returned</param>
		/// <returns>A Work Item typed object which is the parent of <paramref name="obj"/></returns>
		/// <remarks>This methode simply routes to WIGetParent(WorkItem)</remarks>
		public Object GetParentUsingWI(Object obj)
		{
			return WIGetParent(obj as WorkItem);
		}

		/// <summary>
		/// Get the Work Item object corresponding to the given object
		/// </summary>
		/// <param name="obj">The object to get the corresponding Work Item</param>
		/// <returns>The corresponding Work Item</returns>
		/// <remarks>This implementation does nothing but returning a casted version to WorkItem of <paramref name="obj"/></remarks>
		public WorkItem FromObjectUsingWI(Object obj)
		{
			if (obj == null)
			{
				return null;
			}
			return obj as WorkItem;
		}

		public Object CreateFromWIUsingWI(WorkItem wi, Object parent)
		{
			return wi;
		}

		/// <summary>
		/// Get all the children of the given object
		/// </summary>
		/// <param name="obj">The object to get the children</param>
		/// <returns>An enumerable object containing all the children</returns>
		/// <remarks>This method is the Work Item implementation, it relies on the Work Item Links to find the children</remarks>
		public IEnumerable GetChildrenUsingWI(Object obj)
		{
			return WIGetChildren(obj as WorkItem);
		}

		///////////////////////////////////////////////////////////////////////////////////////////////////////////////
		//									HIERARCHY MANAGEMENT WITH WORK ITEMS
		///////////////////////////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>
		/// Retrieve the parent of the given Work Item
		/// </summary>
		/// <param name="wi">The Work Item to get the parent</param>
		/// <returns>The parent Work Item or <c>null</c> if there's no parent</returns>
		public WorkItem WIGetParent(WorkItem wi)
		{
			if (wi == null)
			{
				return null;
			}

			int parentid;
			try
			{
				parentid = (int)wi.Fields["ParentWI"].Value;
			}
			catch (Exception)
			{
				return null;
			}

			if ((parentid == 0) || (parentid == -1))
			{
				return null;
			}

			WorkItem parent = null;
			try
			{
				parent = GetWorkItem(parentid);
			}
			catch (Exception)
			{
				return null;
			}

			return parent;
		}

		/// <summary>
		/// Get all the children of the given Work Item
		/// </summary>
		/// <param name="wi">The Work Item to get the children</param>
		/// <returns>An enumerable that contains the children Work Items</returns>
		public IEnumerable WIGetChildren(WorkItem wi)
		{
			if (wi == null)
			{
				return null;
			}

			// We create the array that will store the children
			ArrayList array = new ArrayList();

			// There're two ways to evaluate the children
			// 1) With the Links (legacy, was the only way before ordering was implemented)
			// 2) With the "FirstChildWI" and "NextWI" fields
			// We favor the 2) of course, if it's available, otherwise we use the 1)

			// Can we use ordered hierarchy for the evaluation ?
			if (GetFieldFirstChildWI(wi) != 0)
			{
				WorkItem cur = GetFirstChildWorkItem(wi);
				while (cur != null)
				{
					array.Add(cur);
					cur = GetNextWorkItem(cur);
				}
			}

			// We use the Links as the ordered hierarchy is not available
			else
			{
				// On parse les enfants
				foreach (Link link in wi.Links)
				{
					// <WIManager> en commentaire pour signifier un lien  grer
					if (link.Comment == "<WIManager>")
					{
						// On rcupre le Work Item associ
						RelatedLink rl = link as RelatedLink;
						if (rl == null)
						{
							continue;
						}
						WorkItem linkedwi = GetWorkItem(rl.RelatedWorkItemId);
						if (linkedwi == null)
						{
							continue;
						}

						// On test si c'est bien un lien descendant
						if (!IsFatherAndSon(wi, linkedwi))
						{
							continue;
						}

						// On ajoute le fils au tableau et on reboucle
						array.Add(linkedwi);
					}
				}
			}
			return array;
		}

		///////////////////////////////////////////////////////////////////////////////////////////////////////////////
		//										HIERARCHY FIELDS ACCESS AND SETUP
		///////////////////////////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>
		/// Set the 'ParentWI' field of the given Work Item
		/// </summary>
		/// <param name="wi">The Work Item to set the field</param>
		/// <param name="parentid">The value of the field to set</param>
		/// <remarks>wi validity is not checked, if the Work Item is not opened, it will be.</remarks>
		public void SetFieldParentWI(WorkItem wi, int parentid)
		{
			if (!wi.IsOpen)
			{
				wi.Open();
			}
			wi.Fields["ParentWI"].Value = parentid;
		}

		/// <summary>
		/// Get the 'ParentWI' field of the given Work Item
		/// </summary>
		/// <param name="wi">The Work Item to get the field value</param>
		/// <returns>The value</returns>
		/// <remarks>If an exception occured, <c>0</c> will be returned</remarks>
		public int GetFieldParentWI(WorkItem wi)
		{
			try
			{
				return (int)wi.Fields["ParentWI"].Value;
			}
			catch (Exception)
			{
				return 0;
			}
		}

		/// <summary>
		/// Set the 'FirstChildWI' field of the given Work Item
		/// </summary>
		/// <param name="wi">The Work Item to set the field</param>
		/// <param name="firstchildwi">The value of the field to set</param>
		/// <remarks>wi validity is not checked, if the Work Item is not opened, it will be.</remarks>
		public void SetFieldFirstChildWI(WorkItem wi, int firstchildwi)
		{
			if (!wi.IsOpen)
			{
				wi.Open();
			}
			wi.Fields["FirstChildWI"].Value = firstchildwi;
		}

		/// <summary>
		/// Get the 'FirstChildWI' field of the given Work Item
		/// </summary>
		/// <param name="wi">The Work Item to get the field value</param>
		/// <returns>The value</returns>
		/// <remarks>If an exception occured, <c>0</c> will be returned</remarks>
		public int GetFieldFirstChildWI(WorkItem wi)
		{
			try
			{
				return (int)wi.Fields["FirstChildWI"].Value;
			}
			catch (Exception)
			{
				return 0;
			}
		}

		/// <summary>
		/// Set the 'NextWI' field of the given Work Item
		/// </summary>
		/// <param name="wi">The Work Item to set the field</param>
		/// <param name="nextid">The value of the field to set</param>
		/// <remarks>wi validity is not checked, if the Work Item is not opened, it will be.</remarks>
		public void SetFieldNextWI(WorkItem wi, int nextid)
		{
			if (!wi.IsOpen)
			{
				wi.Open();
			}
			wi.Fields["NextWI"].Value = nextid;
		}

		/// <summary>
		/// Get the 'NextWI' field of the given Work Item
		/// </summary>
		/// <param name="wi">The Work Item to get the field value</param>
		/// <returns>The value</returns>
		/// <remarks>If an exception occured, <c>0</c> will be returned</remarks>
		public int GetFieldNextWI(WorkItem wi)
		{
			try
			{
				return (int)wi.Fields["NextWI"].Value;
			}
			catch (Exception)
			{
				return 0;
			}
		}

		///////////////////////////////////////////////////////////////////////////////////////////////////////////////
		//												HIERARCHY HELPERS
		///////////////////////////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>
		/// Determine if there's a father/son relationship between two Work Items
		/// </summary>
		/// <param name="father">The Work Item that is supposed to be the father</param>
		/// <param name="son">The Work Item that is supposed to be the child</param>
		/// <returns>Returns <c>true</c> if <paramref name="father"/> is the father of <paramref name="son"/></returns>
		public bool IsFatherAndSon(WorkItem father, WorkItem son)
		{
			try
			{
				int pid = GetFieldParentWI(son);
				return father.Id == pid;
			}
			catch (Exception)
			{
				return false;
			}
		}

		///////////////////////////////////////////////////////////////////////////////////////////////////////////////
		//										CONSTRUCTOR AND INITIALIZATION
		///////////////////////////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>
		/// Constructor of the class
		/// </summary>
		/// <param name="wis">The Work Item Store used for the operations</param>
		/// <param name="xmodel">The XML Network model to rely on</param>
		public WINetworkModel(WorkItemStore wis, XmlDocument xmodel)
		{
			m_Doc = xmodel;
			m_WIS = wis;

			// Default delegates
			GetParent = new DelGetParent(GetParentUsingWI);
			GetChildren = new DelGetChildren(GetChildrenUsingWI);
			FromObject = new DelFromObject(FromObjectUsingWI);
			CreateFromWI = new DelCreateFromWI(CreateFromWIUsingWI);

			m_WINetToSave = new WICache(wis, true, false);
			m_WIPool = new WICache(wis, false, true);

			InitTemplates();
		}

		/// <summary>
		/// Set the delegates responsible for hierarchy parsing
		/// </summary>
		/// <param name="p1">The GetParent delegate</param>
		/// <param name="p2">The GetChildren delegate</param>
		/// <param name="p3">The FromObject delegate</param>
		public void SetDelegates(DelGetParent p1, DelGetChildren p2, DelFromObject p3, DelCreateFromWI p4)
		{
			GetParent = new DelGetParent(p1);
			GetChildren = new DelGetChildren(p2);
			FromObject = new DelFromObject(p3);
			CreateFromWI = new DelCreateFromWI(p4);
		}

		/// <summary>
		/// Set the Work Item Project that will be used
		/// </summary>
		/// <param name="wip">The project to use</param>
		public void SetWIProject(Project wip)
		{
			m_WIP = wip;
		}

		protected void InitTemplates()
		{
			if (m_Doc == null)
			{
				return;
			}
			XmlNodeList tpll = m_Doc.SelectNodes("/WIModel/WITemplates/WITemplate");

			// For each XML Element
			foreach (XmlElement xe in tpll)
			{
				try
				{
					m_WITemplates.Add(xe.GetAttribute("name"), new WITemplate(xe.GetAttribute("name"), xe.GetAttribute("info"), xe.GetAttribute("projects"), xe));
				}
				catch (Exception)
				{
				}
			}
		}

		public Object ApplyTemplate(WITemplate tpl, Object parent)
		{
			// Create the Work Item, copy some fields from its parent
			return ProcessTemplate(parent, tpl.Description);

			//WorkItem wi = m_NetModel.m_WIPool.Create(wit);

			//wi.Title = String.Format("{0} #{1}", wit.Name, wicounter++);
			//wi.IterationPath = wiparent.IterationPath;
			//wi.IterationId = wiparent.IterationId;
			//wi.AreaPath = wiparent.AreaPath;
			//wi.AreaId = wiparent.AreaId;

			//// Create a WIC for the new WorkItem
			//WICObject wic = new WICObject(m_NetModel, WICObjType.DBObj, wi);
			//UpdateDirty(wic, UpdateMode.UM_Modify, false);
			//wicparent.AddChild(wic);
			//UpdateWITree(wicparent, false, true);
			//m_WITree.SelectedNode = wic.TN;
		}

		private Object ProcessTemplate(Object parent, XmlElement elem)
		{
			Object returnobj = null;
			WorkItem wiparent = FromObject(parent);

			XmlNodeList wilist = elem.SelectNodes("WorkItem");
			foreach (XmlElement curelem in wilist)
			{
				WorkItemType wit = m_WIP.WorkItemTypes[curelem.GetAttribute("type")];
				WorkItem wi = m_WIPool.Create(wit);

				wi.IterationPath = wiparent.IterationPath;
				wi.IterationId = wiparent.IterationId;
				wi.AreaPath = wiparent.AreaPath;
				wi.AreaId = wiparent.AreaId;

				XmlNodeList flist = curelem.SelectNodes("Fields/Field");
				foreach (XmlElement curfield in flist)
				{
					wi.Fields[curfield.GetAttribute("name")].Value = curfield.GetAttribute("value");
				}

				Object obj = CreateFromWI(wi, parent);
				if (returnobj == null)
				{
					returnobj = obj;
				}

				ProcessTemplate(obj, curelem);

			}
			return returnobj;
		}



		///////////////////////////////////////////////////////////////////////////////////////////////////////////////
		//												HIERARCHY
		///////////////////////////////////////////////////////////////////////////////////////////////////////////////

		/// <summary>
		/// Determine if the given Work Item is a root one
		/// </summary>
		/// <param name="wi">The Work Item</param>
		/// <returns>Returns <c>true</c> if the WorkItem is a standalone (ParentWI=0) or root (ParentWI=-1)</returns>
		public bool IsRootWorkItem(WorkItem wi)
		{
			return (GetFieldParentWI(wi) == 0) || (GetFieldParentWI(wi) == -1);
		}

		/// <summary>
		/// Return true if the given Work Item is the first in its hierarchy level
		/// </summary>
		/// <param name="wi">The Work Item</param>
		/// <returns><c>true</c> if the Work Item is the first, <c>false</c> otherwise</returns>
		public bool IsFirstChild(WorkItem wi)
		{
			WorkItem parentwi = GetParentWorkItem(wi);
			if (parentwi == null)
			{
				return false;
			}

			return GetFieldFirstChildWI(parentwi) == wi.Id;
		}

		/// <summary>
		/// Return true if the given Work Item is the last in its hierarchy level
		/// </summary>
		/// <param name="wi">The Work Item</param>
		/// <returns><c>true</c> if the Work Item is the last, <c>false</c> otherwise</returns>
		public bool IsLastChild(WorkItem wi)
		{
			return GetFieldNextWI(wi) == 0;
		}

		/// <summary>
		/// Get the Parent of the given Work Item
		/// </summary>
		/// <param name="wi">The Work Item to get the parent</param>
		/// <returns>The Parent Work Item or <c>null</c> if there's no parent</returns>
		/// <remarks>This method uses the WI read-only cache</remarks>
		public WorkItem GetParentWorkItem(WorkItem wi)
		{
			return GetWorkItem(GetFieldParentWI(wi));
		}

		/// <summary>
		/// Get the First Child of th given Work Item
		/// </summary>
		/// <param name="wi">The Work Item to get its first child</param>
		/// <returns>The First Child or <c>null</c> if there's no children</returns>
		/// <remarks>This method uses the WI read-only cache</remarks>
		public WorkItem GetFirstChildWorkItem(WorkItem wi)
		{
			return GetWorkItem(GetFieldFirstChildWI(wi));
		}

		/// <summary>
		/// Get the Next of the given Work Item
		/// </summary>
		/// <param name="wi">The Work Item to get the next one</param>
		/// <returns>The Next Work Item or <c>null</c> if <paramref name="wi"/> is the last</returns>
		/// <remarks>This method uses the WI read-only cache</remarks>
		public WorkItem GetNextWorkItem(WorkItem wi)
		{
			return GetWorkItem(GetFieldNextWI(wi));
		}

		/// <summary>
		/// Get the Previous of the given Work Item
		/// </summary>
		/// <param name="wi">The Work Item to get the previous one</param>
		/// <returns>The Previous Work Item or <c>null</c> if <paramref name="wi"/> is the first</returns>
		/// <remarks>This method uses the WI read-only cache</remarks>
		public WorkItem GetPrevWorkItem(WorkItem wi)
		{
			// Checking
			if (wi == null)
			{
				return null;
			}

			// Get the parent
			WorkItem parentwi = GetParentWorkItem(wi);
			if (parentwi == null)
			{
				return null;
			}

			// Get the first child, and parse its siblings until we find one having its 'nextwi' equals to wi.id
			WorkItem curwi = GetFirstChildWorkItem(parentwi);
			while (GetFieldNextWI(curwi) != wi.Id)
			{
				curwi = GetNextWorkItem(curwi);
				if (curwi == null)
				{
					return null;
				}
			}

			return curwi;
		}

		/// <summary>
		/// Get the last child Work Item of the given one
		/// </summary>
		/// <param name="wi">The Work Item to get its last child</param>
		/// <returns>The last child</returns>
		public WorkItem GetLastWorkItem(WorkItem wi)
		{
			// Checking
			if (wi == null)
			{
				return null;
			}

			// Get the first child, and parse its siblings until we find one having its 'nextwi' equals to wi.id
			WorkItem curwi = GetFirstChildWorkItem(wi);
			while (GetFieldNextWI(curwi) != 0)
			{
				curwi = GetNextWorkItem(curwi);
				if (curwi == null)
				{
					return null;
				}
			}

			return curwi;
		}

		/// <summary>
		/// Action to perform following the operation
		/// </summary>
		public enum WMAction
		{
			/// <summary>
			/// Nothing to do
			/// </summary>
			WMA_NOTHING = 1,

			/// <summary>
			/// Save the changed Work Items
			/// </summary>
			WMA_SAVE = 2,

			/// <summary>
			/// Add the changed Work Items to the ToRecycle list
			/// </summary>
			WMA_TORECYCLE = 3
		}

		/// <summary>
		/// Unlink two Work Items
		/// </summary>
		/// <param name="sonwi">The son Work Item to unlink from its parent</param>
		/// <param name="action">Additional action to perform</param>
		/// <returns>Success code</returns>
		/// <exception cref="System.Exception"> Thrown with the following error codes: 
		/// <list type="bullet">
		/// <item><description>CWC0004: Hierarchy problem, can't unlink properly</description></item>
		/// </list>
		/// </exception>
		public void UnlinkWorkItem(WorkItem sonwi, WMAction action)
		{
			// Get the Parent's ID
			int parentid = GetFieldParentWI(sonwi);

			// Not linked, we quit, quietly...
			if (parentid == 0)
			{
				return;
			}

			// We update right away the hierarchy Fields to make sure it's done even if there's a problem afterward
			// Two cases:

			// 1) The Work Item is the first 
			try
			{
				if (IsFirstChild(sonwi))
				{
					// Get the parent
					WorkItem curparentwi = GetParentWorkItem(sonwi);

					// Set the "FirstChildWI" of the parent to "sonwi.NextWI"
					SetFieldFirstChildWI(curparentwi, GetFieldNextWI(sonwi));

					if (action == WMAction.WMA_SAVE)
					{
						curparentwi.Save();
					}
					else if (action == WMAction.WMA_TORECYCLE)
					{
						AddToRecycle(curparentwi);
					}
				}

				// 2) Not the first
				else
				{
					// Get the Previous Work Item
					WorkItem prevwi = GetPrevWorkItem(sonwi);

					// Change its NextWI to sonwi.NextWI
					SetFieldNextWI(prevwi, GetFieldNextWI(sonwi));

					if (action == WMAction.WMA_SAVE)
					{
						prevwi.Save();
					}
					else if (action == WMAction.WMA_TORECYCLE)
					{
						AddToRecycle(prevwi);
					}
				}
			}
			catch (Exception)
			{
				//There's a problem in the hierarchy, we set the ParentWI to 0, save, and throw an exception
				SetFieldParentWI(sonwi, 0);
				sonwi.Save();
				throw new Exception(String.Format("CWC0004: Work Item {0} has a hierarchy problem, so it can't be unlinked", sonwi.Title));
			}

			// Reset the fields
			SetFieldParentWI(sonwi, 0);
			SetFieldNextWI(sonwi, 0);

			// Save ?
			if (action == WMAction.WMA_SAVE)
			{
				sonwi.Save();
			}
			else if (action == WMAction.WMA_TORECYCLE)
			{
				AddToRecycle(sonwi);
			}
			return;
		}

		/// <summary>
		/// Link two Work Items
		/// </summary>
		/// <param name="parent">The object that will be the parent</param>
		/// <param name="son">The object that will be the son</param>
		/// <returns>Success flag</returns>
		/// <exception cref="System.Exception"> Thrown with the following error codes: 
		/// <list type="bullet">
		/// <item><description>CWC0001: The Work Item has already a parent</description></item>
		/// <item><description>CWC0002: Cyclic linking attempted</description></item>
		/// <item><description>CWC0003: The parent/son relationship is not allowed</description></item>
		/// </list>
		/// </exception>
		public void CanLinkWorkItems(String project, Object parent, Object son)
		{
			WorkItem parentwi = FromObject(parent);
			WorkItem sonwi = FromObject(son);

			// Check for cyclic dependency
			Object cur = parent;
			while (cur != null)
			{
				WorkItem curwi = FromObject(cur);
				if (curwi.Id == sonwi.Id)
				{
					throw new Exception(String.Format("CWC0002: Cyclic linking attempt detected between {0} and {1}", parentwi.Title, sonwi.Title));
				}
				cur = GetParent(cur);
			}

			// We check the link's validity
			if (CheckLinkRules(project, parent, son) == false)
			{
				throw new Exception(String.Format("CWC0003: The relationship rules don't allow you to link {0} to {1}", sonwi.Title, parentwi.Title));
			}

			// We check that the Work Item has no parent
			if (GetFieldParentWI(sonwi) != 0)
			{
				throw new Exception(String.Format("CWC0001: Work Item {0} has already a parent", sonwi.Title));
			}
		}

		/// <summary>
		/// Check if <paramref name="parent"/> can be the parent of <paramref name="son"/>
		/// </summary>
		/// <param name="parent">The supposed parent </param>
		/// <param name="son">The supposed son</param>
		/// <returns>Returns <c>true</c> if <paramref name="parent"/> can be the parent of <paramref name="son"/></returns>
		public bool CheckLinkRules(String project, Object parent, Object son)
		{
			// Get all the types we can create children of for that parent
			ArrayList list = GetCreableWITs(project, parent);

			// Check if the son if of a valid type
			WorkItem sonwi = FromObject(son);
			foreach (WorkItemType wit in list)
			{
				if (sonwi.Type.Name == wit.Name)
				{
					return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Add a given Work Item as a child to another one.
		/// </summary>
		/// <param name="parent">The Parent Work Item</param>
		/// <param name="son">The Work Item to be the child</param>
		/// <param name="save">If <c>true</c> the changed fields of surrounding Work Items will be saved</param>
		/// <returns>Success flag</returns>
		public bool AddChildWorkItem(WorkItem parent, WorkItem son, bool save)
		{
			// Two cases, 1) the parent has no children, 2) the parent has some

			// 1)
			if (GetFieldFirstChildWI(parent) == 0)
			{
				SetFieldFirstChildWI(parent, son.Id);
				SetFieldParentWI(son, parent.Id);
				SetFieldNextWI(son, 0);
			}

			// 2)
			else
			{
				WorkItem lastwi = GetLastWorkItem(parent);
				if (lastwi == null)
				{
					return false;
				}

				SetFieldNextWI(lastwi, son.Id);
				SetFieldParentWI(son, parent.Id);
				SetFieldNextWI(son, 0);
			}

			//Save
			if (save)
			{
				if (parent.IsValid() == false)
				{
					return false;
				}
				parent.Save();


				if (son.IsValid() == false)
				{
					return false;
				}
				son.Save();
			}

			return true;
		}

		/// <summary>
		/// Determine if <paramref name="obj"/> has at least one child of the given type
		/// </summary>
		/// <param name="obj">The object to check the children</param>
		/// <param name="type">The type to look for</param>
		/// <returns>Returns <c>true</c> if there's at least one child of the <paramref name="type"/></returns>
		public bool HasSubWorkItemOfType(Object obj, String type)
		{
			// Get all the children for the given object
			IEnumerable ie = GetChildren(obj);

			// Parse them
			foreach (Object cur in ie)
			{
				WorkItem curwi = FromObject(cur);

				// Is the given type the current's one ?
				if (curwi.Type.Name == type)
				{
					return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Get the count of Work Item of the given type under the given one
		/// </summary>
		/// <param name="obj">The object representing the Work Item to count the children</param>
		/// <param name="type">The type to check for</param>
		/// <returns>The count of Work Item of the <paramref name="type"/> type</returns>
		public int GetCountOfWorkItemOfType(Object obj, String type)
		{
			// Initialize the counter
			int count = 0;

			// Get all the children
			IEnumerable ie = GetChildren(obj);

			// Parse them
			foreach (Object cur in ie)
			{
				WorkItem curwi = FromObject(cur);

				// If the current is of the given type, one more child to count
				if (curwi.Type.Name == type)
				{
					count++;
				}
			}

			return count;
		}

		/// <summary>
		/// Get a list of types of Work Item that can be created under the given object, according to the model rules.
		/// </summary>
		/// <param name="obj">The obj to check.</param>
		/// <returns>A String list of WorkItemType instances corresponding to each type of Work Item that can be created</returns>
		/// <remarks>You can't assume you can create as many instances you want if the type is list, there's also a constraint
		/// of count in the rules definition</remarks>
		public ArrayList GetCreableWITs(string project, Object obj)
		{
			WorkItem wi = (obj == null) ? null : FromObject(obj);
			IEnumerable childrend = GetChildren(obj);
			ArrayList witc = new ArrayList();

			// Get the model corresponding to the type of the given Work Item
			String xpath = String.Format("/WIModel/WIHierarchy/WorkItem[@type=\"{0}\"]", (wi == null) ? "" : wi.Type.Name);
			XmlNodeList witlist = m_Doc.SelectNodes(xpath);
			XmlElement witr = null;
			for (int i = 0; i < witlist.Count; i++)
			{
				XmlElement cur = witlist[i] as XmlElement;
				string projectlist = "$" + cur.GetAttribute("projects") + "$";
				if ((projectlist == "$$") || projectlist.Contains("$" + project + "$"))
				{
					witr = cur;
					break;
				}
			}

			// Is there something to do ?
			if (witr != null)
			{
				XmlNodeList wits = witr.SelectNodes("Allowed/WorkItem");
				foreach (XmlNode node in wits)
				{
					XmlElement wit = node as XmlElement;
					if (wit == null) continue;

					// Get the type of the Work Item in process
					String witname = wit.GetAttribute("type");
					if (witname.Length == 0) continue;

					// Get the 'max' attribute, if it's not specified, there's no limit, so -1 is set
					int maxallowed = -1;
					String attrb = wit.GetAttribute("max");
					if (attrb.Length > 0)
					{
						maxallowed = int.Parse(attrb);
					}

					// Check if the count of instance for that type is reached
					if ((maxallowed != -1) && (GetCountOfWorkItemOfType(obj, witname) >= maxallowed))
					{
						continue;
					}

					// Exception section ?
					XmlNodeList exps = wit.SelectNodes("Exceptions/WorkItem");
					bool failed = false;

					foreach (XmlElement elem in exps)
					{
						String extype = elem.GetAttribute("type");
						int exmin = 1;

						String tpsstr = elem.GetAttribute("min");
						if (tpsstr.Length > 0)
						{
							exmin = int.Parse(tpsstr);
						}

						int excount = GetCountOfWorkItemOfType(obj, extype);

						if (excount >= exmin)
						{
							failed = true;
							break;
						}
					}

					// At least one exception has been encounter, we can't create that kind of Work item
					if (failed == true)
					{
						continue;
					}

					if (m_WIP.WorkItemTypes.Contains(witname))
					{
						WorkItemType witype = m_WIP.WorkItemTypes[witname];
						if (witype != null)
						{
							witc.Add(witype);
						}
					}
				}
			}
			return witc;
		}

		/// <summary>
		/// Detect if the changed field is Time Tracking one, and update the whole hierarchy if the threshold was reached
		/// 
		/// </summary>
		/// <param name="wi"></param>
		/// <returns></returns>
		public String UpdateHierarchyWorkers(XmlDocument docwi, WorkItem wi, XmlDocument xmlts, String tsunc)
		{
			// Set the current context
			m_DocWI = docwi;
			String res = "";
			SetWIProject(wi.Project);

			// Get the list of the Hierarchy Workers to process
			XmlNodeList workers = m_Doc.SelectNodes("WIModel/HiearchyWorkers/HierarchyWorker");

			// Process the Workers to update the timestamp list if anyone match the trigger
			bool docchanged = false;
			foreach (XmlElement el in workers)
			{
				String workername = el.GetAttribute("name");

				// Look for the worker up into the timestamp file
				XmlElement wktsel = xmlts.SelectSingleNode(String.Format("TimeStamps/{0}", workername)) as XmlElement;

				// Already in the list ? Let's loop
				if (wktsel != null) continue;

				// Is the worker filtered by Team Project ?
				if (el.HasAttribute("projects"))
				{
					String tplist = "$" + el.GetAttribute("projects") + "$";
					if (tplist.Contains(String.Format("${0}$", wi.Project.Name)) == false)
					{
						continue;
					}
				}

				// Check if the Worker has for trigger the field that changed
				XmlNodeList triggers = el.SelectNodes("Triggers/Field");
				foreach (XmlElement trigger in triggers)
				{
					string fieldname = trigger.GetAttribute("name");

					string prevvalue = GetStringField(fieldname, false);
					string newvalue = GetStringField(fieldname, true);

					// The field is not present in this Work Item
					if (prevvalue == null && newvalue == null) continue;

					// Detect change from nothing to something, the opposite or value change
					if (prevvalue == null || newvalue == null || prevvalue != newvalue)
					{
						// The trigger is good put it in the Time Stamp list
						XmlElement tsel = xmlts.CreateElement(workername);
						xmlts.DocumentElement.AppendChild(tsel);

						DateTime now = DateTime.Now;
						tsel.SetAttribute("datetime", now.ToString());
						tsel.SetAttribute("tick", now.Ticks.ToString());

						docchanged = true;
						break;
					}
				}
			}

			// Save the XML Document if changed
			if (docchanged == true)
			{
				xmlts.Save(tsunc);
				docchanged = false;
			}

			// Parse again the worker and process the ones that reached the threshold
			foreach (XmlElement el in workers)
			{
				String workername = el.GetAttribute("name");

				// Look for the worker up into the timestamp file
				XmlElement wktsel = xmlts.SelectSingleNode(String.Format("TimeStamps/{0}", workername)) as XmlElement;

				// Nothing ? Loop to the next
				if (wktsel == null) continue;

				// Get the threshold (in minute)
				int th = int.Parse(el.GetAttribute("threshold"));

				// Get the DateTime in the timestamp file
				Int64 tick = Int64.Parse(wktsel.GetAttribute("tick"));
				DateTime dt = new DateTime(tick);

				// Add the threshold in minute
				dt = dt.AddMinutes(th);

				// Compare if the date time we got is after the current time, and loop if it's the case
				if (dt > DateTime.Now)
				{
					continue;
				}

				// Process the Worker
				ProcessHierarchyWorker(el, wi);

				// Remove the node from the XML File as it's processed
				wktsel.ParentNode.RemoveChild(wktsel);

				// Mark the XML TimeStamp document has changed
				docchanged = true;
			}

			// Save the XML Document if changed
			if (docchanged == true)
			{
				xmlts.Save(tsunc);
				docchanged = false;
			}

			m_DocWI = null;
			return res;
		}

		private void ProcessHierarchyWorker(XmlElement el, WorkItem wi)
		{
			XmlNodeList fields = el.SelectNodes("Fields/Field");
			foreach (XmlElement fel in fields)
			{
				string field = fel.GetAttribute("name");
				string op = fel.GetAttribute("operation");

				// Two cases: copy or anything else
				if (op != "copy")
				{
					// Find the root element and update its whole hierarchy
					WorkItem parent = GetParentWorkItem(wi);
					while (true)
					{
						WorkItem _p = GetParentWorkItem(parent);
						if (_p == null)
						{
							break;
						}
						parent = _p;
					}

					UpdateWorkItemTree(parent, op, field);
				}

				else if (op == "copy")
				{
					// Get the direction mode
					string dir = "," + fel.GetAttribute("direction") + ",";
					object valtorep;

					try
					{
						Field f = wi.Fields[field];
						valtorep = f.Value;
					}
					catch (Exception e)
					{
						e.ToString();
						continue;
					}

					// Replicate the value up to the hierarchy
					if (dir.Contains(",up,"))
					{
						ReplicateUp(wi, field, valtorep);
					}

					// Replicate the value to the siblings
					if (dir.Contains(",siblings,"))
					{
						ReplicateSibling(wi, field, valtorep);	
					}

					// Replicate the value down to children
					if (dir.Contains(",down,"))
					{
						ReplicateChildren(wi, field, valtorep);
					}			
				}
			}

			// Remove the Work Items we got during the process
			m_WIPool.Clear();
		}

		private void ReplicateUp(WorkItem wi, string field, object valtorep)
		{
			// Find the root element and update its whole hierarchy
			WorkItem parent = wi;
			while (true)
			{
				// Get the parent, break if we reached the top
				WorkItem _p = GetParentWorkItem(parent);
				if (_p == null)
				{
					break;
				}

				// Open, modify the Work Item and save it
				if (_p.Fields.Contains(field) && _p.Fields[field].IsEditable)
				{
					if (_p.IsOpen == false) _p.Open();
					Field _f = _p.Fields[field];
					_f.Value = valtorep;
					if (_p.IsDirty == true) _p.Save();
				}

				// Loop to the next parent
				parent = _p;
			}
		}

		private void ReplicateSibling(WorkItem wi, string field, object valtorep)
		{
			// Find the root element and update its whole hierarchy
			WorkItem parent = GetParentWorkItem(wi);
			WorkItem cur = GetFirstChildWorkItem(parent);
			while (cur != null)
			{
				if (cur.Fields.Contains(field) && cur.Fields[field].IsEditable)
				{
					// Open, modify the Work Item and save it
					if (cur.IsOpen == false) cur.Open();
					Field _f = cur.Fields[field];
					_f.Value = valtorep;
					if (cur.IsDirty == true) cur.Save();
				}

				// Loop to the next sibling
				cur = GetNextWorkItem(cur);
			}
		}

		private void ReplicateChildren(WorkItem wi, string field, object valtorep)
		{
			// Find the root element and update its whole hierarchy
			WorkItem cur = GetFirstChildWorkItem(wi);
			while (cur != null)
			{
				if (cur.Fields.Contains(field) && cur.Fields[field].IsEditable)
				{
					// Open, modify the Work Item and save it
					if (cur.IsOpen == false) cur.Open();
					Field _f = cur.Fields[field];
					_f.Value = valtorep;
					if (cur.IsDirty == true) cur.Save();
				}

				// Replication to children
				ReplicateChildren(cur, field, valtorep);

				// Loop to the next child
				cur = GetNextWorkItem(cur);
			}
		}

		private bool UpdateWorkItemTree(WorkItem wi, string op, string field)
		{
			try
			{
				// Check if the field exist and is editable
				if (wi.Fields.Contains(field)==false || wi.Fields[field].IsEditable==false)
				{
					return false;
				}

				// Check the field is either of type int or double
				Field f = wi.Fields[field];
				if ((f.FieldDefinition.FieldType != FieldType.Double) && (f.FieldDefinition.FieldType != FieldType.Integer))
				{
					return false;
				}

				// Get the value of the work item's children
				object obj = GetChildValues(wi, op, field);

				// Open, edit and save
				if (obj != null)
				{
					if (wi.IsOpen == false) wi.Open();
					wi.Fields[field].Value = obj;
					if (wi.IsDirty == true)	wi.Save();
				}
			}
			catch (Exception e)
			{
				e.ToString();
				return false;
			}

			return true;
		}

		private object GetChildValues(WorkItem wi, string op, string field)
		{
			if (wi.Id == 300)
			{
				int pipo = 0;
			}
			// Get the first child Work Item, and check if there's one
			WorkItem curwi = GetFirstChildWorkItem(wi);
			if (curwi == null)
			{
				if (wi.Fields.Contains(field)==false) return null;
				return wi.Fields[field].Value;
			}

			// Get the value of its field
			object curval = GetChildValues(curwi, op, field);

			//  Initialize the counter of occurence for avg
			int count = 1;

			// Loop constantly
			while (true)
			{
				// Get the next Work Item
				curwi = GetNextWorkItem(curwi);

				// Break of the loop if there's no more
				if (curwi == null) break;

				// Get the value of the next WI
				object val = GetChildValues(curwi, op, field);

				// Check if we've got nothing, loop if it's the case
				if (val == null) continue;

				// Check if there were initially a value
				if (curval == null)
				{
					curval = val;
					continue;
				}

				// Do the operation we've been asked for
				switch (op)
				{
					case "avg":
					case "sum":
						if (curval is int)
						{
							curval = (int)curval + (int)val;
						}
						else if (curval is double)
						{
							curval = (double)curval + (double)val;
						}

						break;

					case "min":
						if ((double)val < (double)curval) curval = val;
						break;

					case "max":
						if ((double)val > (double)curval) curval = val;
						break;
				}
				count++;
			}

			if (op == "avg")
			{
				if (curval is int) curval = (int)curval / count;
				else if (curval is double) curval = (double)curval / count;
			}

			if (wi.IsOpen == false) wi.Open();
			wi.Fields[field].Value = curval;
			if (wi.IsDirty == true)	wi.Save();

			return curval;
		}


		private XmlDocument m_DocWI = null;

		/// <summary>
		/// Get a field of type <c>string</c> in the WorkItemChangedEvent XML feed.
		/// </summary>
		/// <param name="name">Name of the field to get the value</param>
		/// <param name="newversion">If <c>true</c> returns the new version of the field, if <c>false</c> the old
		/// one will be.</param>
		/// <returns>The value corresponding to <c>name</c>, or an empty string if the field doesn't exists</returns>
		private string GetStringField(String name, bool newversion)
		{
			try
			{
				// Let's build the XPath request to get the field element
				String xpath = String.Format(@"WorkItemChangedEvent/*/StringFields/Field[ReferenceName='{0}']/{1}", name, newversion ? "NewValue" : "OldValue");
				return m_DocWI.SelectSingleNode(xpath).InnerText;
			}
			catch (Exception)
			{
				return String.Empty;
			}
		}

		/// <summary>
		/// Get a field of type <c>integer</c> in the WorkItemChangedEvent XML feed.
		/// </summary>
		/// <param name="name">Name of the field to get the value</param>
		/// <param name="newversion">If <c>true</c> returns the new version of the field, if <c>false</c> the old
		/// one will be.</param>
		/// <returns>The value corresponding to <c>name</c>, or 0 if the field doesn't exists</returns>
		private int GetIntField(String name, bool newversion)
		{
			try
			{
				// Let's build the XPath request to get the field element
				String xpath = String.Format(@"WorkItemChangedEvent/*/IntegerFields/Field[ReferenceName='{0}']/{1}", name, newversion ? "NewValue" : "OldValue");
				return int.Parse(m_DocWI.SelectSingleNode(xpath).InnerText);
			}
			catch (Exception)
			{
				return 0;
			}
		}



		/// <summary>
		/// Process the Models for the given Work Item, executing the actions where possible
		/// </summary>
		/// <param name="wi">The Work Item to process</param>
		/// <returns>A result string for information purpose</returns>
		public String ProcessModels(WorkItem wi)
		{
			String res = "";
			SetWIProject(wi.Project);

			try
			{
				// We get all available models
				XmlNodeList models = m_Doc.SelectNodes("WIModel/NetworkModels/NetworkModel");

				// and process each one of them
				foreach (XmlElement model in models)
				{
					string projectlist = "$" + model.GetAttribute("projects") + "$";
					if ((projectlist == "$$") || (projectlist.Contains(wi.Project.Name)))
					{
						ProcessModel(model, wi, ref res);
					}
				}

				// Save the Work Item that were modified
				m_WINetToSave.Save(ref res);

			}

			// If an exception was raised, return the error string
			catch (Exception ex)
			{
				res += ex.ToString();
			}

			return res;
		}

		/// <summary>
		/// Current Model in processing
		/// </summary>
		public String m_CurModelName;

		/// <summary>
		/// Current State in processing
		/// </summary>
		public String m_CurStateName;

		/// <summary>
		/// Process the given model on the given work item
		/// </summary>
		/// <param name="model">Model definition</param>
		/// <param name="wi">Work Item to process</param>
		/// <param name="res">Result string for information purpose</param>
		void ProcessModel(XmlElement model, WorkItem wi, ref String res)
		{
			// Name of the model
			m_CurModelName = model.GetAttribute("name");

			// We evaluate all the types
			XmlNodeList xwits = model.SelectNodes("WorkItem");

			foreach (XmlElement xwit in xwits)
			{
				// Get the type
				StringCollection wits = new StringCollection();
				String curtype = "";
				try
				{
					String stypes = xwit.GetAttribute("type");
					wits.AddRange(stypes.Split(new Char[] { '$' }));

					// We check if every returned types are existing in the Project
					// An exception will be thrown if it's not the case
					foreach (String type in wits)
					{
						curtype = type;
						WorkItemType wit = m_WIP.WorkItemTypes[type];
					}
				}
				catch (Exception ex)
				{
					// Error output
					res += String.Format("ERROR: Unknown Work Item Type {0} for the model {1}\r\n", curtype, m_CurModelName);
					ex.ToString();

					// Process the next WIT
					continue;
				}

				// We evaluate all the states for that type
				XmlNodeList xstates = xwit.SelectNodes("States/State");
				foreach (XmlElement xstate in xstates)
				{
					// State name
					m_CurStateName = xstate.GetAttribute("name");

					// Trigger
					String trigger = xstate.GetAttribute("trigger");

					// The trigger tells us what Work Item(s) to process

					// The Work Item that triggered the alert is supposed to be a child, so we analyse its parent
					if ((trigger == "") || (trigger == "Children"))
					{
						// Get the parent
						WorkItem wiparent = WIGetParent(wi);

						// Process it
						ProcessState(wiparent, xstate, wits, ref res);
					}

					// Current kind of trigger, we use the Work Item that triggered the alert
					else if (trigger == "Current")
					{
						ProcessState(wi, xstate, wits, ref res);
					}

					// Parent kind of trigger, we process all its children, one by one
					else if (trigger == "Parent")
					{
						IEnumerable wis = WIGetChildren(wi);

						foreach (WorkItem cur in wis)
						{
							ProcessState(cur, xstate, wits, ref res);
						}

					}
					else
					{
						res += String.Format("ERROR: Unknown trigger {0} for the model {1}/{2}\r\n", trigger, m_CurModelName, m_CurStateName);
					}
				}
			}
		}

		/// <summary>
		/// Current Work Item
		/// </summary>
		/// <value>Set/get the Work Item that is in process</value>
		WorkItem m_CurWI
		{
			set
			{
				// No change ?
				if ((_CurWI != null) && (value != null) && (_CurWI.Id == value.Id))
				{
					// Same Work Item or no Work Item, so quit immediately
					return;
				}

				// Update the variable
				_CurWI = value;

				// If there's a new Work Item
				if (_CurWI != null)
				{
					// We immediately get its parent
					_CurParentWI = WIGetParent(_CurWI);
				}
				else
				{
					_CurParentWI = null;
				}
				_CurChildrenWI = null;					// Empty the children list
				_CurChildrenEval = false;				// Flag we didn't get the children for the current WI
				// It will be done the first time we ask for them because
				// it's a VERY costly operation
			}
			get
			{
				return _CurWI;
			}
		}

		/// <summary>
		/// Parent of the current Work Item
		/// </summary>
		/// <value>Get the parent of the current Work Item</value>
		WorkItem m_CurParentWI
		{
			get
			{
				return _CurParentWI;
			}
		}

		/// <summary>
		/// Children of the current Work Item
		/// </summary>
		/// <value>Get the children of the current Work Item</value>
		IEnumerable m_CurChildrenWI
		{
			get
			{
				// Not the first time we access the children for the current Work Item ?
				if (_CurChildrenEval == true)
				{
					// Return the list
					return _CurChildrenWI;
				}

				// First time we want the children for that Work Item, let get them
				_CurChildrenWI = WIGetChildren(m_CurWI);
				_CurChildrenEval = true;						// and flag it

				return _CurChildrenWI;
			}
		}

		/// <summary>
		/// Current Work Item
		/// </summary>
		/// <remarks>Used by the <see cref="m_CurWI"/> property</remarks>
		private WorkItem _CurWI = null;

		/// <summary>
		/// Parent of the current Work Item
		/// </summary>
		/// <remarks>Used by the <see cref="m_CurParentWI"/> property</remarks>
		private WorkItem _CurParentWI = null;

		/// <summary>
		/// Children of the current Work Item
		/// </summary>
		/// <remarks>Used by the <see cref="m_CurChildrenWI"/> property</remarks>
		private IEnumerable _CurChildrenWI = null;

		/// <summary>
		/// Flag that indicates if the children of the current Work Item were already evaluated
		/// </summary>
		private bool _CurChildrenEval = false;

		/// <summary>
		/// Process a given state on a given Work Item
		/// </summary>
		/// <param name="wi">The WorkItem to process</param>
		/// <param name="xstate">The current state definition</param>
		/// <param name="wits">Possible types supported by the current model</param>
		/// <param name="res">Result string, for information purpose</param>
		/// <remarks>If the type of the Work Item is not supported, nothing is done.
		/// <para>This method will check if the Work Item is in the required state, then check the rules defined
		/// in the model, and if everything is correct it will execute the actions stored in the model</para></remarks>
		void ProcessState(WorkItem wi, XmlElement xstate, StringCollection wits, ref String res)
		{
			// First thing, we check if the Work Item is suitable for the current model
			if ((wi == null) || (wits.Contains(wi.Type.Name) == false))
			{
				return;
			}

			// Set the Work Item as the current one
			m_CurWI = wi;

			// Check if all the fields apply to the model definition
			if (CheckFields(wi, xstate.SelectNodes("Fields/Field"), ref res) == false)
			{
				// The field don't match, we quit
				return;
			}

			// Processing rules
			String rulelogop;
			XmlElement xrules = xstate.SelectSingleNode("Rules") as XmlElement;
			if (xrules == null)
			{
				res += String.Format("Missing the 'Rules' in {0}/{1}\r\n",
				m_CurModelName, m_CurStateName);
				return;
			}

			// Get the logical operation to perform among rules
			rulelogop = xrules.GetAttribute("logop");

			// No attribute ? Logical AND is the default one
			if (rulelogop == "")
			{
				rulelogop = "AND";
			}

			// Process the rules
			if (ProcessRules(wi, xstate.SelectNodes("Rules/Rule"), rulelogop, ref res) == false)
			{
				// Failed, we quit
				return;
			}

			// All the rules were satisfied, so action!
			res += String.Format("Executing {0}/{1} on Work Item [{2}] - {3}\r\n",
				m_CurModelName, m_CurStateName, wi.Id.ToString(), wi.Title);

			// Execute the actions states in the model definition
			ExecuteActions(wi, xstate.SelectNodes("Actions/Action"), ref res);
		}

		/// <summary>
		/// Check if the fields of the given Work Item.
		/// </summary>
		/// <param name="wi">The Work Item to check the fields</param>
		/// <param name="xfields">A list of Field elements that defines what fields to check and with what value</param>
		/// <param name="res">The result string, for information purpose</param>
		/// <returns>Returns <c>true</c> if all the fields are correct, <c>false</c> otherwise</returns>
		/// <remarks>This method will test all the fields stated in the XML Model against the values of the given Work Item</remarks>
		bool CheckFields(WorkItem wi, XmlNodeList xfields, ref String res)
		{
			foreach (XmlElement xfield in xfields)
			{
				Field field = null;
				String fname = "";

				// Get the field
				try
				{
					fname = xfield.GetAttribute("name");
					field = wi.Fields[fname];
				}

				// Unknown or invalid field...
				catch (Exception)
				{
					res += String.Format("Unknown field '{0}' in Work Item {1} of type {2}\r\n", fname, wi.Title, wi.Type.Name);
					return false;
				}

				// Check if the field is empty
				if (field.Value == null)
				{
					return false;
				}

				// We have to check the field's current value ?
				if (xfield.HasAttribute("current"))
				{
					// Get the value that is supposed to be
					String current = xfield.GetAttribute("current");

					// Compare against the current value of the Work Item
					if (CompareFieldValues(field.Value.ToString(), current) == false)
					{
						// Failed, quit in error
						return false;
					}
				}

				// We have to check the field's previous value ?
				if (xfield.HasAttribute("previous"))
				{
					String previous = xfield.GetAttribute("previous");
					String curfieldvalue = field.Value.ToString();

					// We look backward to find the first value that is different from the current one
					for (int i = wi.Revision - 1; i > 0; i--)
					{
						Field prev = wi.Revisions[i - 1].Fields[fname];
						if (prev.Value == null)
						{
							// Field is empty, quit in error
							return false;
						}

						// We got it ?
						String prevvalue = prev.Value.ToString();
						if (prevvalue != curfieldvalue)
						{
							// Check both field values
							if (CompareFieldValues(prevvalue, previous) == false)
							{
								// Failed, quit in error
								return false;
							}

							// We found the previous value and it matchs, break the loop
							break;
						}
					}
				}
			}
			return true;
		}

		/// <summary>
		/// Compare field values.
		/// </summary>
		/// <param name="wifield">Value of the field in the Work Item</param>
		/// <param name="modelfield">Value(s) of the field in the model definition. (see remarks)</param>
		/// <returns>Returns <c>true</c> if the fields match, <c>false</c> otherwise.</returns>
		/// <remarks><paramref name="modelfield"/> can have multiple values, they must be separated by the '$'
		/// character, without whitespace characters</remarks>
		bool CompareFieldValues(String wifield, String modelfield)
		{
			String[] modellist = modelfield.Split(new Char[] { '$' });

			foreach (String curmodelvalue in modellist)
			{
				if (curmodelvalue == wifield)
				{
					return true;
				}
			}
			return false;
		}

		/// <summary>
		/// Get the XML element corresponding to the type of the given Work Item
		/// </summary>
		/// <param name="wi">The Work Item to get the type from</param>
		/// <param name="parent">The XML element that contains all definitions of WITs</param>
		/// <returns>The XML element corresponding to the type of the given Work Item, or <c>null</c> if not found</returns>
		XmlElement GetWorkItemElem(WorkItem wi, XmlElement parent)
		{
			if (wi == null)
			{
				return null;
			}

			String witn = "$" + wi.Type.Name + "$";
			XmlNodeList wis = parent.SelectNodes("WorkItem");
			foreach (XmlElement curwi in wis)
			{
				String type = "$" + curwi.GetAttribute("type") + "$";
				if (type.Contains(witn))
				{
					return curwi;
				}
			}
			return null;
		}

		/// <summary>
		/// Process the rules define in the XML Model against the given Work Item
		/// </summary>
		/// <param name="wi">The Work Item to check</param>
		/// <param name="xrules">List of XML Rule elements</param>
		/// <param name="logop">Logical operation to perform between rules, can be "AND", "OR" or "XOR"</param>
		/// <param name="res">The result string, for information purpose</param>
		/// <returns>Returns <c>true</c> if the rules were validated successfully against the given Work Item</returns>
		/// <remarks>
		/// <para>A rule can't succeed if there's no Work Item processed</para>
		/// <para>This method will return <c>false</c> if no rule were processed</para>
		/// </remarks>
		bool ProcessRules(WorkItem wi, XmlNodeList xrules, String logop, ref String res)
		{
			int successrules = 0;		// Counter that gives the count of rules that were successfully processed
			int currulecount = 0;		// Index of the rule in process
			bool prevruleres = false;	// Result of the previous rule
			bool curruleres = false;	// Result of the current rule

			// We evaluate all the rules, their result are compared using the given logical operator (logop)
			foreach (XmlElement xrule in xrules)
			{
				String rulename = xrule.GetAttribute("name");

				// "OneChild" and "AllChildren" are processed in the same code block
				//   - OneChild: at least one children must comply to the rule
				//   - AllChildren: all the children must comply to the rule
				if ((rulename == "OneChild") || (rulename == "AllChildren"))
				{
					bool onechild = (rulename == "OneChild");
					curruleres = !onechild;				// false in OneChild mode, true in AllChildren
					int wicount = 0;						// Count of processed Work Item

					foreach (WorkItem curchild in m_CurChildrenWI)
					{
						// We get the XML description for the type of the Work Item in process
						XmlElement xwit = GetWorkItemElem(curchild, xrule);

						// No description, no problem, this Work Item is ignored by the rule, we keep going
						if (xwit == null)
						{
							continue;
						}

						// The Work Item's type is stated in the rule, so we count it as one more analyzed
						wicount++;

						// We check the fields:
						//  - OneChild: If CheckFields returns true, we can quit successfully right away
						//  - AllChildre: If CheckFields return false, we can quit in failure right away
						if (CheckFields(curchild, xwit.SelectNodes("State/Field"), ref res) == onechild)
						{
							curruleres = onechild;
							break;
						}
					}

					// If nothing was processed, the rule is considered as failed
					if (wicount == 0)
					{
						curruleres = false;
					}
				}

				// "Parent" mode, the parent must comply with the rule
				else if (rulename == "Parent")
				{
					// We get the XML description for the type of the Work Item in process
					XmlElement xwit = GetWorkItemElem(m_CurParentWI, xrule);

					// No description, so it's a failure because the parent doesn't suit for that rule
					if (xwit == null)
					{
						// Flag the failure
						curruleres = false;
					}

					// Check the fields and store the result
					else
					{
						curruleres = CheckFields(m_CurParentWI, xwit.SelectNodes("State/Field"), ref res);
					}
				}

				// Check the "AND" LogOp here
				if (logop == "AND")
				{
					// If the rule fails, we fail right now
					if (curruleres == false)
					{
						return false;
					}
				}

				// Check the "XOR" LogOp
				if (logop == "XOR")
				{
					// We check for it only if it's not the first rule (because we check it against the previous rule's result)
					if (currulecount > 0)
					{
						// Check XOR invalidity
						if ((prevruleres == true) && (curruleres == true))
						{
							return false;
						}
					}
				}

				// Store the current result in the previous one
				prevruleres = curruleres;

				// One more rule processed
				currulecount++;

				// If the rule succeed, we inc the corresponding counter
				if (curruleres)
				{
					successrules++;
				}

				// Loop to the next rule
			}

			// For a success to be returned, at least one rule had to succeed
			if (successrules == 0)
			{
				return false;
			}

			// "OR" LogOp
			if (logop == "OR")
			{
				// At least one rule had to succeed
				return successrules > 0;
			}

			// In case of AND and XOR, if we survived until here, it's good !
			return true;
		}

		/// <summary>
		/// Execute the actions
		/// </summary>
		/// <param name="wi">Work Item to execute the actions on</param>
		/// <param name="xactions">List of Action elements</param>
		/// <param name="res">The result string, for information purpose</param>
		void ExecuteActions(WorkItem wi, XmlNodeList xactions, ref String res)
		{
			// We execute all the actions
			foreach (XmlElement xaction in xactions)
			{
				String actionname = xaction.GetAttribute("name");
				if (actionname != "")
				{
					WorkItem modwi = m_WINetToSave.Get(wi.Id);
					if (modwi.History == null)
					{
						modwi.History = "WorkItem Network solver: " + actionname;
					}
					else
					{
						modwi.History += "\r\n" + "WorkItem Network solver: " + actionname;
					}
				}

				// Process all the elements
				foreach (XmlNode xcur in xaction.ChildNodes)
				{
					XmlElement xa = xcur as XmlElement;
					if (xa == null)
					{
						continue;
					}

					// Action of type Field ?
					if (xa.Name == "Field")
					{
						// Get the Field's name
						String fname = xa.GetAttribute("name");

						// Get the Field's value
						String value = xa.GetAttribute("newvalue");

						// Set it
						SetField(wi, fname, value, ref res);
					}
				}
			}
		}

		/// <summary>
		/// Change the value of a field of a given Work Item
		/// </summary>
		/// <param name="wi">The Work Item to change</param>
		/// <param name="name">The name of the field to change</param>
		/// <param name="value">The new value</param>
		/// <param name="res">The result string, for information purpose</param>
		/// <returns>Success flag</returns>
		/// <remarks>Note that the <paramref name="wi"/> instance won't be modifie, a new Work Item instance will be
		/// retrieved from the <see cref="m_WINetToSave"/>
		/// <para>The cache will be save when all the models will be executed</para>
		/// </remarks>
		bool SetField(WorkItem wi, String name, String value, ref String res)
		{
			WorkItem modwi = m_WINetToSave.Get(wi.Id);
			Field f = null;

			try
			{
				f = modwi.Fields[name];
			}
			catch (Exception)
			{
				res += String.Format("ERROR The field {0} in Work Item of type {1} doesn't exist\r\n", name, modwi.Type.Name);
				return false;
			}

			if (f.IsEditable == false)
			{
				res += String.Format("ERROR The field {0} in Work Item of type {1} can't be edited in the current state\r\n", name, modwi.Type.Name);
				return false;
			}

			// We assign the value depending of field's type (in a try block because formatting can lead to exceptions)
			try
			{
				switch (f.FieldDefinition.FieldType)
				{
					// String, easy....
					case FieldType.String:
						f.Value = value;
						break;

					// Plain Text, just like a string
					case FieldType.PlainText:
						f.Value = value;
						break;

					// Integer
					case FieldType.Integer:
						int i = int.Parse(value);
						f.Value = i;
						break;

					// Double
					case FieldType.Double:
						Double d = Double.Parse(value);
						f.Value = d;
						break;
				}
			}
			catch (Exception)
			{
				res += String.Format("ERROR The value {0} is of a wrong type for the field {1} in Work Item of type {2}\r\n",
					value, name, modwi.Type.Name);
				return false;
			}
			return true;
		}
	}
}
